//	CurvedSpaces-Common.h
//
//	Platform-independent definitions for Curved Spaces' internal code.
//	The internal code doesn't know or care what platform (Win32 or MacOS)
//	it's running on.
//	However, the platform-specific user interface *does* include
//	the present file.  For that reason we take care not to use any
//	OpenGL-specific definitions (i.e. nothing from gl.h) because
//	the user interface doesn't know about OpenGL.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#pragma once

#include <stdlib.h>	//	for malloc() and free()
#include <stdbool.h>
#include "Matrix44.h"

//	Acknowledge unused parameters.
#define UNUSED_PARAMETER(x)	((void)(x))

//	Define a byte.
typedef unsigned char	Byte;

//	Internal functions pass error messages as pointers
//	to zero-terminated UTF-8 strings.  The strings are defined
//	statically, so the caller need not (must not!) free them.
typedef const char		*ErrorText;

//	Define GET_MEMORY and FREE_MEMORY macros to manage memory so
//	that in future we can easily change memory allocation schemes.
//	Use a variable gMemCount to test for memory leaks.
//	Define FREE_MEMORY_SAFELY() to test for and set NULL pointers.
#define GET_MEMORY(n)			(gMemCount++, malloc(n))
#define FREE_MEMORY(p)			(gMemCount--, free(p))
#define FREE_MEMORY_SAFELY(p)	{if (p != NULL) {FREE_MEMORY(p); p = NULL;}}

//	BUFFER_LENGTH() measures the number of items in an array,
//	not the number of bytes, and automatically adjusts to changes
//	in the number of elements or the size of each element.
#define BUFFER_LENGTH(a)	( sizeof(a) / sizeof((a)[0]) )

//	Macro to compare two 2-letter language codes.
#define SAME_LANGUAGE_CODE(A,B)	((A)[0] == (B)[0] && (A)[1] == (B)[1])

//	When the 3-torus is first introduced in the Shape of Space lecture,
//	it's nice to start with walls closed and zero speed.
//#define START_STILL

//	For the Shape of Space presentation it’s nice
//	to be able to move the Earth.  I've suppressed this feature
//	from the public version because it introduces artifacts:
//	1.	The centerpiece may extend beyond the fundamental domain,
//		leading to a "popping" artifact when the fundamental domain
//		passes outside the view pyramid and no longer gets drawn.
//	2.	The centerpiece also suffers a noticeable "popping" artifact
//		when it's near the back of the visible region in the Euclidean
//		case.  In such cases the beams also "pop" in theory, but in
//		practice they're not noticeable.
//	3.	The wrong resolution might get used for the spinning Earth,
//		because it's no longer at the center.  DrawEarthVBOs()
//		include a hack to render the Earth more smoothly,
//		so even an off-center Earth looks OK.  (This feature is
//		for my personal use only. Release-quality code would decide
//		the Earth's resolution based on its true position in world space.)
//#define CENTERPIECE_DISPLACEMENT

//	For the Shape of Space presentation it’s nice
//	to start outside the dodecahedron and then move in.
//	I’m not sure whether this will work well
//	in the publicly released version or not, partly because
//	of conflicts with existing features and partly because
//	it might not look so good with spaces like slab spaces.
//	So for now let’s keep it optional.
//	We could eventually make it permanent if desired.
//#define START_OUTSIDE

//	For the Curvature presentation it's nice to start 
//	with the walls mostly open, leaving beams to represent
//	the geodesics (at least in spaces where the beams
//	continue straight through the vertices and out the other side).
//	This option also leaves the spaceship hidden by default.
//#define START_WALLS_OPEN

//	To produce a good screenshot, tile deep (can also omit the centerpiece
//	in the home cell to avoid obscuring the view) (might also want
//	to improve the level-of-detail of the spinning Earths).
//	[2009/10/28 JRW  This option is currently unmaintained.  Next time I need
//	a high-resolution screenshot, I'll have to adjust the relevant settings.]
//#define HIGH_RESOLUTION_SCREENSHOT


//	Assign the vertex attributes to well-defined locations in the vertex shader.
//	ATTRIBUTE_POSITION must bind to location 0, but the remaining choices are arbitrary.
#define ATTRIBUTE_POSITION				 0
#define ATTRIBUTE_TEX_COORD				 1
#define ATTRIBUTE_COLOR					 2
#define ATTRIBUTE_MVP_MATRIX_ROW_0		 3
#define ATTRIBUTE_MVP_MATRIX_ROW_1		 4
#define ATTRIBUTE_MVP_MATRIX_ROW_2		 5
#define ATTRIBUTE_MVP_MATRIX_ROW_3		 6
#define ATTRIBUTE_MV_MATRIX_ROW_0		 7
#define ATTRIBUTE_MV_MATRIX_ROW_1		 8
#define ATTRIBUTE_MV_MATRIX_ROW_2		 9
#define ATTRIBUTE_MV_MATRIX_ROW_3		10

//	Keep an array of shader programs, each referenced by an unsigned integer 
//	that glCreateProgram() provides to refer to the given texture.
//	Each program contains a vertex shader and a fragment shader.
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.
typedef enum
{
	ShaderSph = 0,
	ShaderEuc,
	ShaderHyp,
	NumShaders
} ShaderIndex;

//	Keep an array of texture "names", each "name" being an unsigned integer 
//	that glGenTextures() generates to refer to the given texture.  
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.
typedef enum
{
	TextureWallPaper = 0,
	TextureWallWood,
	TextureEarth,
	TextureGalaxy,
	TextureGyroscope,
	TextureObserver,
	TextureVertexFigures,
	TextureClifford,
	NumTextures
} TextureIndex;

//	Keep an array of vertex buffer "names" and a parallel array 
//	of so-called index buffer "names", each "name" being an unsigned integer
//	that glGenBuffers() generates to refer to the given vertex or index buffer.  
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.  
typedef enum
{
	VertexBufferDirichlet = 0,
	VertexBufferEarth,
	VertexBufferGalaxy,
	VertexBufferGyroscope,
	VertexBufferObserver,
	VertexBufferVertexFigures,
	VertexBufferClifford,
	NumVertexBuffers
} VertexBufferIndex;


//	π
#define PI	3.14159265358979323846

//	The  up -arrow key decreases the user's speed by USER_SPEED_INCREMENT.
//	The down-arrow key increases the user's speed by USER_SPEED_INCREMENT.
//	The space bar sets the user's speed to zero.
#define USER_SPEED_INCREMENT	0.02


//	Opaque typedef
typedef struct HEPolyhedron	DirichletDomain;


//	Transparent typedefs

typedef enum
{
	SpaceNone,
    SpaceSpherical,
    SpaceFlat,
    SpaceHyperbolic
} SpaceType;

typedef enum
{
	ImagePositive,	//	not mirror-reversed
	ImageNegative	//	mirror-reversed
} ImageParity;

typedef struct
{
	double	v[4];
} Vector;

typedef struct
{
	double		m[4][4];
	ImageParity	itsParity;	//	Is determinant positive or negative?
} Matrix;

typedef struct
{
	unsigned int	itsNumMatrices;
	Matrix			*itsMatrices;
} MatrixList;

//	Technical note:  Why does a Honeycell use a Dirichlet domain's
//	full set of vertices instead of a bounding box?
//	1.	For the most common manifolds, the number of vertices is fairly small.
//	2.	Computing a bounding box is a small nuisance when the fundamental domain
//		may extend as far as -- or even into -- the southern hemisphere of S³,
//		as happens with lens spaces and slab spaces.
typedef struct
{
	Matrix			itsMatrix;
	Vector			itsCenter,
					*itsVertices;
	unsigned int	itsNumVertices;
	double			itsDistance;	//	distance from origin to cell center after applying view matrix
} Honeycell;
typedef struct
{
	//	A fixed list of the cells, sorted relative
	//	to their distance from the basepoint (0,0,0,1).
	unsigned int	itsNumCells;
	Honeycell		*itsCells;

	//	At render time, make a temporary list of the visible cells
	//	and sort them according to their distance from the observer.
	unsigned int	itsNumVisibleCells;
	Honeycell		**itsVisibleCells;
} Honeycomb;

typedef enum
{
	SceneryNone,
	SceneryEarth,
	SceneryGalaxy,
	SceneryGyroscope
} SceneryType;

#ifdef START_OUTSIDE
typedef enum
{
	ViewpointIntrinsic,		//	normal operation
	ViewpointExtrinsic,		//	external view of fundamental domain
	ViewpointEntering		//	transition from extrinsic to intrinsic
} Viewpoint;
#endif


//	All platform-independent data about the space and
//	how it's displayed live in the ModelData structure.
typedef struct
{
	//	Have the various OpenGL elements been pre-initialized?
	bool			itsPreparedExtensions,
					itsPreparedShaders,
					itsPreparedTextures,
					itsPreparedVBOs,
					itsPreparedAperture;	//	Set to false to recompute Dirichlet faces VBO.
	
	//	Which OpenGL extensions does the current context support?
	bool			itsMultisamplingAvailableFlag,
					itsAnisotropicAvailableFlag,
					itsShadersAvailableFlag;
	
	//	Which OpenGL extensions would the user like enabled?
	bool			itsMultisamplingEnabledFlag,
					itsAnisotropicEnabledFlag,
					itsShadersEnabledFlag;
	
	//	OpenGL shaders, textures and vertex/index buffers.
	unsigned int	itsShaderPrograms[NumShaders],
					itsTextureNames[NumTextures],
					itsVertexBufferNames[NumVertexBuffers],
					itsIndexBufferNames [NumVertexBuffers];
	
	//	Draw in anaglyphic 3D?
	bool			itsStereoFlag;

	//	Synchronize to the monitor's refresh rate?
	bool			itsSyncFlag;

	//	Environmental measurements permit accurate simulation.
	//	These quantities correspond to the variables d, w, h, e and c
	//	in the file FlatSpaceProjectionMatrices.pdf.
	double			itsWindowDistance,
					itsWindowHalfWidth,
					itsWindowHalfHeight,
					itsEyeOffset,		//	for stereo 3D only
					itsIntrinsicUnit;	//	for stereo 3D only
	
	//	When some part of the program (for example the mouse-handling code)
	//	wants to request a redraw, it sets itsRedrawRequestFlag.
	//	The idle-time routine will handle the request and clear the flag.
	//
	//	Note:  It's essential that all redraws funnel through the idle-time routine.
	//	For example, if mouse movements generated redraw events directly, this could
	//	saturate the event loop, thus blocking normal idle-time motion.
	bool			itsRedrawRequestFlag;
	
	//	Keep track of ModelView and Projection matrices.
	double			itsModelViewMatrix[4][4],
					itsProjectionMatrix[4][4];

	//	Most of the code doesn't need to know the curvature
	//	of space.  However, some parts do, for example
	//	the part that draws the back hemisphere of S³.
	SpaceType		itsSpaceType;
	
	//	For flat and hyperbolic spaces, this flag is ignored.
	bool			itsDrawBackHemisphere;
	
	//	How far out should we tile?
	//	The tiling radius is the maximum distance that
	//	a group element may translate the basepoint (0,0,0,1).
	//	The tiling radius should be greater than π
	//	in the spherical case if you want to tile all of S³.
	//	In the hyperbolic case, try a radius in the range 3.0 - 6.0,
	//	depending on your graphics card.
	//	The drawing radius is similar, but is slightly smaller.
	//	By omitting a few of the outermost tiles, the number
	//	of visible tiles -- and therefore the frame rate -- remains
	//	more constant as we pass through a face of the Dirichlet domain.
	//	This avoids a sudden "lurching" effect that would otherwise
	//	occur when the frame rate suddenly drops.
	double			itsTilingRadius,
					itsDrawingRadius;
	
	//	Keep track of the user's placement in the world.  The transformation moves
	//	the eye from its default position (0,0,0,1) with right vector (1,0,0,0),
	//	up vector (0,1,0,0) and forward vector (0,0,1,0) to the user's current placement.
	Matrix			itsUserPlacement;
	
	//	How fast is the user moving?
	//	The only sustained momentum is straight forward.
	double			itsUserSpeed;

#ifdef CENTERPIECE_DISPLACEMENT
	//	Keep track of the centerpiece's placement in the world.  The transformation moves
	//	the centerpiece from its default position (0,0,0,1) with right vector (1,0,0,0),
	//	up vector (0,1,0,0) and forward vector (0,0,1,0) to its current placement.
	Matrix			itsCenterpiecePlacement;
#endif
	
	//	Keep a Dirichlet domain for the discrete group.
	//	Assume no group element fixes the origin.
	DirichletDomain	*itsDirichletDomain;
	
	//	Keep a list of all translates of the Dirichlet domain
	//	that sit sufficiently close to the origin.
	//	For a spherical manifold the list will typically include
	//	the whole finite group.  For all manifolds the list
	//	is sorted near-to-far. 
	Honeycomb		*itsHoneycomb;
	
	//	The aperture in each face of the Dirichlet domain may be
	//	fully open (0.0), fully closed (1.0), or anywhere in between.
	//	Use the left and right arrow keys to control it.
	//	The arrow keys change itsDesiredAperture immediately,
	//	but itsCurrentAperture catches up only gradually,
	//	providing a smooth animation.
	double			itsDesiredAperture,
					itsCurrentAperture;
	
	//	What scenery should we display within each translate
	//	of the fundamental cell?
	SceneryType		itsSceneryType;
	
	//	Let the centerpiece (Earth, galaxy or gyroscope) rotate.
	double			itsRotationAngle;	//	in radians
	
	//	Draw the observer (as a small colored dart,
	//	representing the user's spaceship) ?
	bool			itsShowObserver;
	
	//	Color code the faces?
	bool			itsShowColorCoding;
	
	//	Draw Clifford parallels in spherical spaces?
	bool			itsShowCliffordParallels;
	
	//	Draw vertex figures?
	bool			itsShowVertexFigures;
	
	//	Enable fog?
	bool			itsFogFlag;			//	Does the user want fog?
	double			itsFogSaturation;	//	Current saturation level
										//		0.0 = fully transparent
										//		1.0 = fully opaque

#ifdef START_OUTSIDE
	//	View the fundamental domain from within, from without,
	//	or somewhere in between?
	Viewpoint		itsViewpoint;

	//	If we’re viewing the fundamental domain from somewhere
	//	in between, how far along are we?
	//		0.0 = ViewpointIntrinsic
	//		1.0 = ViewpointExtrinsic
	double			itsViewpointTransition;
	
	//	Keep the fundamental domain spinning.
	double			itsExtrinsicRotation;
#endif

} ModelData;


//	Ordinary rendering uses a single viewpoint,
//	while anaglyphic 3D uses separate left- and right-eye views.
typedef enum
{
	EyeOnly,
	EyeLeft,
	EyeRight
} EyeType;

//	Color definitions

typedef struct
{
	double	h,
			s,
			l,
			a;
} HSLAColor;

typedef struct
{
	double	r,
			g,
			b,
			a;
} RGBAColor;


//	A dictionary will translate all user-visible phrases
//	to the currently active language.  
//	The dictionary's structure is private to CurvedSpacesLocalization.c.
//	For all other purposes the following opaque typedef suffices.
typedef struct LocalizedPhrase	LocalizedPhrase;


//	Platform-independent global variables

//	Two-letter language code, such as "en" or "ja".
extern char				gLanguageCode[3];	//	zero-terminated ASCII string

//	Keep a dictionary for translating all user-interface text
//	into the current language.
extern LocalizedPhrase	*gDictionary;

//	Show frame rate?
extern bool				gShowFrameRate;

//	Test for memory leaks.
extern signed int		gMemCount;


//	Platform-dependent global functions,
//	in CurvedSpaces-Win32-Utilities.c or CurvedSpaces-MacOS-Utilities.c
extern bool			GetBasePath(char *aPath, unsigned int aPathSize);
extern ErrorText	GetFileContents(const char *aPathName, unsigned int *aNumRawBytes, Byte **someRawBytes);
extern void			FreeFileContents(unsigned int *aNumRawBytes, Byte **someRawBytes);
extern bool			GetUserPref(char *aKey);
extern void			SetUserPref(char *aKey, bool aValue);
extern void			ErrorMessage(ErrorText aMessage, ErrorText aTitle);

//	Platform-independent global functions

//	in CurvedSpacesInit.c
extern bool			ShadersAreAvailable(void);
extern void			SetUpModelData(ModelData *md);
extern void			ShutDownModelData(ModelData *md);
extern ErrorText	SetUpGraphicsAsNeeded(ModelData *md);
extern void			ShutDownGraphicsAsNeeded(ModelData *md);

//	in CurvedSpacesUtilities.c
extern ErrorText	GetAbsolutePath(const char *aDirectory, const char *aFileName, char *aPath, unsigned int aPathSize);

//	in CurvedSpacesOptions.c
extern void			SetSceneryType(ModelData *md, SceneryType aSceneryTypeChoice);
extern void			SetShowObserver(ModelData *md, bool aShowObserverChoice);
extern void			SetShowColorCoding(ModelData *md, bool aShowColorCodingChoice);
extern void			SetShowCliffordParallels(ModelData *md, bool aShowCliffordParallelsChoice);
extern void			SetShowVertexFigures(ModelData *md, bool aShowVertexFiguresChoice);
extern void			SetFogFlag(ModelData *md, bool aFogFlag);
extern void			SetMultisamplingEnabledFlag(ModelData *md, bool aMultisamplingEnabledFlag);
extern void			SetStereo3DFlag(ModelData *md, bool aStereo3DFlag);
extern void			SetAnisotropicEnabledFlag(ModelData *md, bool anAnisotropicEnabledFlag);
extern void			SetShadersEnabledFlag(ModelData *md, bool aShadersEnabledFlag);

//	in CurvedSpacesSimulation.c
extern bool			SimulationWantsUpdates(ModelData *md);
extern void			SimulationUpdate(ModelData *md, double aFramePeriod);
extern void			ChangeAperture(ModelData *md, bool aDilationFlag);
extern void			FastGramSchmidt(Matrix *aMatrix, SpaceType aSpaceType);

//	in CurvedSpacesMouse.c
extern void			MouseMove(ModelData *md, double dx, double dy, bool aShiftKeyIsDown, bool aCtrlKeyIsDown, bool anAltKeyIsDown);

//	in CurvedSpacesFileIO.c
extern ErrorText	LoadGenerators(ModelData *md, Byte *anInputText);

//	in CurvedSpacesTiling.c
extern ErrorText	ConstructHolonomyGroup(MatrixList *aGeneratorList, double aTilingRadius, MatrixList **aHolonomyGroup);
extern ErrorText	NeedsBackHemisphere(MatrixList *aHolonomyGroup, SpaceType aSpaceType, bool *aDrawBackHemisphereFlag);

//	in CurvedSpacesGraphics.c
extern char			*GetRendererName(void);
extern void			InitTexture(ModelData *md, unsigned int aTextureIndex, bool aClampFlag);
extern void			SetTextureImage(ModelData *md, unsigned int aTextureIndex, unsigned int aWidth, unsigned int aHeight, Byte (*aPixelArray)[4]);
extern bool			IsPowerOfTwo(unsigned int n);
extern void			ResizeGLWindow(ModelData *md, unsigned int aNewScreenWidth, unsigned int aNewScreenHeight);
extern ErrorText	ReadRGBAPixels(unsigned int aWidth, unsigned int aHeight, Byte *aBuffer, bool aReadFrontBufferFlag);
extern ErrorText	Render(ModelData *md);
extern void			SendTransformationsToShader(ModelData *md);

//	in CurvedSpacesDirichlet.c
extern ErrorText	ConstructDirichletDomain(MatrixList *aHolonomyGroup, DirichletDomain **aDirichletDomain);
extern void			FreeDirichletDomain(DirichletDomain **aDirichletDomain);
extern void			StayInDirichletDomain(DirichletDomain *aDirichletDomain, Matrix *aPlacement);
extern ErrorText	ConstructHoneycomb(MatrixList *aHolonomyGroup, DirichletDomain *aDirichletDomain, Honeycomb **aHoneycomb);
extern void			FreeHoneycomb(Honeycomb **aHoneycomb);
extern ErrorText	MakeDirichletVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, DirichletDomain *aDirichletDomain, double anAperture, bool aColorCodingFlag, bool aStereoFlag);
extern void			BindDirichletVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag);
extern void			DrawDirichletVBOs(ModelData *md, DirichletDomain *aDirichletDomain, Honeycomb *aHoneycomb, Matrix *aWorldPlacement, bool aColorCodingFlag);
extern ErrorText	MakeVertexFiguresVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, DirichletDomain *aDirichletDomain);
extern void			BindVertexFiguresVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag);
extern void			DrawVertexFiguresVBOs(ModelData *md, DirichletDomain *aDirichletDomain, Honeycomb *aHoneycomb, Matrix *aWorldPlacement);
extern void			SortVisibleCells(Honeycomb *aHoneycomb, Matrix *aViewProjectionMatrix, Matrix *aViewMatrix, double aDrawingRadius);

//	in CurvedSpacesEarth.c
extern ErrorText	MakeEarthVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName);
extern void			BindEarthVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag);
extern void			DrawEarthVBOs(ModelData *md, Honeycomb *aHoneycomb, Matrix *aWorldPlacement, Matrix *anEarthPlacement);

//	in CurvedSpacesGalaxy.c
extern void			MakeGalaxyVBO(unsigned int aVertexBufferName, unsigned int anIndexBufferName);
extern void			BindGalaxyVBO(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag);
extern void			DrawGalaxyVBO(ModelData *md, Honeycomb *aHoneycomb, Matrix *aWorldPlacement, Matrix *aGalaxyPlacement);

//	in CurvedSpacesGyroscope.c
extern void			MakeGyroscopeVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aStereoFlag);
extern void			BindGyroscopeVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag);
extern void			DrawGyroscopeVBOs(ModelData *md, Honeycomb *aHoneycomb, Matrix *aWorldPlacement, Matrix *aGyroscopePlacement);

//	in CurvedSpacesObserver.c
extern void			MakeObserverVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aStereoFlag);
extern void			BindObserverVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag);
extern void			DrawObserverVBOs(ModelData *md, Honeycomb *aHoneycomb, Matrix *aWorldPlacement, Matrix *anObserverPlacement);

//	in CurvedSpacesClifford.c
extern void			MakeCliffordVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName);
extern void			BindCliffordVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag);
extern void			DrawCliffordVBOs(ModelData *md, Matrix *aWorldPlacement);

//	in CurvedSpacesMatrices.c
extern void			MatrixIdentity(Matrix *aMatrix);
extern bool			MatrixIsIdentity(Matrix *aMatrix);
extern void			MatrixTranslation(Matrix *aMatrix, SpaceType aSpaceType, double dx, double dy, double dz);
extern void			MatrixRotation(Matrix *aMatrix, double da, double db, double dc);
extern void			MatrixGeometricInverse(Matrix *aMatrix, Matrix *anInverse);
extern double		MatrixDeterminant(Matrix *aMatrix);
extern void			VectorTernaryCrossProduct(Vector *aFactorA, Vector *aFactorB, Vector *aFactorC, Vector *aProduct);
extern bool			MatrixEquality(Matrix *aMatrixA, Matrix *aMatrixB, double anEpsilon);
extern void			MatrixProduct(Matrix *aMatrixA, Matrix *aMatrixB, Matrix *aProduct);
extern void			VectorNegate(Vector *aVector, Vector *aNegation);
extern void			VectorSum(Vector *aVectorA, Vector *aVectorB, Vector *aSum);
extern void			VectorDifference(Vector *aVectorA, Vector *aVectorB, Vector *aDifference);
extern void			VectorInterpolate(Vector *aVectorA, Vector *aVectorB, double t, Vector *aResult);
extern double		VectorDotProduct(Vector *aVectorA, Vector *aVectorB);
extern ErrorText	VectorNormalize(Vector *aRawVector, SpaceType aSpaceType, Vector *aNormalizedVector);
extern double		VectorGeometricDistance(Vector *aVector);
extern double		VectorGeometricDistance2(Vector *aVectorA, Vector *aVectorB);
extern void			VectorTimesMatrix(Vector *aVector, Matrix *aMatrix, Vector *aProduct);
extern void			ScalarTimesVector(double aScalar, Vector *aVector, Vector *aProduct);
extern MatrixList	*AllocateMatrixList(unsigned int aNumMatrices);
extern void			FreeMatrixList(MatrixList **aMatrixList);

//	in CurvedSpacesSafeMath.c
extern double		SafeAcos(double x);
extern double		SafeAcosh(double x);

//	in CurvedSpacesColors.c
extern void			HSLAtoRGBA(HSLAColor *anHSLAColor, RGBAColor *anRGBAColor);

//	in CurvedSpacesLocalization.c
extern void			RefreshLocalization(const char aTwoLetterLanguageCode[3]);
extern const char	*GetLocalizedText(const char *aKey);
