//	CurvedSpacesClifford.c
//
//	Makes, binds and draws Vertex Buffer Objects for Clifford parallels.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#include "CurvedSpaces-Common.h"
#include <stddef.h>	//	for offsetof()
#include <math.h>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#endif
#ifdef __WIN32__
#include "GeometryGames-Win32-OpenGLEntryPoints.h"
#endif


//	What is the (approximate) radius of each Clifford parallel?
#define	R					0.01

//	How finely should we subdivide each meridian?
#define M					8

//	How finely should we subdivide each longitude?  (Must be even)
#define N					8

//	How many times should the longitudinal texture coordinate cycle
//	within each longitudinal segment?
#define TEXTURE_MULTIPLE	25


//	The Clifford parallel Vertex Buffer Object (VBO) will contain
//	the following data for each of its vertices.
typedef struct
{
	float	pos[4],	//	position (x,y,z,w)
			tex[2];	//	texture coordinates (u,v)
} CliffordVBOData;

//	The Clifford parallel Index Buffer Object (IBO) will contain
//	the following data for each of its faces.
typedef struct
{
	unsigned short	vtx[3];	//	three vertices
} CliffordIBOData;


//	How many Clifford parallels should each (coaxial, toroidal) layer contain?
//	Radially the layers are spaced π/12 units apart.
//	Both endpoints (at distance 0 and at distance π) are included,
//	so there are 13 levels total.
static const unsigned int	gNumParallels[13]	= {1, 4, 8, 11, 14, 16, 16, 16, 14, 11, 8, 4, 1};
static Matrix				gTranslations[1 + 4 + 8 + 11 + 14 + 16 + 16 + 16 + 14 + 11 + 8 + 4 + 1];


static void		MakeTransformation(Matrix *aTransformation, double aTheta, double aPhi);


void MakeCliffordVBOs(
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName)
{
	CliffordVBOData	theVertices[N][M];
	CliffordIBOData	theFaces[N][M][2];
	unsigned int	i,
					j;
	Matrix			*theTranslation;
	unsigned int	n,
					m;

	//	Create the vertices for a single Clifford parallel
	//	running along the axis {x² + y² = 0, w² + z² = 1}.

	//	Position the vertices like this
	//
	//		00--01--02--03--00					|
	//		 \  /\  /\  /\  /\					|
	//		  30--31--32--33--30				|
	//		 /  \/  \/  \/  \/					|
	//		20--21--22--23--20					|
	//		 \  /\  /\  /\  /\					|
	//		  10--11--12--13--10				|
	//		 /  \/  \/  \/  \/					|
	//		00--01--02--03--00					|
	//
	//	with a half-notch rotation from each meridian
	//	to the next.
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < M; j++)
		{
			theVertices[i][j].pos[0] = R*cos(2*PI*(((i & 0x00000001) ? 0.5 : 0.0) + j)/M);
			theVertices[i][j].pos[1] = R*sin(2*PI*(((i & 0x00000001) ? 0.5 : 0.0) + j)/M);
			theVertices[i][j].pos[2] =   cos(2*PI*i/N);
			theVertices[i][j].pos[3] =   sin(2*PI*i/N);

			theVertices[i][j].tex[0] = 0.0;	//	irrelevant
			theVertices[i][j].tex[1] = ((i & 0x00000001) ? TEXTURE_MULTIPLE : 0.0);
		}
	}
	
	//	List the faces.
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < M; j++)
		{
			if (i & 0x00000001)
			{
				//	odd-numbered row

				theFaces[i][j][1].vtx[0] = M*((i+1)%N) +    j   ;
				theFaces[i][j][1].vtx[1] = M*(   i   ) +    j   ;
				theFaces[i][j][1].vtx[2] = M*((i+1)%N) + (j+1)%M;
				
				theFaces[i][j][0].vtx[0] = M*((i+1)%N) + (j+1)%M;
				theFaces[i][j][0].vtx[1] = M*(   i   ) +    j   ;
				theFaces[i][j][0].vtx[2] = M*(   i   ) + (j+1)%M;
			}
			else
			{
				//	even-numbered row

				theFaces[i][j][1].vtx[0] = M*(   i   ) +    j   ;
				theFaces[i][j][1].vtx[1] = M*(   i   ) + (j+1)%M;
				theFaces[i][j][1].vtx[2] = M*((i+1)%N) +    j   ;
				
				theFaces[i][j][0].vtx[0] = M*((i+1)%N) +    j   ;
				theFaces[i][j][0].vtx[1] = M*(   i   ) + (j+1)%M;
				theFaces[i][j][0].vtx[2] = M*((i+1)%N) + (j+1)%M;
			}
		}
	}

	//	Send the data to the GPU.

	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);
	glBufferData(	GL_ARRAY_BUFFER,
					sizeof(theVertices),
					theVertices,
					GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, anIndexBufferName);
	glBufferData(	GL_ELEMENT_ARRAY_BUFFER,
					sizeof(theFaces),
					theFaces,
					GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	//	Initialize the translation matrices.
	n = BUFFER_LENGTH(gNumParallels) - 1;
	theTranslation = gTranslations;
	for (i = 0; i <= n; i++)
	{
		m = gNumParallels[i];
		for (j = 0; j < m; j++)
			MakeTransformation(theTranslation++, i*PI/n, j*2*PI/m);
	}
}

static void MakeTransformation(
	Matrix	*aTransformation,	//	output
	double	aTheta,				//	input:  angle down from pole, in range [0,π]
	double	aPhi)				//	input:  rotation angle about pole, in range [0,2π]
{
	//	Construct an isometry of S³ for which the corresponding
	//	isometry of S² takes the north pole to the indicated point.
	//	In S³ such a map takes the "central" Clifford parallel
	//	to an desired Clifford parallel.  Two matrices are possible,
	//	different by minus the identity, so choose an arbitrary one.

	Matrix	theFirstFactor =
			{
				{
					{ cos(aTheta/2),       0.0,            0.0,      -sin(aTheta/2)},
					{      0.0,       cos(aTheta/2),  sin(aTheta/2),       0.0     },
					{      0.0,      -sin(aTheta/2),  cos(aTheta/2),       0.0     },
					{ sin(aTheta/2),       0.0,            0.0,       cos(aTheta/2)}
				},
				ImagePositive
			},
			theSecondFactor =
			{
				{
					{  cos(aPhi/2),  sin(aPhi/2),      0.0,          0.0    },
					{ -sin(aPhi/2),  cos(aPhi/2),      0.0,          0.0    },
					{      0.0,          0.0,      cos(aPhi/2), -sin(aPhi/2)},
					{      0.0,          0.0,      sin(aPhi/2),  cos(aPhi/2)}
				},
				ImagePositive
			};

	MatrixProduct(&theFirstFactor, &theSecondFactor, aTransformation);
}

void BindCliffordVBOs(
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName,
	bool			aShadersFlag)
{
	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);

	if (aShadersFlag)
	{
		glEnableVertexAttribArray(ATTRIBUTE_POSITION);
		glVertexAttribPointer(ATTRIBUTE_POSITION,  4, GL_FLOAT, GL_FALSE, sizeof(CliffordVBOData), (void *)offsetof(CliffordVBOData, pos));

		glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
		glVertexAttribPointer(ATTRIBUTE_TEX_COORD, 2, GL_FLOAT, GL_FALSE, sizeof(CliffordVBOData), (void *)offsetof(CliffordVBOData, tex));

		glDisableVertexAttribArray(ATTRIBUTE_COLOR);
	}
	else
	{
		//	Internet folklore claims that glVertexPointer() should be called last,
		//	after glTexCoordPointer(), glColorPointer(), etc., but my own testing
		//	revealed no difference.

		glVertexPointer(4, GL_FLOAT, sizeof(CliffordVBOData), (void *)offsetof(CliffordVBOData, pos));
		glEnableClientState(GL_VERTEX_ARRAY);

		glTexCoordPointer(2, GL_FLOAT, sizeof(CliffordVBOData), (void *)offsetof(CliffordVBOData, tex));
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glDisableClientState(GL_COLOR_ARRAY);
	}

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, anIndexBufferName);
}

void DrawCliffordVBOs(
	ModelData	*md,
	Matrix		*aWorldPlacement)	//	the world's placement in eye space
{
	Matrix			*theTranslation;
	const float		*theColor = (float [4]){0.0, 0.0, 0.0, 1.0};
	unsigned int	n,
					i,
					j;

	static const float	theColorDarkBlue[4]		= {0.0, 0.0, 1.0, 1.0},
						theColorGreyBlue[4]		= {0.5, 0.5, 1.0, 1.0},
						theColorWhite[4]		= {1.0, 1.0, 1.0, 1.0},
						theColorGreyGreen[4]	= {0.5, 1.0, 0.5, 1.0},
						theColorDarkGreen[4]	= {0.0, 0.5, 0.0, 1.0};

	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glFrontFace(aWorldPlacement->itsParity == ImagePositive ? GL_CCW : GL_CW);
	glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureClifford]);

	//	To apply the various translations, we can use a pair of nested loops
	//	just like the ones that MakeCliffordVBOs() used to create the translations
	//	in the first place.
	n = BUFFER_LENGTH(gNumParallels) - 1;
	theTranslation = gTranslations;
	for (i = 0; i <= n; i++)
	{
		//	Select a color.
		if (i == 0)			theColor = theColorDarkBlue;	//	nearest Clifford parallel
		if (i == 1)			theColor = theColorGreyBlue;	//	near Clifford parallels
		if (i == n/2)		theColor = theColorWhite;		//	half-way Clifford parallels
		if (i == n/2 + 1)	theColor = theColorGreyGreen;	//	far Clifford parallels
		if (i == n)			theColor = theColorDarkGreen;	//	furthest Clifford parallel
		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
			glVertexAttrib4fv(ATTRIBUTE_COLOR, theColor);
		else
			glColor4fv(theColor);

		//	Draw all the parallels found at the present distance.
		for (j = gNumParallels[i]; j-- > 0; )
		{
			//	Compose theTranslation with aWorldPlacement, write the result
			//	to itsModelViewMatrix, and send it to the shader.
			Matrix44Product(theTranslation->m, aWorldPlacement->m, md->itsModelViewMatrix);
			SendTransformationsToShader(md);

			//	Draw one Clifford parallel.
			glDrawRangeElements(	GL_TRIANGLES,
									0,
									N*M - 1,
									3*2*N*M,
									GL_UNSIGNED_SHORT,
									0);
			
			//	Move on to the next translation.
			theTranslation++;
		}
	}
}
