//	CurvedSpacesGalaxy.c
//
//	Makes, binds and draws Vertex Buffer Objects for the rotating galaxy.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#include "CurvedSpaces-Common.h"
#include <stddef.h>	//	for offsetof()
#include <math.h>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#endif
#ifdef __WIN32__
#include "GeometryGames-Win32-OpenGLEntryPoints.h"
#endif


//	The galaxy's corners sit at (±GALAXY_SIZE, ±GALAXY_SIZE, 0).
#define GALAXY_SIZE		0.25


//	The galaxy Vertex Buffer Object (VBO) will contain
//	the following data for each of its vertices.
typedef struct
{
	float	pos[4],	//	position (x,y,z,w)
			tex[2];	//	texture coordinates (u,v)
} GalaxyVBOData;


void MakeGalaxyVBO(
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName)
{
	//	The galaxy is so simply (a single quad!) that it's hardly worth
	//	creating a Vertex Buffer Object for it.  I'm making one anyhow,
	//	though, for consistency with the other centerpieces (the Earth
	//	and the gyroscope) which do merit Vertex Buffer Objects.

	static const GalaxyVBOData	theVertices[4] =
								{
									{{-GALAXY_SIZE, -GALAXY_SIZE, 0.0, 1.0}, {0.0, 0.0}},
									{{+GALAXY_SIZE, -GALAXY_SIZE, 0.0, 1.0}, {1.0, 0.0}},
									{{+GALAXY_SIZE, +GALAXY_SIZE, 0.0, 1.0}, {1.0, 1.0}},
									{{-GALAXY_SIZE, +GALAXY_SIZE, 0.0, 1.0}, {0.0, 1.0}}
								};

	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);
	glBufferData(	GL_ARRAY_BUFFER,
					4 * sizeof(GalaxyVBOData),
					theVertices,
					GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	UNUSED_PARAMETER(anIndexBufferName);
}

void BindGalaxyVBO(
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName,
	bool			aShadersFlag)
{
	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);

	if (aShadersFlag)
	{
		glEnableVertexAttribArray(ATTRIBUTE_POSITION);
		glVertexAttribPointer(ATTRIBUTE_POSITION,  4, GL_FLOAT, GL_FALSE, sizeof(GalaxyVBOData), (void *)offsetof(GalaxyVBOData, pos));

		glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
		glVertexAttribPointer(ATTRIBUTE_TEX_COORD, 2, GL_FLOAT, GL_FALSE, sizeof(GalaxyVBOData), (void *)offsetof(GalaxyVBOData, tex));

		glDisableVertexAttribArray(ATTRIBUTE_COLOR);
	}
	else
	{
		//	Internet folklore claims that glVertexPointer() should be called last,
		//	after glTexCoordPointer(), glColorPointer(), etc., but my own testing
		//	revealed no difference.

		glVertexPointer(	4, GL_FLOAT, sizeof(GalaxyVBOData), (void *)offsetof(GalaxyVBOData, pos));
		glEnableClientState(GL_VERTEX_ARRAY);

		glTexCoordPointer(	2, GL_FLOAT, sizeof(GalaxyVBOData), (void *)offsetof(GalaxyVBOData, tex));
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glDisableClientState(GL_COLOR_ARRAY);
	}

	UNUSED_PARAMETER(anIndexBufferName);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

void DrawGalaxyVBO(
	ModelData	*md,
	Honeycomb	*aHoneycomb,
	Matrix		*aWorldPlacement,	//	the world's placement in eye space
	Matrix		*aGalaxyPlacement)	//	the galaxy's placement in the Dirichlet domain
{
	unsigned int	i;
	Matrix			*theDirichletPlacement;	//	the (translated) Dirichlet domain's placement in world space

	if (aHoneycomb == NULL)
		return;

	glEnable(GL_BLEND);
	glDisable(GL_CULL_FACE);
	glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureGalaxy]);
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		glVertexAttrib4fv(ATTRIBUTE_COLOR, (float [4]){1.0, 1.0, 1.0, 1.0});
	else
		glColor4f(1.0, 1.0, 1.0, 1.0);
	
	//	Draw the spinning galaxies in far-to-near order,
	//	to get the transparency right.
	for (i = aHoneycomb->itsNumVisibleCells; i-- > 0; )
	{
		//	Each element of the tiling group defines
		//	a placement of the Dirichlet domain in world space.
		theDirichletPlacement = &aHoneycomb->itsVisibleCells[i]->itsMatrix;

		//	Compose aGalaxyPlacement, theDirichletPlacement and aWorldPlacement,
		//	write the result to itsModelViewMatrix, and send it to the shader.
		Matrix44Product(aGalaxyPlacement->m,    theDirichletPlacement->m, md->itsModelViewMatrix);
		Matrix44Product(md->itsModelViewMatrix, aWorldPlacement->m,       md->itsModelViewMatrix);
		SendTransformationsToShader(md);

		//	Draw.
		glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
	}

	glDisable(GL_BLEND);
}
