//	CurvedSpacesGraphics.c
//
//	Drawing code for Curved Spaces.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#include "CurvedSpaces-Common.h"
#include <math.h>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#endif
#ifdef __WIN32__
#include "GeometryGames-Win32-OpenGLEntryPoints.h"
#endif


//	The near clipping distance is 1/INVERSE_NEAR_CLIP.
//
//	See comments in CurvedSpacesObserver.c for an explanation
//	of why INVERSE_NEAR_CLIP should be at least 1/(0.004/2) ≈ 500.
//	However, it shouldn't be unnecessarily large, to avoid
//	needless loss of precision in the depth buffer.
#define INVERSE_NEAR_CLIP	512.0

//	How fast should the galaxy, Earth and gyroscope spin?
//	Express their speeds as integer multiples of the default rotation speed.
//	The reason they're integer multiples is that itsRotationAngle
//	occasionally jumps by 2π.
#define GALAXY_SPEED			1
#define EARTH_SPEED				2
#define GYROSCOPE_SPEED			6

#ifdef START_OUTSIDE
//	When viewing the fundamental polyhedron from outside,
//	how far away should it sit?
#define EXTRINSIC_VIEWING_DISTANCE	0.75
#endif


typedef enum
{
	BoxFull,	//	render into the full clipping box -w ≤ z ≤ w
	BoxFront,	//	render into the front half        -w ≤ z ≤ 0
	BoxBack		//	render into the back  half         0 ≤ z ≤ w
} ClippingBoxPortion;


static void			HalveTextureWidth(unsigned int *aWidth, unsigned int aHeight, Byte (*aPixelArray)[4]);
static void			HalveTextureHeight(unsigned int aWidth, unsigned int *aHeight, Byte (*aPixelArray)[4]);
static void			ProjectAndDraw(EyeType anEyeType, ModelData *md);
static void			SetProjectionMatrix(EyeType anEyeType, ClippingBoxPortion aClippingBoxPortion, bool aSceneryInversionFlag, ModelData *md);
static void			DrawTheScene(ModelData *md);
static void			DrawTheSceneIntrinsically(ModelData *md);
#ifdef START_OUTSIDE
static void			DrawTheSceneExtrinsically(ModelData *md);
#endif
static ErrorText	GetErrorString(void);


char *GetRendererName(void)
{
	return (char *)glGetString(GL_RENDERER);
}


void InitTexture(
	ModelData		*md,
	unsigned int	aTextureIndex,
	bool			aClampFlag)
{
	float	theMaxAnisotropy;

	//	Bind the texture.
	glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[aTextureIndex]);

	//	Set minification and magnification filters.
	//	Mipmapping is almost always a good idea.
	//
	//	On mobile platforms (PowerVR) GL_LINEAR_MIPMAP_NEAREST is much faster
	//	than GL_LINEAR_MIPMAP_LINEAR (40-42 fps instead of 25-27 fps in torus Jigsaw).
#ifdef GL_VERSION_ES_CM_1_1
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
#else
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
#endif
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR               );

	//	Set anisotropic texture filtering.
	if (md->itsAnisotropicAvailableFlag)
	{
		if (md->itsAnisotropicEnabledFlag)
			glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &theMaxAnisotropy);	//	Allow the maximum.
		else
			theMaxAnisotropy = 1.0f;	//	Suppress anisotropic texture filtering.
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, theMaxAnisotropy);
	}

	//	How shall we handle the edges?
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, aClampFlag ? GL_CLAMP_TO_EDGE : GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, aClampFlag ? GL_CLAMP_TO_EDGE : GL_REPEAT);
}


void SetTextureImage(
	ModelData		*md,
	unsigned int	aTextureIndex,
	unsigned int	aWidth,
	unsigned int	aHeight,
	Byte			(*aPixelArray)[4])
{
	//	Creates a standard RGBA texture.
	//	To create more general textures, such as alpha-only textures,
	//	please see the Torus Games version of SetTextureImage().

	unsigned int	theMipmapLevel;

	if (IsPowerOfTwo(aWidth ) == false
	 || IsPowerOfTwo(aHeight) == false)
		return;

	glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[aTextureIndex]);

	theMipmapLevel = 0;
	while (true)
	{
		glTexImage2D(	GL_TEXTURE_2D,
						theMipmapLevel,
						GL_RGBA,
						aWidth,
						aHeight,
						0,	//	Modern graphics cards don't support borders -- see TextureBorders.html.
						GL_RGBA,
						GL_UNSIGNED_BYTE,
						aPixelArray);

		if (aWidth == 1 && aHeight == 1)
			break;

		if (aWidth  > 1)
			HalveTextureWidth (&aWidth, aHeight, aPixelArray);
		if (aHeight > 1)
			HalveTextureHeight(aWidth, &aHeight, aPixelArray);

		theMipmapLevel++;
	}
}


static void HalveTextureWidth(
	unsigned int	*aWidth,
	unsigned int	aHeight,
	Byte			(*aPixelArray)[4])
{
	unsigned int	theOldWidth,
					theNewWidth,
					r,
					c,
					i;

	//	Just to be safe...
	if (IsPowerOfTwo(*aWidth ) == false
	 || IsPowerOfTwo(aHeight) == false
	 || *aWidth  < 2)
		return;

	//	Note the old and new widths.
	theOldWidth = *aWidth;
	theNewWidth = theOldWidth/2;

	//	Let each texel with coordinates (r, 2c) be the average
	//	of the texels at (r, 2c) and (r, 2c + 1).
	for (r = 0; r < aHeight; r++)
		for (c = 0; c < theNewWidth; c++)
			for (i = 0; i < 4; i++)
				    aPixelArray[r*theOldWidth + (2*c    )][i]
				 = (aPixelArray[r*theOldWidth + (2*c    )][i]
				  + aPixelArray[r*theOldWidth + (2*c + 1)][i])
				  / 2;

	//	Compress the array, so only the averaged texels remain.
	for (r = 0; r < aHeight; r++)
		for (c = 0; c < theNewWidth; c++)
			for (i = 0; i < 4; i++)
				aPixelArray[r*theNewWidth + c][i] = aPixelArray[r*theOldWidth + 2*c][i];

	//	Set the new image width.
	*aWidth = theNewWidth;
}


static void HalveTextureHeight(
	unsigned int	aWidth,
	unsigned int	*aHeight,
	Byte			(*aPixelArray)[4])
{
	unsigned int	theOldHeight,
					theNewHeight,
					r,
					c,
					i;

	//	Just to be safe...
	if (IsPowerOfTwo(aWidth ) == false
	 || IsPowerOfTwo(*aHeight) == false
	 || *aHeight < 2)
		return;

	//	Note the old and new heights.
	theOldHeight = *aHeight;
	theNewHeight = theOldHeight/2;

	//	Let each texel with coordinates (2r, c) be the average
	//	of the texels at (2r, c) and (2r + 1, c).
	for (r = 0; r < theNewHeight; r++)
		for (c = 0; c < aWidth; c++)
			for (i = 0; i < 4; i++)
				    aPixelArray[(2*r    )*aWidth + c][i]
				 = (aPixelArray[(2*r    )*aWidth + c][i]
				  + aPixelArray[(2*r + 1)*aWidth + c][i])
				  / 2;

	//	Compress the array, so only the averaged texels remain.
	for (r = 0; r < theNewHeight; r++)
		for (c = 0; c < aWidth; c++)
			for (i = 0; i < 4; i++)
				aPixelArray[r*aWidth + c][i] = aPixelArray[(2*r)*aWidth + c][i];

	//	Set the new image height.
	*aHeight = theNewHeight;
}


bool IsPowerOfTwo(unsigned int n)
{
	if (n == 0)
		return false;

	while ((n & 0x00000001) == 0)
		n >>= 1;

	return (n == 0x00000001);
}


void ResizeGLWindow(
	ModelData		*md,
	unsigned int	aNewScreenWidth,	//	in pixels
	unsigned int	aNewScreenHeight)	//	in pixels
{
	double	theOldInternalSize,	//	in intrinsic units
			theNewScreenSize;	//	in pixels

	//	Let OpenGL know about the new window size.
	glViewport(0, 0, aNewScreenWidth, aNewScreenHeight);

	//	At this point one might want to adjust the environmental measurements
	//	to match the new window size.  Various approaches are possible.
	//	For now let's just do the bare minimum to keep the aspect ratio correct.
	//
	//	Ignore calls with aWidth or aHeight equal to zero, because
	//	they'll foul up wd->md.itsWindowHalfWidth and wd->md.itsWindowHalfHeight.
	//	(Win32 generates such calls when the window is minimized.)
	if (aNewScreenWidth != 0 && aNewScreenHeight != 0)
	{
		theOldInternalSize	= sqrt(md->itsWindowHalfWidth * md->itsWindowHalfHeight);
		theNewScreenSize	= sqrt(aNewScreenWidth * aNewScreenHeight);
		md->itsWindowHalfWidth	= aNewScreenWidth  * (theOldInternalSize / theNewScreenSize);
		md->itsWindowHalfHeight	= aNewScreenHeight * (theOldInternalSize / theNewScreenSize);
	}
}


ErrorText ReadRGBAPixels(
	unsigned int	aWidth,
	unsigned int	aHeight,
	Byte			*aBuffer,
	bool			aReadFrontBufferFlag)
{
	//	The Radeon Catalyst driver on Windows can't read
	//	the back buffer when multisampling is enabled --
	//	instead it fills the whole image with background color.
	//	So on Windows, read the front buffer instead.
	//	On Mac OS X, the driver reads the back buffer correctly,
	//	even with multisampling.
	//
	//	As of May 2009 this Catalyst bug still persists,
	//	so I can't get rid of this awkward little workaround.
	//	To get the code to compile cleanly under OpenGL ES,
	//	I bracketed the glReadBuffer() call with #ifdef GL_VERSION_1_1,
	//	so that the call remains visible under standard OpenGL 
	//	but disappears under OpenGL ES.  If at some point in the future
	//	the Catalyst bug disappears, then we can eliminate
	//	both the glReadBuffer() call and aReadFrontBufferFlag. 

#ifdef GL_VERSION_1_1
	glReadBuffer(aReadFrontBufferFlag ? GL_FRONT : GL_BACK);
#endif

	glReadPixels(	0, 0,
					(GLsizei) aWidth, (GLsizei) aHeight,
					GL_RGBA, GL_UNSIGNED_BYTE,
					(GLvoid *) aBuffer);
	
	return GetErrorString();
}


ErrorText Render(ModelData *md)
{
	unsigned int	theShaderProgram;
	float			theFogRangeParameter;

	//	Clear color buffer and the depth buffer.
	//	For Curved Spaces an opaque black background works well.
	//	For other programs, such as KaleidoTile and Kali,
	//	a transparent background would let the user superimpose
	//	copied-and-pasted or saved images on top of the background
	//	of his/her choice.
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//	Depth testing serves us well.
	glEnable(GL_DEPTH_TEST);
	
	//	Set up a single shader for the whole scene.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		//	Select a shader according to the SpaceType.
		switch (md->itsSpaceType)
		{
			case SpaceSpherical:	theShaderProgram = md->itsShaderPrograms[ShaderSph];	break;
			case SpaceFlat:			theShaderProgram = md->itsShaderPrograms[ShaderEuc];	break;
			case SpaceHyperbolic:	theShaderProgram = md->itsShaderPrograms[ShaderHyp];	break;
			
			//	At launch no space is present.  This is fine.
			//	The user will select a space momentarily.
			case SpaceNone:
			default:
				return NULL;
		}

		//	Enable the selected shader.
		glUseProgram(theShaderProgram);

		//	Set the amount of fog, ranging from 0.0 (fully transparent) to 1.0 (fully opaque).
		glUniform1f(glGetUniformLocation(theShaderProgram, "uniFogFactor"), md->itsFogSaturation);
		
		//	Adjust the fog range parameter, which conveys the distance
		//	at which the fog saturates.  Rather than passing that saturation distance
		//	directly, and thus forcing the vertex shader to compute an inverse
		//	along with a square or a cosh, pass the saturation distance
		//	in a pre-digested form.  Better to compute the inverse, etc.,
		//	once and for all here, rather than over and over in the shader,
		//	once for each vertex.
		switch (md->itsSpaceType)
		{
			case SpaceSpherical:
				//	Use 75% fog saturation at the antipode.
				theFogRangeParameter = 0.75;
				break;
			
			case SpaceFlat:
				//	Fog is proportional to d².
				//
				//	Setting full fog saturation at itsTilingRadius is a little bolder
				//	than setting it at itsDrawingRadius, but the results look good:
				//	the scenery becomes invisible just at the point you want it to,
				//	and no sooner (at least on my monitor).
				theFogRangeParameter = 1.0 / (md->itsTilingRadius*md->itsTilingRadius);
				break;
			
			case SpaceHyperbolic:
				//	Fog is proportional to log(w) = log(cosh(d)).
				//
				//	Letting the fog saturate at itsTilingRadius instead of 
				//	at itsDrawingRadius isn't quite as perfect as in the flat case --
				//	on my monitor you can see a tiny bit of "popping" if you
				//	look for it -- but all in all it seems to be a good compromise
				//	between image depth and anti-popping.
				theFogRangeParameter = 1.0 / log(cosh(md->itsTilingRadius));
				break;
			
			default:
				return NULL;
		}
		glUniform1f(glGetUniformLocation(theShaderProgram, "uniFogRangeParameter"), theFogRangeParameter);
	}
	else
	{
	}

	//	Texturing will be active at (almost) all times.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
	}
	else
	{
		glEnable(GL_TEXTURE_2D);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	}

	//	Disable fixed-function lighting.  Directional lights are problematic
	//	in a periodic scene, so we won't have any.
	//	As for ambient light, the brightness of a surface viewed
	//	in a pure vacuum (no fog) is independent of distance,
	//	no matter what the geometry (spherical, Euclidean or hyperbolic).
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
	}
	else
	{
		glDisable(GL_LIGHTING);
	}

	//	Blending determines how the final fragment
	//	blends in with the previous color buffer contents.
	//	For opaque surfaces we can disable blending.
	//	For partially transparent surfaces, such as the galaxy,
	//	we may enable blending but must take care to draw the scene
	//	in back-to-front order.
	glDisable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	//	According to the formal specification, one may call
	//
	//			glEnable(GL_MULTISAMPLE)
	//	or
	//			glDisable(GL_MULTISAMPLE)
	//
	//	but in practice the pixel format (with or without a multisample buffer)
	//	wholly determines whether multisampling takes place.
	//	This latter approach makes sense, given that 
	//	multisampling into a single-sample buffer would be impossible, and
	//	single-sampling into a multisample buffer would be pointless.
	//	Nevertheless, let's call glEnable/glDisable(GL_MULTISAMPLE),
	//	just in case some future OpenGL graphics driver wants the call.
	if (md->itsMultisamplingAvailableFlag && md->itsMultisamplingEnabledFlag)
		glEnable(GL_MULTISAMPLE);
	else
		glDisable(GL_MULTISAMPLE);

	//	Disable unneeded arrays.
	if (md->itsShadersAvailableFlag)
	{
		if (md->itsShadersEnabledFlag)
		{
			//	We'll use only the forward-compatible glVertexAttribPointer(),
			//	not the backward-compatible glVertexPointer(), glTexCoordPointer(), etc.
			glDisableClientState(GL_VERTEX_ARRAY		);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY	);
			glDisableClientState(GL_COLOR_ARRAY			);
		}
		else
		{
			//	We'll use only the backward-compatible glVertexPointer(), glTexCoordPointer(), etc.,
			//	not the forward-compatible glVertexAttribPointer().
			glDisableVertexAttribArray(ATTRIBUTE_POSITION			);
			glDisableVertexAttribArray(ATTRIBUTE_TEX_COORD			);
			glDisableVertexAttribArray(ATTRIBUTE_COLOR				);
			glDisableVertexAttribArray(ATTRIBUTE_MVP_MATRIX_ROW_0	);
			glDisableVertexAttribArray(ATTRIBUTE_MVP_MATRIX_ROW_1	);
			glDisableVertexAttribArray(ATTRIBUTE_MVP_MATRIX_ROW_2	);
			glDisableVertexAttribArray(ATTRIBUTE_MVP_MATRIX_ROW_3	);
			glDisableVertexAttribArray(ATTRIBUTE_MV_MATRIX_ROW_0	);
			glDisableVertexAttribArray(ATTRIBUTE_MV_MATRIX_ROW_1	);
			glDisableVertexAttribArray(ATTRIBUTE_MV_MATRIX_ROW_2	);
			glDisableVertexAttribArray(ATTRIBUTE_MV_MATRIX_ROW_3	);
		}
	}


	//	Draw the scene.
	if (md->itsStereoFlag == false)
	{
		//	Draw a full color image for a single eye.
		ProjectAndDraw(EyeOnly, md);
	}
	else
	{
		//	Restrict to the red channel.
		glColorMask(true, false, false, true);

		//	Draw the left eye image.
		ProjectAndDraw(EyeLeft, md);

		//	Clear the z-buffer.
		glClear(GL_DEPTH_BUFFER_BIT);

		//	Restrict to the "cyan channel".
		glColorMask(false, true, true, true);

		//	Draw the right eye image.
		ProjectAndDraw(EyeRight, md);

		//	Re-enable all color channels.
		glColorMask(true, true, true, true);
	}

	//	Return a string describing the first of any OpenGL errors
	//	that may have occurred.  If no errors occurred, return NULL.
	return GetErrorString();
}


static void ProjectAndDraw(
	EyeType		anEyeType,
	ModelData	*md)
{
	if (md->itsSpaceType != SpaceSpherical
	 || md->itsDrawBackHemisphere == false)
	{
		SetProjectionMatrix(anEyeType, BoxFull,  false, md);
		DrawTheScene(md);
	}
	else
	{
		SetProjectionMatrix(anEyeType, BoxFront, false, md);
		DrawTheScene(md);

		SetProjectionMatrix(anEyeType, BoxBack,  true,  md);
		DrawTheScene(md);
	}
}


static void SetProjectionMatrix(
	EyeType				anEyeType,
	ClippingBoxPortion	aClippingBoxPortion,
	bool				aSceneryInversionFlag,	//	Invert the scenery to draw the back half of S³?
	ModelData			*md)
{
	//	How to Build a Projection Matrix
	//
	//	The key is to think... projectively!  After applying our projection matrix,
	//	the graphics card will divide through by the last coordinate,
	//	so points (x,y,z,w) and c(x,y,z,w) are equivalent for all positive c.
	//	(They'd be equivalent for negative c as well, except for clipping considerations.)
	//	Thus each projective point corresponds to a ray from the origin.
	//	Rays from the origin correspond, in turn, to points on S³,
	//	so we may visualize world space as S³ if we wish.
	//
	//	The curvature of the space being modelled (spherical, flat or hyperbolic)
	//	is almost irrelevant.  The only difference is that spherical space occupies
	//	all of S³, while flat space occupies only the northern hemisphere
	//	(excluding the equator, which corresponds to the Euclidean sphere at infinity)
	//	and hyperbolic space occupies only the disk above 45° north latitude.
	//
	//	OpenGL clips to a "clipping wedge" bounded by the six hyperplanes
	//
	//			-w ≤ x ≤ w
	//			-w ≤ y ≤ w
	//			-w ≤ z ≤ w
	//
	//	This wedge lies entirely in the upper half space w ≥ 0, which is why
	//	points (x,y,z,w) and c(x,y,z,w) are *not* equivalent when c is negative.
	//	The clipping wedge intersects the hyperplane w == 1 in a cube
	//	-1 ≤ x ≤ +1, -1 ≤ y ≤ +1, -1 ≤ z ≤ +1, which is a convenient way
	//	to visualize it.  The whole purpose of a projection matrix is to move
	//	a desired "view volume" (the portion of world space that we want to see)
	//	into the "clipping wedge".
	//
	//	Let us construct a projection matrix in steps.
	//	Each step corresponds to a block of code below,
	//	but the order in which we explain the steps differs
	//	from the order in which we apply them in the code.
	//
	//	Step 0.  The simplest possible starting point is the identity matrix
	//
	//			1  0  0  0
	//			0  1  0  0
	//			0  0  1  0
	//			0  0  0  1
	//
	//	The clipping wedge itself defines the view volume, which
	//	the user sees from the perspective of an observer at (0,0,-1,0).
	//	In the flat case, this is effectively an observer
	//	at negative infinity on the z-axis.
	//	In all cases, the observer's field of view extends 45°
	//	to the left, to the right, above and below the centerline.
	//
	//	Step 1.  Initialize the projection matrix to the quarter turn
	//
	//			1  0  0  0
	//			0  1  0  0
	//			0  0  0  1
	//			0  0 -1  0
	//
	//		which rotates a view volume in the front hemisphere (z > 0)
	//		onto the clipping wedge.  The observer sits at (0,0,0,1)
	//		in world coordinates (before applying the quarter turn),
	//		in agreement with usual practices.
	//		In the flat case, the near clipping plane is at z = 1
	//		while the far clipping plane lies "beyond infinity".
	//		In the spherical case, the view volume extends from π/4 to 3π/4 radians.
	//
	//	Step 2.  Adjust the near and far clipping planes.
	//
	//		In all three geometries we want the near clipping plane to pass through
	//		the point (0, 0, NEAR_CLIP, 1), which after the rotation of Step 1
	//		becomes (0, 0, -1, NEAR_CLIP) or equivalently (0, 0, -1/NEAR_CLIP, 1).
	//
	//		In the spherical case the far clipping plane should lie at an equal
	//		distance from the south pole, at the point (0, 0, NEAR_CLIP, -1), which
	//		rotates to (0, 0, +1, NEAR_CLIP) or equivalently (0, 0, +1/NEAR_CLIP, 1).
	//
	//		In the flat case the far clipping plane passes through (0, 0, 1, 0),
	//		which rotates to (0, 0, 0, 1).
	//
	//		In the hyperbolic case the far clipping plane passes through (0, 0, 1, 1),
	//		which rotates to (0, 0, -1, 1).  However, the distance from
	//		(0, 0, -1/NEAR_CLIP, 1) to (0, 0, -1, 1) is almost as great as
	//		the distance from (0, 0, -1/NEAR_CLIP, 1) to (0, 0, 0, 1), so there's
	//		no harm in folding the hyperbolic case in with the flat case.
	//		(Folding it in could keep all quantities exact powers of 2,
	//		even though it's hard to image that the speed or accuracy
	//		of the subsequent matrix algebra would ever be an issue.)
	//
	//		In all three geometries, a shear transformation translates
	//		the hyperplane w == 1 through a distance (NearClip + FarClip)/2
	//		to center the desired view volume, and a compression by a factor
	//		of (FarClip - NearClip)/2 aligns it with the front and back
	//		of the standard clipping wedge.
	//
	//	Step 3.  Adjust the left, right, bottom and top clipping planes
	//			to accommodate a field of view other than ±45°.
	//
	//		Similar triangles dictate that the desired view volume's
	//		left and right faces pass through the points
	//		( ±WindowHalfWidth, 0, 0, WindowDistance) or equivalently
	//		( ±WindowHalfWidth/WindowDistance, 0, 0, 1), so rescale by
	//		a factor of WindowDistance/WindowHalfWidth to align
	//		the view volume's left and right faces with those of the standard
	//		clipping box, and similarly for the top and bottom faces.
	//
	//	Step 4.  Adjust the left, right, bottom and top clipping planes
	//			to allow off-axis viewing for stereoscopic 3D.
	//
	//		The user's left eye sees the display screen slightly to its right,
	//		so we need to use a view volume that's also slightly to the right.
	//		In other words, we want to translate the desired view volume
	//		leftwards by NoseToEyeDistance/WindowHalfWidth.
	//
	//	Step 5.  Translate the scenery to accommodate stereoscopic 3D.
	//
	//		The left eye sees the scenery slightly to the right of where
	//		a cyclops would see it.  So strictly speaking the last factor
	//		in the modelview matrix should be a rightwards translation
	//		through a distance NoseToEyeDistance/WindowHalfWidth.
	//		For convenience, though, let's make it the first factor
	//		in the projection matrix.
	//
	//	Step 6.  Handle the back hemisphere of S³, if required.
	//
	//		Most spherical manifolds yield tilings that are symmetrical
	//		under the antipodal map, in which case there's no need to draw
	//		the contents of the back hemisphere.  However, for tilings lacking
	//		antipodal symmetry (such as those arising from odd-order lens spaces)
	//		we do need to draw the back hemisphere as well as the front one.
	//		To do this, two modifications are required.
	//
	//		a.	The program draws the front hemisphere into the front half
	//			of the clipping box ( -w ≤ z ≤ 0 ) and draws the back hemisphere
	//			into the back half of the clipping box.
	//
	//		b.	For the back hemisphere, the program applies the antipodal map
	//			to all back-hemisphere scenery to bring it to an equivalent position
	//			in the front hemisphere.

	double	theFactor[4][4],
			n,	//	near clipping plane passes through (0, 0, n, 1)
			f,	//	far  clipping plane passes through (0, 0, f, 1)
			d;	//	nose-to-eye separation in intrinsic units

	//	Initialize itsProjectionMatrix to the identity.
	Matrix44Identity(md->itsProjectionMatrix);

	//	Check our inputs just to be safe.
	if (md->itsWindowDistance	<= 0.0
	 || md->itsWindowHalfWidth	<= 0.0
	 || md->itsWindowHalfHeight	<= 0.0
	 || md->itsIntrinsicUnit	<= 0.0)
		return;

	//	Step 6a.  When drawing both hemispheres of S³,
	//	compress the front hemisphere into the front half of the clipping box,
	//	and the back hemisphere into the back half of the clipping box.
	if (aClippingBoxPortion != BoxFull)
	{
		Matrix44Identity(theFactor);
		theFactor[2][2] = 0.5;
		theFactor[3][2] = (aClippingBoxPortion == BoxFront ? -0.5 : +0.5);
		Matrix44Product(theFactor, md->itsProjectionMatrix, md->itsProjectionMatrix);
	}

	//	Step 4.  Adjust the left and right clipping planes for off-axis viewing.
	if (anEyeType != EyeOnly)
	{
		Matrix44Identity(theFactor);
		theFactor[3][0] = (anEyeType == EyeLeft ?
							- md->itsEyeOffset / md->itsWindowHalfWidth :
							+ md->itsEyeOffset / md->itsWindowHalfWidth );
		Matrix44Product(theFactor, md->itsProjectionMatrix, md->itsProjectionMatrix);
	}

	//	Step 3.  Adjust the side clipping planes to match the window geometry.
	Matrix44Identity(theFactor);
	theFactor[0][0] = md->itsWindowDistance / md->itsWindowHalfWidth;
	theFactor[1][1] = md->itsWindowDistance / md->itsWindowHalfHeight;
	Matrix44Product(theFactor, md->itsProjectionMatrix, md->itsProjectionMatrix);

	//	Step 2.  Adjust the near and far clipping planes.
	n = -INVERSE_NEAR_CLIP;
	f = (md->itsSpaceType == SpaceSpherical) ? +INVERSE_NEAR_CLIP : 0.0;
	Matrix44Identity(theFactor);
	theFactor[2][2] = 2/(f - n);
	theFactor[3][2] = (n + f)/(n - f);
	Matrix44Product(theFactor, md->itsProjectionMatrix, md->itsProjectionMatrix);

	//	Step 1.  Apply the quarter turn.
	Matrix44Identity(theFactor);
	theFactor[2][2] =  0.0;
	theFactor[2][3] = +1.0;
	theFactor[3][2] = -1.0;
	theFactor[3][3] =  0.0;
	Matrix44Product(theFactor, md->itsProjectionMatrix, md->itsProjectionMatrix);

	//	Step 5.  Translate the scenery to accommodate stereoscopic 3D.
	if (anEyeType != EyeOnly)
	{
		Matrix44Identity(theFactor);

		d = (anEyeType == EyeLeft) ?
			+ md->itsEyeOffset / md->itsIntrinsicUnit :
			- md->itsEyeOffset / md->itsIntrinsicUnit;

		switch (md->itsSpaceType)
		{
			case SpaceSpherical:
				theFactor[0][0] =  cos(d);	theFactor[0][3] = -sin(d);
				theFactor[3][0] =  sin(d);	theFactor[3][3] =  cos(d);
				break;

			case SpaceFlat:
				theFactor[0][0] =   1.0;  	theFactor[0][3] =   0.0;
				theFactor[3][0] =    d;   	theFactor[3][3] =   1.0;
				break;

			case SpaceHyperbolic:
				theFactor[0][0] = cosh(d);	theFactor[0][3] = sinh(d);
				theFactor[3][0] = sinh(d);	theFactor[3][3] = cosh(d);
				break;

			case SpaceNone:
				break;
		}

		Matrix44Product(theFactor, md->itsProjectionMatrix, md->itsProjectionMatrix);
	}

	//	Step 6b.  To draw the back hemisphere, invert all scenery.
	if (aSceneryInversionFlag == true)
	{
		Matrix44Identity(theFactor);
		theFactor[0][0] = -1.0;
		theFactor[1][1] = -1.0;
		theFactor[2][2] = -1.0;
		theFactor[3][3] = -1.0;
		Matrix44Product(theFactor, md->itsProjectionMatrix, md->itsProjectionMatrix);
	}
}


static void DrawTheScene(ModelData *md)
{
#ifdef START_OUTSIDE
	if (md->itsViewpoint == ViewpointIntrinsic)
		DrawTheSceneIntrinsically(md);
	else
		DrawTheSceneExtrinsically(md);
#else
	DrawTheSceneIntrinsically(md);
#endif
}


static void DrawTheSceneIntrinsically(ModelData *md)
{
	Matrix	theViewMatrix,
			theProjectionMatrix,
			theViewProjectionMatrix,
			theSpin,
			theTilt,
			theOrientation,	//	“orientation” in the sense of an element of O(3),
							//		not a connected component of O(3)
			thePlacement;

	//	Set the current placement.
	//	The view matrix is the inverse of the eye matrix.
	MatrixGeometricInverse(&md->itsUserPlacement, &theViewMatrix);

	//	Compute the viewprojection transformation
	//	(into clipping coordinates) for use in culling.
	Matrix44Copy(theProjectionMatrix.m, md->itsProjectionMatrix);
	theProjectionMatrix.itsParity = ImagePositive;	//	parity will be ignored
	MatrixProduct(&theViewMatrix, &theProjectionMatrix, &theViewProjectionMatrix);

	//	Determine which cells are visible relative
	//	to the current viewprojection matrix, and sort them
	//	in order of increasing distance from the observer
	//	(so transparency effects come out right).
	SortVisibleCells(	md->itsHoneycomb,
						&theViewProjectionMatrix,
						&theViewMatrix,
						md->itsDrawingRadius);

	//	Draw all visible translates of the Dirichlet domain.
	if (md->itsCurrentAperture > 0.0)
	{
		BindDirichletVBOs(	md->itsVertexBufferNames[VertexBufferDirichlet],
							md->itsIndexBufferNames [VertexBufferDirichlet],
							md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
		DrawDirichletVBOs(	md,
							md->itsDirichletDomain,
							md->itsHoneycomb,
							&theViewMatrix,
							md->itsShowColorCoding);
	}

	//	Draw all visible translates of the observer if desired.
	if (md->itsShowObserver)
	{
		//	The central image of the Dirichlet domain never moves relative to space,
		//	so the observer's position in the Dirichlet domain is the same
		//	as his/her position in space.
		BindObserverVBOs(	md->itsVertexBufferNames[VertexBufferObserver],
							md->itsIndexBufferNames [VertexBufferObserver],
							md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
		DrawObserverVBOs(	md,
							md->itsHoneycomb,
							&theViewMatrix,
							&md->itsUserPlacement);
	}

	//	Draw all visible translates of the vertex figures if desired.
	if (md->itsShowVertexFigures == true)
	{
		BindVertexFiguresVBOs(	md->itsVertexBufferNames[VertexBufferVertexFigures],
								md->itsIndexBufferNames [VertexBufferVertexFigures],
								md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
		DrawVertexFiguresVBOs(	md,
								md->itsDirichletDomain,
								md->itsHoneycomb,
								&theViewMatrix);
	}

	//	Draw Clifford parallels if desired.
	if (md->itsSpaceType == SpaceSpherical
	 && md->itsShowCliffordParallels == true)
	{
		BindCliffordVBOs(	md->itsVertexBufferNames[VertexBufferClifford],
							md->itsIndexBufferNames [VertexBufferClifford],
							md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
		DrawCliffordVBOs(	md,
							&theViewMatrix);
	}

	//	Draw all visible translates of the centerpiece if desired.
	//	The centerpiece gets drawn last, because it may be partially transparent
	//	(at present the galaxy is partially transparent, but the Earth and gyroscope are not).
	//	Transparent objects must be draw last, and in strict back-to-front order.
	if (md->itsSceneryType != SceneryNone)
	{
		//	How should we position the centerpiece?
		switch (md->itsSceneryType)
		{
			case SceneryEarth:
				MatrixRotation(&theSpin, 0.0, 0.0,   EARTH_SPEED   * md->itsRotationAngle);
				MatrixRotation(&theTilt, -PI/2, 0.0, 0.0);	//	takes Earth's spin axis from z-axis to y-axis
				break;

			case SceneryGalaxy:
				MatrixRotation(&theSpin, 0.0, 0.0,   GALAXY_SPEED  * md->itsRotationAngle);
				MatrixRotation(&theTilt, 0.2, 0.3, 0.0);	//	arbitrary spin axis that looks nice
				break;

			case SceneryGyroscope:
				MatrixRotation(&theSpin, 0.0, 0.0, GYROSCOPE_SPEED * md->itsRotationAngle);
				MatrixRotation(&theTilt, -PI/2, 0.0, 0.0);	//	takes gyroscope's spin axis from z-axis to y-axis
				break;

			default:	//	should never occur
				MatrixRotation(&theSpin, 0.0, 0.0, 0.0);
				MatrixRotation(&theTilt, 0.0, 0.0, 0.0);
				break;
		}
		
		//	Keeping in mind our left-to-right matrix conventions 
		//	(see “Curved Spaces 3/Read Me.txt”), first apply theSpin,
		//	then theTilt, and then the overall placement if it's enabled.
		MatrixProduct(&theSpin, &theTilt, &theOrientation);
#ifdef CENTERPIECE_DISPLACEMENT
		MatrixProduct(&theOrientation, &md->itsCenterpiecePlacement, &thePlacement);
#else
		thePlacement = theOrientation;
#endif

		//	Draw all visible translates of the centerpiece.
		switch (md->itsSceneryType)
		{
			case SceneryEarth:
				BindEarthVBOs(	md->itsVertexBufferNames[VertexBufferEarth],
								md->itsIndexBufferNames [VertexBufferEarth],
								md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
				DrawEarthVBOs(	md,
								md->itsHoneycomb,
								&theViewMatrix,
								&thePlacement);
				break;

			case SceneryGalaxy:
				BindGalaxyVBO(	md->itsVertexBufferNames[VertexBufferGalaxy],
								md->itsIndexBufferNames [VertexBufferGalaxy],
								md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
				DrawGalaxyVBO(	md,
								md->itsHoneycomb,
								&theViewMatrix,
								&thePlacement);
				break;

			case SceneryGyroscope:
				BindGyroscopeVBOs(	md->itsVertexBufferNames[VertexBufferGyroscope],
									md->itsIndexBufferNames [VertexBufferGyroscope],
									md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
				DrawGyroscopeVBOs(	md,
									md->itsHoneycomb,
									&theViewMatrix,
									&thePlacement);
				break;

			default:	//	should never occur
				break;
		}
	}
}


#ifdef START_OUTSIDE

static void DrawTheSceneExtrinsically(ModelData *md)
{
	Matrix	theTranslation,
			theRotation,
			thePlacement,
			theSpin,
			theTilt,
			theOrientation;	//	“orientation” in the sense of an element of O(3),
							//		not a connected component of O(3)
	
	//	Set up a Honeycomb containing the identity matrix alone.
	//	Omit fields related to depth sorting -- DrawDirichletVBOs() will ignore them.
	static Honeycell	theIdentityCell =
						{
							{
								{
									{1.0, 0.0, 0.0, 0.0},
									{0.0, 1.0, 0.0, 0.0},
									{0.0, 0.0, 1.0, 0.0},
									{0.0, 0.0, 0.0, 1.0}
								},
								ImagePositive
							},
							{{0.0, 0.0, 0.0, 1.0}},	//	ignored (but nevertheless correct!)
							NULL,					//	ignored
							0,						//	ignored
							0.0						//	ignored (but nevertheless correct!)
						},
						*theSingletonArray[1] =
						{
							&theIdentityCell
						};
	static Honeycomb	theSingletonHoneycomb =
						{
							0,		//	ignored
							NULL,	//	ignored
							1,
							theSingletonArray
						};

	//	Do we have a space loaded?
	if (md->itsSpaceType == SpaceNone)
		return;

	//	Disable depth testing.
	//	Normally it would be harmless, except that the transparent corners
	//	of the galaxy square extend slightly beyond the faces of the
	//	Poincaré dodecahedral space's fundamental polyhedron.
	glDisable(GL_DEPTH_TEST);

	//	The current placement consists of a rotation
	//	followed by a translation.
	MatrixGeometricInverse(&md->itsUserPlacement, &theRotation);
	MatrixTranslation(	&theTranslation,
						md->itsSpaceType,
						0.0,
						0.0,
						md->itsViewpointTransition * EXTRINSIC_VIEWING_DISTANCE);
	MatrixProduct(&theRotation, &theTranslation, &thePlacement);

	//	Draw the Dirichlet domain's inside faces.
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	BindDirichletVBOs(	md->itsVertexBufferNames[VertexBufferDirichlet],
						md->itsIndexBufferNames [VertexBufferDirichlet],
						md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
	DrawDirichletVBOs(	md,
						md->itsDirichletDomain,
						&theSingletonHoneycomb,
						&thePlacement,
						md->itsShowColorCoding);

	//	Draw the centerpiece.
	//	Please remember that DrawTheSceneExtrinsically()
	//	is just a quick hack for personal use!
	MatrixRotation(&theSpin, 0.0, 0.0, GALAXY_SPEED  * md->itsRotationAngle);
	MatrixRotation(&theTilt, 0.2, 0.3, 0.0);	//	arbitrary spin axis that looks nice
	MatrixProduct(&theSpin, &theTilt, &theOrientation);
	BindGalaxyVBO(	md->itsVertexBufferNames[VertexBufferGalaxy],
					md->itsIndexBufferNames [VertexBufferGalaxy],
					md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
	DrawGalaxyVBO(	md,
					&theSingletonHoneycomb,
					&thePlacement,
					&theOrientation);

	//	Draw the Dirichlet domain's outside faces.
	glEnable(GL_CULL_FACE);
	glCullFace(GL_FRONT);
	BindDirichletVBOs(	md->itsVertexBufferNames[VertexBufferDirichlet],
						md->itsIndexBufferNames [VertexBufferDirichlet],
						md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
	DrawDirichletVBOs(	md,
						md->itsDirichletDomain,
						&theSingletonHoneycomb,
						&thePlacement,
						md->itsShowColorCoding);

	//	Just for good form...
	glCullFace(GL_BACK);

	//	Re-enable depth testing.
	glEnable(GL_DEPTH_TEST);
}

#endif


void SendTransformationsToShader(ModelData *md)
{
	double	theModelViewProjectionMatrix[4][4];
	float	theModelViewProjectionMatrixFloat[4][4],
			theModelViewMatrixFloat[4][4];

	Matrix44Product(md->itsModelViewMatrix, md->itsProjectionMatrix, theModelViewProjectionMatrix);
	Matrix44DoubleToFloat(theModelViewProjectionMatrixFloat, theModelViewProjectionMatrix);
	Matrix44DoubleToFloat(theModelViewMatrixFloat, md->itsModelViewMatrix);

	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		//	for position
		glVertexAttrib4fv(ATTRIBUTE_MVP_MATRIX_ROW_0, theModelViewProjectionMatrixFloat[0]);
		glVertexAttrib4fv(ATTRIBUTE_MVP_MATRIX_ROW_1, theModelViewProjectionMatrixFloat[1]);
		glVertexAttrib4fv(ATTRIBUTE_MVP_MATRIX_ROW_2, theModelViewProjectionMatrixFloat[2]);
		glVertexAttrib4fv(ATTRIBUTE_MVP_MATRIX_ROW_3, theModelViewProjectionMatrixFloat[3]);

		//	for fog
		glVertexAttrib4fv(ATTRIBUTE_MV_MATRIX_ROW_0, theModelViewMatrixFloat[0]);
		glVertexAttrib4fv(ATTRIBUTE_MV_MATRIX_ROW_1, theModelViewMatrixFloat[1]);
		glVertexAttrib4fv(ATTRIBUTE_MV_MATRIX_ROW_2, theModelViewMatrixFloat[2]);
		glVertexAttrib4fv(ATTRIBUTE_MV_MATRIX_ROW_3, theModelViewMatrixFloat[3]);
	}
	else	//	no shaders
	{
		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf((float *)theModelViewProjectionMatrixFloat);
	}
}


static ErrorText GetErrorString(void)
{
	switch (glGetError())
	{
		case GL_NO_ERROR:
			return NULL;
		case GL_INVALID_ENUM:
			return "GL_INVALID_ENUM:  GLenum argument out of range.";
		case GL_INVALID_VALUE:
			return "GL_INVALID_VALUE:  Numeric argument out of range.";
		case GL_INVALID_OPERATION:
			return "GL_INVALID_OPERATION:  Operation illegal in current state.";
		case GL_STACK_OVERFLOW:
			return "GL_STACK_OVERFLOW:  Command would cause a stack overflow.";
		case GL_STACK_UNDERFLOW:
			return "GL_STACK_UNDERFLOW:  Command would cause a stack underflow.";
		case GL_OUT_OF_MEMORY:
			return "GL_OUT_OF_MEMORY:  Not enough memory left to execute command.";
		default:
			return "Unknown OpenGL error.";
	}
}
