//	CurvedSpacesInit.c
//
//	Initializes the ModelData.
//	Prepares OpenGL textures, Vertex Buffer Objects and vertex shaders.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#include "CurvedSpaces-Common.h"
#include <math.h>	//	for floor()
#include <string.h>	//	for strlen()

#ifdef __APPLE__
#include <OpenGL/gl.h>
#endif
#ifdef __WIN32__
#include "GeometryGames-Win32-OpenGLEntryPoints.h"
#endif


//	RGBA pixel
typedef struct
{
	Byte	r,	//	red
			g,	//	green
			b,	//	blue
			a;	//	alpha = opacity
} PixelRGBA;

//	RGBA image, as read from a .rgba file
typedef struct
{
	//	Keep a pointer to the raw data,
	//	exactly as it was read from the file.
	unsigned int	itsNumRawBytes;
	Byte			*itsRawBytes;
	
	//	Interpret the data as a width, a height and some pixels.
	unsigned int	itsWidth,
					itsHeight;
	PixelRGBA		*itsPixels;	//	= itsRawBytes + 8

} ImageRGBA;


static void			SetUpExtensions(ModelData *md);
static bool			WholeWordSubstring(const char *aString, const char *aPotentialSubstring);
static void			ShutDownExtensions(ModelData *md);
static ErrorText	SetUpShaders(ModelData *md);
static ErrorText	SetUpOneShaderProgram(unsigned int *aShaderProgram, ShaderIndex aShaderIndex);
static void			ShutDownShaders(ModelData *md);
static ErrorText	SetUpTextures(ModelData *md);
static void			SetUpOneTexture(ModelData *md, unsigned int aTextureIndex, bool aClampFlag, bool aGreyscaleFlag, ErrorText *aFirstError);
static ErrorText	ReadImageRGBA(unsigned int aTextureIndex, bool aGreyscaleFlag, ImageRGBA **anImageRGBA);
static void			FreeImageRGBA(ImageRGBA **anImage);
static char			*GetTextureFileName(unsigned int aTextureIndex);
static void			ShutDownTextures(ModelData *md);
static ErrorText	SetUpVBOs(ModelData *md);
static void			ShutDownVBOs(ModelData *md);


bool ShadersAreAvailable(void)
{
	const char	*theExtensionsString;

	//	Assume the OpenGL context has already been set.
	
	theExtensionsString = (const char *) glGetString(GL_EXTENSIONS);

	if (theExtensionsString != NULL)
		return
			WholeWordSubstring(theExtensionsString, "GL_ARB_shader_objects" )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_vertex_shader"  )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_fragment_shader");
	else
		return false;
}


void SetUpModelData(ModelData *md)
{
	unsigned int	i;

	//	All OpenGL elements require initialization.
	md->itsPreparedExtensions	= false;
	md->itsPreparedShaders		= false;
	md->itsPreparedTextures		= false;
	md->itsPreparedVBOs			= false;

	//	We'll test for OpenGL extensions at the last minute,
	//	just before rendering the first frame.
	//	For now set safe defaults.
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;
	
	//	No shaders, textures or vertex/index buffers are present yet.

	for (i = 0; i < NumShaders; i++)
		md->itsShaderPrograms[i] = 0;

	for (i = 0; i < NumTextures; i++)
		md->itsTextureNames[i] = 0;

	for (i = 0; i < NumVertexBuffers; i++)
	{
		md->itsVertexBufferNames[i] = 0;
		md->itsIndexBufferNames [i] = 0;
	}

	//	Leave stereo off by default.
	md->itsStereoFlag	= false;

	//	Enable synchronization by default, not only to save cycles
	//	but also to allow smoother transitions across faces
	//	of the Dirichlet domain, where rendering time takes a jump.
	md->itsSyncFlag		= true;

	//	Multisampling takes place iff itsMultisamplingAvailableFlag
	//	and itsMultisamplingEnabledFlag are both true.
	//
	//	StarterCodeMacOSView's initWithFrame method (Mac)
	//	or GetUserPref() itself (Win32) sets the default value.
	md->itsMultisamplingEnabledFlag	= GetUserPref("multisampling");

	//	Enable anisotropic texture filtering by default.
	//	(Anisotropic texture filtering takes place iff
	//	itsAnisotropicAvailableFlag and itsAnisotropicEnabledFlag
	//	are both true.)
	md->itsAnisotropicEnabledFlag	= true;
	
	//	Programmable shaders are used iff itsShadersAvailableFlag 
	//	and itsShadersEnabledFlag are both true.
	//
	//	CurvedSpacesMacOSView's initWithFrame method (Mac)
	//	or GetUserPref() itself (Win32) sets the default value.
//	md->itsShadersEnabledFlag		= GetUserPref("shaders");
//IGNORE USER PREF AND ALWAYS START WITH SHADERS DISABLED,
//FOR MAXIMAL ROBUSTNESS IN THIS ARCHIVED VERSION OF THE PROGRAM.
md->itsShadersEnabledFlag = false;

	//	Set up window dimensions.
	//
	//	For a realistic simulation, the program's internal parameters
	//	must match the relative proportions of the user's physical display device,
	//	the distance from the display to the user's face, and the separation
	//	between the user's eyes.
	//	Initialize the parameters for a 90 field of view.
	//	In monoscopic 3D this gives a nice wide-angle view.
	//	In stereoscopic 3D a 90 field of view will be practical
	//	only with a very large display -- for example in a lecture hall --
	//	relative to which the separation between the user's eyes is small.
	//	Accurate stereoscopic 3D is more difficult on an ordinary monitor,
	//	but is workable in the flat space case if you set itsWindowDistance,
	//	itsWindowHalfWidth and itsWindowHalfHeight correctly, but underreport
	//	the separation between your eyes.
	//	At runtime the user may use the Preferences panel
	//	to override the defaults given here.
	md->itsWindowDistance	= 100.0;
	md->itsWindowHalfWidth	= 100.0;
	md->itsWindowHalfHeight	= 100.0;
	md->itsEyeOffset		=   2.5;	//	nose to eye
	md->itsIntrinsicUnit	= 400.0;

	//	No redraw request is pending.
	md->itsRedrawRequestFlag	= false;


	//	Set up application-specific data.

	md->itsSpaceType				= SpaceNone;
	md->itsDrawBackHemisphere		= false;
	md->itsTilingRadius				= 0.0;	//	LoadGenerators() will set the tiling radius.
	md->itsDrawingRadius			= 0.0;	//	LoadGenerators() will set the drawing radius.
	MatrixIdentity(&md->itsUserPlacement);
	md->itsUserSpeed				= 0.0;	//	LoadGenerators() will set the speed.

#ifdef CENTERPIECE_DISPLACEMENT
	MatrixIdentity(&md->itsCenterpiecePlacement);
#endif

	md->itsDirichletDomain			= NULL;
	md->itsHoneycomb				= NULL;

#if defined(START_OUTSIDE)
	md->itsDesiredAperture			= 1.00;
	md->itsCurrentAperture			= 1.00;
	md->itsSceneryType				= SceneryGalaxy;
#elif defined(CENTERPIECE_DISPLACEMENT)
	md->itsDesiredAperture			= 1.00;
	md->itsCurrentAperture			= 1.00;
	md->itsSceneryType				= SceneryEarth;
#elif defined(START_STILL)
	md->itsDesiredAperture			= 1.00;
	md->itsCurrentAperture			= 1.00;
	md->itsSceneryType				= SceneryEarth;
#elif defined(START_WALLS_OPEN)
	md->itsDesiredAperture			= 0.125;
	md->itsCurrentAperture			= 0.125;
	md->itsSceneryType				= SceneryEarth;
#else
	md->itsDesiredAperture			= 0.75;
	md->itsCurrentAperture			= 0.75;
	md->itsSceneryType				= SceneryEarth;
#endif

	md->itsRotationAngle			= 0.0;

#ifdef START_WALLS_OPEN
	md->itsShowObserver				= false;
#else
	md->itsShowObserver				= true;
#endif

#ifdef CENTERPIECE_DISPLACEMENT
	md->itsShowColorCoding			= true;
#else
	md->itsShowColorCoding			= false;
#endif

	md->itsShowCliffordParallels	= false;
	md->itsShowVertexFigures		= false;
	md->itsFogFlag					= true;
	md->itsFogSaturation			= 1.0;

#ifdef START_OUTSIDE
	md->itsViewpoint				= ViewpointExtrinsic;
	md->itsViewpointTransition		= 1.0;
	md->itsExtrinsicRotation		= 0.0;
#endif
}


void ShutDownModelData(ModelData *md)
{
	//	Free any allocated memory.
	//	Leave other information untouched.
	FreeDirichletDomain(&md->itsDirichletDomain);
	FreeHoneycomb(&md->itsHoneycomb);
}


ErrorText SetUpGraphicsAsNeeded(ModelData *md)
{
	//	These are one-time initializations,
	//	so usually no work is required here.
	//	Only in exceptional circumstances, for example
	//	when the user has selected a new manifold,
	//	will something need an update.

	//	Assume the OpenGL context has already been set.
	
	ErrorText	theError;

	if (md->itsPreparedExtensions == false)
	{
		SetUpExtensions(md);
		md->itsPreparedExtensions = true;

		//	Availability of extensions may influence
		//	the construction of textures, etc.,
		//	so rebuild everything from scratch.
		md->itsPreparedShaders	= false;
		md->itsPreparedTextures	= false;
		md->itsPreparedVBOs		= false;
	}

	if (md->itsPreparedShaders == false)
	{
		if ((theError = SetUpShaders(md)) != NULL)
			return theError;
		md->itsPreparedShaders	= true;
	}

	if (md->itsPreparedTextures == false)
	{
		if ((theError = SetUpTextures(md)) != NULL)
			return theError;
		md->itsPreparedTextures	= true;
	}

	if (md->itsPreparedVBOs == false)
	{
		SetUpVBOs(md);
		md->itsPreparedVBOs		= true;
		md->itsPreparedAperture	= true;
	}
	
	//	When the user resizes the aperture, recompute the Dirichlet VBOs.
	if (md->itsPreparedAperture == false)
	{
		if ((theError = MakeDirichletVBOs(	md->itsVertexBufferNames[VertexBufferDirichlet],
											md->itsIndexBufferNames [VertexBufferDirichlet],
											md->itsDirichletDomain,
											md->itsCurrentAperture,
											md->itsShowColorCoding,
											md->itsStereoFlag)) != NULL)
			return theError;
		md->itsPreparedAperture = true;
	}
	
	return NULL;
}


void ShutDownGraphicsAsNeeded(ModelData *md)
{
	//	Assume the OpenGL context has already been set.

	//	Shut down the various elements in the opposite order
	//	from that in which they were created.  In particular,
	//	be sure to shut down the extensions last, so the other
	//	shutdown routines will know what extensions have been
	//	available.  For example, ShutDownShaders() will need
	//	to know whether shaders have been available or not.
	ShutDownVBOs(md);
	ShutDownTextures(md);
	ShutDownShaders(md);
	ShutDownExtensions(md);

	md->itsPreparedExtensions	= false;
	md->itsPreparedShaders		= false;
	md->itsPreparedTextures		= false;
	md->itsPreparedVBOs			= false;
}


static void SetUpExtensions(ModelData *md)
{
	const char	*theExtensionsString;

	//	Desired OpenGL extensions are assumed absent until found.
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;

	//	Get the renderer's extensions string.
	theExtensionsString = (const char *) glGetString(GL_EXTENSIONS);

	if (theExtensionsString != NULL)
	{
		//	Is multisampling available?
		if (WholeWordSubstring(theExtensionsString, "GL_ARB_multisample"))
			md->itsMultisamplingAvailableFlag = true;

		//	Is anisotropic texture filtering available?
		if (WholeWordSubstring(theExtensionsString, "GL_EXT_texture_filter_anisotropic"))
			md->itsAnisotropicAvailableFlag = true;

		//	Are programmable shaders available?
		if (WholeWordSubstring(theExtensionsString, "GL_ARB_shader_objects" )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_vertex_shader"  )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_fragment_shader"))
			md->itsShadersAvailableFlag = true;
	}
}


static bool WholeWordSubstring(
	const char	*aString,				//	null-terminated
	const char	*aPotentialSubstring)	//	null-terminated
{
	const char	*thePotentialMatch,
				*s,
				*p;
	bool		theMatchFlag;

	//	Does aPotentialSubstring occur as a complete space-delimited word in aString?

	//	Irrelevant technical detail:  an empty aPotentialSubstring
	//	(consisting of a terminating zero alone) counts as a substring
	//	of any string, even an empty one.

	//	Consider the first character in each space-delimited word of aString
	//	as the starting point for a potential match.
	thePotentialMatch = aString;
	while (*thePotentialMatch == ' ')
		thePotentialMatch++;
	do
	{
		//	Compare each character in aPotentialSubstring
		//	(excluding the terminating zero) to thePotentialMatch.
		s = thePotentialMatch;
		p = aPotentialSubstring;
		theMatchFlag = true;
		while (theMatchFlag == true && *p != 0)
		{
			if (*p++ != *s++)
				theMatchFlag = false;
		}

		//	If a space or terminating 0 follows a match, return true.
		if (theMatchFlag == true
		 && (*s == ' ' || *s == 0))
			return true;

		//	Skip to the beginning of the next whole word in thePotentialMatch.
		//		Skip by non-spaces.
		while (*thePotentialMatch != ' ' && *thePotentialMatch != 0)
			thePotentialMatch++;
		//		Skip by intervening spaces.
		while (*thePotentialMatch == ' ')
			thePotentialMatch++;

		//	Continue with the loop to examine the next whole word (if any).

	} while (*thePotentialMatch != 0);

	return false;
}


static void ShutDownExtensions(ModelData *md)
{
	//	Just for good form...
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;
}


static ErrorText SetUpShaders(ModelData *md)
{
	ErrorText		theError;
	unsigned int	i;

	if (md->itsShadersAvailableFlag)
	{
		//	Delete any pre-existing shaders.
		ShutDownShaders(md);

		//	If shaders are desired, then...
		if (md->itsShadersEnabledFlag)
		{
			//	...create a shader of each type.
			//
			//	At run time, swapping simple shaders in and out as needed turns out
			//	to be faster than running a single all-purpose shader with options,
			//	because (much to my surprise) an if/then block containing
			//	texture sampling slows the fragment shader to a crawl
			//	(at least on my Radeon X1600).
			for (i = 0; i < NumShaders; i++)
			{
				theError = SetUpOneShaderProgram(&md->itsShaderPrograms[i], (ShaderIndex) i );
				if (theError != NULL)
					return theError;
			}
		}
	}
	
	return NULL;
}


static ErrorText SetUpOneShaderProgram(
	unsigned int	*aShaderProgram,	//	output
	ShaderIndex		aShaderIndex)		//	input
{
	ErrorText		theErrorMessage					= NULL;
	char			*theVertexShaderFileName		= NULL,
					*theFragmentShaderFileName		= NULL,
					theFilePath[2048];
	unsigned int	theVertexShaderSourceLength		= 0,
					theFragmentShaderSourceLength	= 0;
	Byte			*theVertexShaderSource			= NULL,
					*theFragmentShaderSource		= NULL;
	unsigned int	theVertexShader					= 0,
					theFragmentShader				= 0;
	GLint			theSuccessFlag;

	static GLchar	theInfoLog[1024] = {0};	//	static so it persists after this function returns
	
	//	For safe error handling.
	//	(*aShaderProgram should already be 0, but reset it anyway.)
	*aShaderProgram = 0;
	
	//	Select an appropriate pair of shaders for the given ShaderIndex.
	switch (aShaderIndex)
	{
		case ShaderSph:
			theVertexShaderFileName		= "CurvedSpaces-Sph.vs";
			theFragmentShaderFileName	= "CurvedSpaces-Sph.fs";
			break;

		case ShaderEuc:
			theVertexShaderFileName		= "CurvedSpaces-Euc.vs";
			theFragmentShaderFileName	= "CurvedSpaces-Euc.fs";
			break;

		case ShaderHyp:
			theVertexShaderFileName		= "CurvedSpaces-Hyp.vs";
			theFragmentShaderFileName	= "CurvedSpaces-Hyp.fs";
			break;

		default:
			theErrorMessage = "Invalid shader index";
			goto CleanUpSetUpOneShaderProgram;
	}

	//	Load the vertex shader source code.
	if ((theErrorMessage = GetAbsolutePath(	"Shaders", 
											theVertexShaderFileName,
											theFilePath,
											BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	if ((theErrorMessage = GetFileContents(theFilePath, &theVertexShaderSourceLength, &theVertexShaderSource)) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	
	//	Load the fragment shader source code.
	if ((theErrorMessage = GetAbsolutePath(	"Shaders",
											theFragmentShaderFileName,
											theFilePath,
											BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	if ((theErrorMessage = GetFileContents(theFilePath, &theFragmentShaderSourceLength, &theFragmentShaderSource)) != NULL)
		goto CleanUpSetUpOneShaderProgram;

	//	Create the vertex shader.
	theVertexShader = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(	theVertexShader,
					1,
					(const char *[1]){(const char *)theVertexShaderSource},
					(GLint [1]){theVertexShaderSourceLength});
	glCompileShader(theVertexShader);
	glGetShaderiv(theVertexShader, GL_COMPILE_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetShaderInfoLog(theVertexShader, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

	//	Create the fragment shader.
	theFragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(	theFragmentShader,
					1,
					(const char *[1]){(const char *)theFragmentShaderSource},
					(GLint [1]){theFragmentShaderSourceLength});
	glCompileShader(theFragmentShader);
	glGetShaderiv(theFragmentShader, GL_COMPILE_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetShaderInfoLog(theFragmentShader, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

	//	Create the program and attach the vertex and fragment shaders.
	*aShaderProgram = glCreateProgram();
	if (*aShaderProgram == 0)
	{
		theErrorMessage = "Couldn't create *aShaderProgram";
		goto CleanUpSetUpOneShaderProgram;
	}
	glAttachShader(*aShaderProgram, theVertexShader  );
	glAttachShader(*aShaderProgram, theFragmentShader);
	
	//	Assign locations to vertex shader attributes.
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_POSITION,			"atrPosition"					);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_TEX_COORD,			"atrTextureCoordinates"			);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_COLOR,				"atrColor"						);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_MVP_MATRIX_ROW_0,	"atrModelViewProjectionMatrix"	);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_MV_MATRIX_ROW_0,	"atrModelViewMatrix"			);
	
	//	Link the program.
	glLinkProgram(*aShaderProgram);
	glGetProgramiv(*aShaderProgram, GL_LINK_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetProgramInfoLog(*aShaderProgram, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

CleanUpSetUpOneShaderProgram:

	//	Free our private copy of the source code.
	FreeFileContents(&theVertexShaderSourceLength,   &theVertexShaderSource  );
	FreeFileContents(&theFragmentShaderSourceLength, &theFragmentShaderSource);

	//	Release our temporary references (if any) to the vertex and fragments shaders.
	//	The shader program will keep its own references to the vertex and fragments shaders,
	//	and will automatically delete them when it itself gets deleted.  
	//	Passing 0 is harmless.
	glDeleteShader(theVertexShader);
	glDeleteShader(theFragmentShader);

	//	If an error occured, delete the shader program (if any).
	if (theErrorMessage != NULL)
		glDeleteProgram(*aShaderProgram);	//	OK to pass 0

	return theErrorMessage;
}


static void ShutDownShaders(ModelData *md)
{
	unsigned int	i;

	if (md->itsShadersAvailableFlag)
	{
		glUseProgram(0);

		for (i = 0; i < NumShaders; i++)
		{
			glDeleteProgram(md->itsShaderPrograms[i]);
			md->itsShaderPrograms[i] = 0;
		}
	}
}


static ErrorText SetUpTextures(ModelData *md)
{
	ErrorText	theFirstError	= NULL;
	
	//	Release any pre-existing textures.
	ShutDownTextures(md);
	
	//	Generate new texture names.
	glGenTextures(NumTextures, (GLuint *) md->itsTextureNames);

	//	Set up the individual textures.
	SetUpOneTexture(md, TextureWallPaper,		false,	false,				&theFirstError);
	SetUpOneTexture(md, TextureWallWood,		false,	false,				&theFirstError);
	SetUpOneTexture(md, TextureEarth,			true,	md->itsStereoFlag,	&theFirstError);
	SetUpOneTexture(md, TextureGalaxy,			true,	md->itsStereoFlag,	&theFirstError);
	SetUpOneTexture(md, TextureGyroscope,		true,	false,				&theFirstError);
	SetUpOneTexture(md, TextureObserver,		true,	false,				&theFirstError);
	SetUpOneTexture(md, TextureVertexFigures,	false,	md->itsStereoFlag,	&theFirstError);
	SetUpOneTexture(md, TextureClifford,		false,	false,				&theFirstError);

	return theFirstError;
}


static void SetUpOneTexture(
	ModelData		*md,
	unsigned int	aTextureIndex,
	bool			aClampFlag,
	bool			aGreyscaleFlag,
	ErrorText		*aFirstError)
{
	ErrorText	theErrorMessage	= NULL;
	ImageRGBA	*theImageRGBA	= NULL;

	//	Initialize the OpenGL texture options.
	InitTexture(md, aTextureIndex, aClampFlag);

	//	Read the texture file.
	if ((theErrorMessage = ReadImageRGBA(aTextureIndex, aGreyscaleFlag, &theImageRGBA)) != NULL)
		goto CleanUpSetUpOneTexture;

	//	Pass the pixel data to OpenGL.
	SetTextureImage(md,
					aTextureIndex,
					theImageRGBA->itsWidth,
					theImageRGBA->itsHeight,
					(Byte (*)[4]) theImageRGBA->itsPixels);

CleanUpSetUpOneTexture:

	//	Free the image (if necessary).
	FreeImageRGBA(&theImageRGBA);

	if (theErrorMessage != NULL)
	{
		//	Substitute a pure red texture for the missing one.
		SetTextureImage(md, aTextureIndex, 1, 1, (Byte [1][4]){{0xFF, 0x00, 0x00, 0xFF}});
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}

static ErrorText ReadImageRGBA(
	unsigned int	aTextureIndex,	//	input
	bool			aGreyscaleFlag,	//	input
	ImageRGBA		**anImageRGBA)	//	output
{
	ErrorText		theErrorMessage		= NULL;
	ImageRGBA		*theImage			= NULL;
	char			theFilePath[4096]	= "";
	Byte			theLuminance;
	unsigned int	i;
	
	//	Check the output variable.
	if (anImageRGBA == NULL)
	{
		theErrorMessage = "anImageRGBA == NULL in ReadImageRGBA()";
		goto CleanUpReadImageRGBA;
	}
	if (*anImageRGBA != NULL)
	{
		theErrorMessage = "*anImageRGBA != NULL in ReadImageRGBA()";
		goto CleanUpReadImageRGBA;
	}

	//	Allocate and initialize an empty ImageRGBA.
	theImage = GET_MEMORY(sizeof(ImageRGBA));
	if (theImage == NULL)
		goto CleanUpReadImageRGBA;
	theImage->itsNumRawBytes	= 0;
	theImage->itsRawBytes		= NULL;
	theImage->itsWidth			= 0;
	theImage->itsHeight			= 0;
	theImage->itsPixels			= NULL;

	//	Read the texture file's bytes.
	if ((theErrorMessage = GetAbsolutePath("Textures", GetTextureFileName(aTextureIndex), theFilePath, BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpReadImageRGBA;
	if ((theErrorMessage = GetFileContents(theFilePath, &theImage->itsNumRawBytes, &theImage->itsRawBytes)) != NULL)
		goto CleanUpReadImageRGBA;

	//	Did we get a header?
	if (theImage->itsNumRawBytes < 8)
	{
		theErrorMessage = "Texture file lacks width and/or height specification.";
		goto CleanUpReadImageRGBA;
	}

	//	Parse the image width and height as big-endian 4-byte integers.
	theImage->itsWidth  = ((unsigned int)theImage->itsRawBytes[0] << 24)
						+ ((unsigned int)theImage->itsRawBytes[1] << 16)
						+ ((unsigned int)theImage->itsRawBytes[2] <<  8)
						+ ((unsigned int)theImage->itsRawBytes[3] <<  0);
	theImage->itsHeight = ((unsigned int)theImage->itsRawBytes[4] << 24)
						+ ((unsigned int)theImage->itsRawBytes[5] << 16)
						+ ((unsigned int)theImage->itsRawBytes[6] <<  8)
						+ ((unsigned int)theImage->itsRawBytes[7] <<  0);

	//	OpenGL requires power-of-two dimensions.
	if (IsPowerOfTwo(theImage->itsWidth ) == false
	 || IsPowerOfTwo(theImage->itsHeight) == false)
	{
		theErrorMessage = "OpenGL requires that each texture's width and height be powers of two.";
		goto CleanUpReadImageRGBA;
	}

	//	Is the total file length correct?
	if (theImage->itsNumRawBytes !=  8  +  4 * theImage->itsWidth * theImage->itsHeight)
	{
		theErrorMessage = "Number of pixels in texture file does not match stated width and height.";
		goto CleanUpReadImageRGBA;
	}

	//	Locate the pixels.
	theImage->itsPixels = (PixelRGBA *) (theImage->itsRawBytes + 8);

	//	Convert to greyscale if desired.
	//	The formula
	//
	//		luminance = 30% red + 59% green + 11% blue
	//
	//	appears widely on the internet, but with little explanation.
	//	Presumably its origins lie in human color perception.
	if (aGreyscaleFlag)
	{
		for (i = 0; i < theImage->itsWidth * theImage->itsHeight; i++)
		{
			theLuminance = (Byte) floor(0.5 + (
				0.30 * theImage->itsPixels[i].r
			  + 0.59 * theImage->itsPixels[i].g
			  + 0.11 * theImage->itsPixels[i].b));

			theImage->itsPixels[i].r = theLuminance;
			theImage->itsPixels[i].g = theLuminance;
			theImage->itsPixels[i].b = theLuminance;
		}
	}
	
	//	Assign theImage to the output variable.
	*anImageRGBA = theImage;

CleanUpReadImageRGBA:
	
	if (theErrorMessage != NULL)
	{
		FreeImageRGBA(&theImage);
		if (anImageRGBA != NULL)
			*anImageRGBA = NULL;
	}
	
	return theErrorMessage;
};

static void FreeImageRGBA(ImageRGBA **anImage)
{
	if (anImage != NULL && *anImage != NULL)
	{
		FreeFileContents(&(*anImage)->itsNumRawBytes, &(*anImage)->itsRawBytes);

		FREE_MEMORY_SAFELY(*anImage);
	}
}

static char *GetTextureFileName(	//	Returns texture file name as zero-terminated UTF-8 string.
	unsigned int	aTextureIndex)	//	input
{
	switch (aTextureIndex)
	{
		case TextureWallPaper:		return "Paper.rgba";
		case TextureWallWood:		return "Wood.rgba";
		case TextureEarth:			return "Earth.rgba";
		case TextureGalaxy:			return "Galaxy.rgba";
		case TextureGyroscope:		return "White.rgba";
		case TextureObserver:		return "White.rgba";
		case TextureVertexFigures:	return "Stone.rgba";
		case TextureClifford:		return "Clifford.rgba";

		default:				return NULL;
	}
}


static void ShutDownTextures(ModelData *md)
{
	unsigned int	i;

	//	The platform-dependent UI #includes StarterCode-Common.h,
	//	but shouldn't know about OpenGL.  To keep things clean,
	//	we've defined itsTextureNames as an array of unsigned int's
	//	rather than an array of GLuint's.
	//	For MacOS X compatibility, though, we need to explicitly cast them
	//	to GLuint's before passing the array to glDeleteTextures().

	//	Delete all textures.
	//	glDeleteTextures() will silently ignore zeros and invalid names.
	glDeleteTextures(NumTextures, (GLuint *) md->itsTextureNames);

	//	Clear all texture names.
	for (i = 0; i < NumTextures; i++)
		md->itsTextureNames[i] = 0;
}


static ErrorText SetUpVBOs(ModelData *md)
{
	ErrorText	theNonfatalError	= NULL,
				theReturnError		= NULL;

	//	Release any pre-existing buffers.
	ShutDownVBOs(md);
	
	//	Generate new buffer names.
	glGenBuffers(NumVertexBuffers, (GLuint *) md->itsVertexBufferNames);
	glGenBuffers(NumVertexBuffers, (GLuint *) md->itsIndexBufferNames );

	//	Set up the individual buffers.

	theNonfatalError = MakeDirichletVBOs(	md->itsVertexBufferNames[VertexBufferDirichlet],
											md->itsIndexBufferNames [VertexBufferDirichlet],
											md->itsDirichletDomain,
											md->itsCurrentAperture,
											md->itsShowColorCoding,
											md->itsStereoFlag);
	if (theNonfatalError != NULL && theReturnError == NULL)
		theReturnError = theNonfatalError;

	switch (md->itsSceneryType)
	{
		case SceneryNone:
			break;

		case SceneryEarth:
			//	The Earth is the only case that can return an error message,
			//	because it's the only centerpiece that relies on dynamically
			//	allocated memory.
			theNonfatalError = MakeEarthVBOs(	md->itsVertexBufferNames[VertexBufferEarth],
												md->itsIndexBufferNames [VertexBufferEarth]);
			if (theNonfatalError != NULL && theReturnError == NULL)
				theReturnError = theNonfatalError;
			break;

		case SceneryGalaxy:
			MakeGalaxyVBO(	md->itsVertexBufferNames[VertexBufferGalaxy],
							md->itsIndexBufferNames [VertexBufferGalaxy]);
			break;

		case SceneryGyroscope:
			MakeGyroscopeVBOs(	md->itsVertexBufferNames[VertexBufferGyroscope],
								md->itsIndexBufferNames [VertexBufferGyroscope],
								md->itsStereoFlag);
			break;
	}
	
	if (md->itsShowObserver)
		MakeObserverVBOs(	md->itsVertexBufferNames[VertexBufferObserver],
							md->itsIndexBufferNames [VertexBufferObserver],
							md->itsStereoFlag);

	if (md->itsShowVertexFigures)
	{
		theNonfatalError = MakeVertexFiguresVBOs(	md->itsVertexBufferNames[VertexBufferVertexFigures],
													md->itsIndexBufferNames [VertexBufferVertexFigures],
													md->itsDirichletDomain);
		if (theNonfatalError != NULL && theReturnError == NULL)
			theReturnError = theNonfatalError;
	}

	if (md->itsShowCliffordParallels)
		MakeCliffordVBOs(	md->itsVertexBufferNames[VertexBufferClifford],
							md->itsIndexBufferNames [VertexBufferClifford]);

	return theReturnError;
}


static void ShutDownVBOs(ModelData *md)
{
	unsigned int	i;

	//	The platform-dependent UI #includes StarterCode-Common.h,
	//	but shouldn't know about OpenGL.  To keep things clean,
	//	we've defined itsVertexBufferNames and itsIndexBufferNames
	//	as arrays of unsigned int's rather than arrays of GLuint's.
	//	For MacOS X compatibility, though, we need to explicitly cast them
	//	to GLuint's before passing the arrays to glDeleteBuffers().

	//	Delete all Vertex Buffer Objects.
	//	glDeleteBuffers() will silently ignore zeros and invalid names.
	glDeleteBuffers(NumVertexBuffers, (GLuint *) md->itsVertexBufferNames);
	glDeleteBuffers(NumVertexBuffers, (GLuint *) md->itsIndexBufferNames );

	//	Clear all buffer names.
	for (i = 0; i < NumVertexBuffers; i++)
	{
		md->itsVertexBufferNames[i] = 0;
		md->itsIndexBufferNames [i] = 0;
	}
}
