//	CurvedSpacesObserver.c
//
//	Makes, binds and draws Vertex Buffer Objects
//	for the colored dart that marks the observer's position.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#include "CurvedSpaces-Common.h"
#include <stddef.h>	//	for offsetof()
#include <math.h>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#endif
#ifdef __WIN32__
#include "GeometryGames-Win32-OpenGLEntryPoints.h"
#endif


//	How big should the dart be?
//
//	Note:  In a 3-sphere, the image of the dart at the user's origin
//	and/or antipodal point will display correctly iff the clipping distance
//	is at most about WIDTH/2.
#define HALF_LENGTH	0.050
#define RADIUS		0.017
#define WIDTH		0.004

//	What colors should the fletches be?
#define COLOR_FLETCH_LEFT	{1.0000, 0.0000, 0.0000, 1.0000}
#define COLOR_FLETCH_RIGHT	{0.0000, 1.0000, 0.0000, 1.0000}
#define COLOR_FLETCH_BOTTOM	{0.0000, 0.0000, 1.0000, 1.0000}
#define COLOR_FLETCH_TOP	{1.0000, 1.0000, 0.0000, 1.0000}


//	The observer Vertex Buffer Object (VBO) will contain
//	the following data for each of its vertices.
typedef struct
{
	float	pos[4],	//	position (x,y,z,w)
			col[4];	//	color (r,g,b,a)
} ObserverVBOData;

//	The observer Index Buffer Object (IBO) will contain
//	the following data for each of its faces.
typedef struct
{
	unsigned short	vtx[3];	//	three vertices
} ObserverIBOData;


//	Represent the observer's spaceship as a dart with four fletches.
static const ObserverVBOData	gVertices[4*5] =
								{
									//	left fletch
									{{  0.0,     0.0,   -HALF_LENGTH, 1.0}, COLOR_FLETCH_LEFT	},
									{{-WIDTH,  +WIDTH,  -HALF_LENGTH, 1.0}, COLOR_FLETCH_LEFT	},
									{{-WIDTH,  -WIDTH,  -HALF_LENGTH, 1.0}, COLOR_FLETCH_LEFT	},
									{{-RADIUS,   0.0,   -HALF_LENGTH, 1.0}, COLOR_FLETCH_LEFT	},
									{{  0.0,     0.0,   +HALF_LENGTH, 1.0}, COLOR_FLETCH_LEFT	},

									//	right fletch
									{{  0.0,     0.0,   -HALF_LENGTH, 1.0}, COLOR_FLETCH_RIGHT	},
									{{+WIDTH,  -WIDTH,  -HALF_LENGTH, 1.0}, COLOR_FLETCH_RIGHT	},
									{{+WIDTH,  +WIDTH,  -HALF_LENGTH, 1.0}, COLOR_FLETCH_RIGHT	},
									{{+RADIUS,   0.0,   -HALF_LENGTH, 1.0}, COLOR_FLETCH_RIGHT	},
									{{  0.0,     0.0,   +HALF_LENGTH, 1.0}, COLOR_FLETCH_RIGHT	},

									//	bottom fletch
									{{  0.0,     0.0,   -HALF_LENGTH, 1.0}, COLOR_FLETCH_BOTTOM	},
									{{-WIDTH,  -WIDTH,  -HALF_LENGTH, 1.0}, COLOR_FLETCH_BOTTOM	},
									{{+WIDTH,  -WIDTH,  -HALF_LENGTH, 1.0}, COLOR_FLETCH_BOTTOM	},
									{{  0.0,   -RADIUS, -HALF_LENGTH, 1.0}, COLOR_FLETCH_BOTTOM	},
									{{  0.0,     0.0,   +HALF_LENGTH, 1.0}, COLOR_FLETCH_BOTTOM	},

									//	top fletch
									{{  0.0,     0.0,   -HALF_LENGTH, 1.0}, COLOR_FLETCH_TOP	},
									{{+WIDTH,  +WIDTH,  -HALF_LENGTH, 1.0}, COLOR_FLETCH_TOP	},
									{{-WIDTH,  +WIDTH,  -HALF_LENGTH, 1.0}, COLOR_FLETCH_TOP	},
									{{  0.0,   +RADIUS, -HALF_LENGTH, 1.0}, COLOR_FLETCH_TOP	},
									{{  0.0,     0.0,   +HALF_LENGTH, 1.0}, COLOR_FLETCH_TOP	}
								};

static const ObserverIBOData	gFaces[4*4] =
								{
									//	left fletch
									{{ 0,  1,  2}},
									{{ 2,  1,  3}},
									{{ 3,  1,  4}},
									{{ 3,  4,  2}},

									//	right fletch
									{{ 5,  6,  7}},
									{{ 7,  6,  8}},
									{{ 8,  6,  9}},
									{{ 8,  9,  7}},

									//	bottom fletch
									{{10, 11, 12}},
									{{12, 11, 13}},
									{{13, 11, 14}},
									{{13, 14, 12}},

									//	top fletch
									{{15, 16, 17}},
									{{17, 16, 18}},
									{{18, 16, 19}},
									{{18, 19, 17}}
								};


void MakeObserverVBOs(
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName,
	bool			aStereoFlag)
{
	const ObserverVBOData	*theVertices;
	ObserverVBOData			theGreyscaleVertices[BUFFER_LENGTH(gVertices)];
	unsigned int			i;
	float					theLuminance;
	
	if ( ! aStereoFlag )
	{
		//	For normal (non-stereo) viewing,
		//	use gVertices exactly as they are.
		theVertices = gVertices;
	}
	else
	{
		//	For anaglyphic 3D, convert gVertices's colors to greyscale.

		for (i = 0; i < BUFFER_LENGTH(gVertices); i++)
		{
			theGreyscaleVertices[i] = gVertices[i];

			//	The greyscale conversion formula
			//
			//		luminance = 30% red + 59% green + 11% blue
			//
			//	appears widely on the internet, but with little explanation.
			//	Presumably its origins lie in human color perception.

			theLuminance = (float) (
					0.30 * theGreyscaleVertices[i].col[0]
				  + 0.59 * theGreyscaleVertices[i].col[1]
				  + 0.11 * theGreyscaleVertices[i].col[2] );

			theGreyscaleVertices[i].col[0] = theLuminance;
			theGreyscaleVertices[i].col[1] = theLuminance;
			theGreyscaleVertices[i].col[2] = theLuminance;
		}
		
		theVertices = theGreyscaleVertices;
	}

	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);
	glBufferData(	GL_ARRAY_BUFFER,
					sizeof(gVertices),
					theVertices,
					GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, anIndexBufferName);
	glBufferData(	GL_ELEMENT_ARRAY_BUFFER,
					sizeof(gFaces),
					gFaces,
					GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

void BindObserverVBOs(
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName,
	bool			aShadersFlag)
{
	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);

	if (aShadersFlag)
	{
		glEnableVertexAttribArray(ATTRIBUTE_POSITION);
		glVertexAttribPointer(ATTRIBUTE_POSITION,  4, GL_FLOAT, GL_FALSE, sizeof(ObserverVBOData), (void *)offsetof(ObserverVBOData, pos));

		glDisableVertexAttribArray(ATTRIBUTE_TEX_COORD);

		glEnableVertexAttribArray(ATTRIBUTE_COLOR);
		glVertexAttribPointer(ATTRIBUTE_COLOR,     4, GL_FLOAT, GL_FALSE, sizeof(ObserverVBOData), (void *)offsetof(ObserverVBOData, col));
	}
	else
	{
		//	Internet folklore claims that glVertexPointer() should be called last,
		//	after glTexCoordPointer(), glColorPointer(), etc., but my own testing
		//	revealed no difference.

		glVertexPointer(4, GL_FLOAT, sizeof(ObserverVBOData), (void *)offsetof(ObserverVBOData, pos));
		glEnableClientState(GL_VERTEX_ARRAY);

		glDisableClientState(GL_TEXTURE_COORD_ARRAY);

		glColorPointer(	4, GL_FLOAT, sizeof(ObserverVBOData), (void *)offsetof(ObserverVBOData, col));
		glEnableClientState(GL_COLOR_ARRAY);
	}

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, anIndexBufferName);
}

void DrawObserverVBOs(
	ModelData	*md,
	Honeycomb	*aHoneycomb,
	Matrix		*aWorldPlacement,		//	the world's placement in eye space
	Matrix		*anObserverPlacement)	//	the observer's placement in the Dirichlet domain
{
	ImageParity		thePartialParity;
	unsigned int	i;
	Matrix			*theDirichletPlacement;	//	the (translated) Dirichlet domain's placement in world space

	if (aHoneycomb == NULL)
		return;

	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		//	It's simpler to bind a pure white texture for the observer
		//	than it would be to write a special-purpose texture-free shader for it.
		//	Using a flag to toggle texturing on and off would be the worst approach of all:
		//	on my Radeon X1600 it slowed KaleidoTile's frame rate by a factor of four!
		glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureObserver]);
		
		//	Set a pair of texture coordinates once and for all,
		//	to avoid having to pass them in arrays.
		//	(0.5, 0.5) points to the texture's center,
		//	which is as good as spot as any.
		glVertexAttrib2fv(ATTRIBUTE_TEX_COORD, (float [2]){0.5, 0.5});
	}
	else
	{
		glDisable(GL_TEXTURE_2D);
	}
	
	//	Compose the parity of the observer's placement in the Dirichlet domain
	//	with the parity of the world's placement in eye space.  
	//	Later on, as we traverse the tiling, we'll factor in the parity 
	//	of the Dirichlet domain's placement in world space, which may 
	//	be negative for some tiles and positive for others.
	thePartialParity = (aWorldPlacement->itsParity == anObserverPlacement->itsParity
						? ImagePositive : ImageNegative);
	
	//	Draw the images of the observer in near-to-far order.
	//
	//	Skip the i=0 element of the group, which should be the identity.
	//	Normally it's invisible (because we sit inside it looking out),
	//	but in stereoscopic 3D it makes a mess because our eyes sit outside it.
	for (i = 1; i < aHoneycomb->itsNumVisibleCells; i++)
	{
		//	Each element of the tiling group defines
		//	a placement of the Dirichlet domain in world space.
		theDirichletPlacement = &aHoneycomb->itsVisibleCells[i]->itsMatrix;

		//	Let front faces wind counterclockwise (resp. clockwise)
		//	when the observer's placement in eye space preserves (resp. reverses) parity.
		glFrontFace(theDirichletPlacement->itsParity == thePartialParity ? GL_CCW : GL_CW);

		//	Compose anObserverPlacement, theDirichletPlacement and aWorldPlacement,
		//	write the result to itsModelViewMatrix, and send it to the shader.
		Matrix44Product(anObserverPlacement->m, theDirichletPlacement->m, md->itsModelViewMatrix);
		Matrix44Product(md->itsModelViewMatrix, aWorldPlacement->m,       md->itsModelViewMatrix);
		SendTransformationsToShader(md);

		//	Draw.
		glDrawRangeElements(	GL_TRIANGLES,
								0,
								BUFFER_LENGTH(gVertices) - 1,
								3 * BUFFER_LENGTH(gFaces),
								GL_UNSIGNED_SHORT,
								0);
	}

	//	Leave texturing enabled by default.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
	}
	else
	{
		glEnable(GL_TEXTURE_2D);
	}
}
