//	CurvedSpaces-MacOS-Utilities.c
//
//	A few platform-dependent functions that the platform-independent code will need.
//	(Along with a couple global variables whose use I'll need to ponder further.)
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#import <Cocoa/Cocoa.h>
#import "CurvedSpaces-Common.h"
#import "CurvedSpacesMacOSUtilities.h"


NSString *GetLocalizedTextAsNSString(
	const char	*aKey)
{
	return [NSString stringWithUTF8String:GetLocalizedText(aKey)];
}


bool GetBasePath(
	char			*aPath,
	unsigned int	aPathSize)
{
	return [[[[NSBundle mainBundle] resourcePath] stringByAppendingString:@"/"]
		getCString:aPath maxLength:aPathSize encoding:NSUTF8StringEncoding];
}


ErrorText GetFileContents(
	const char		*aPathName,		//	input, zero-terminated UTF-8 string containing absolute path name
	unsigned int	*aNumRawBytes,	//	output, the file size in bytes
	Byte			**someRawBytes)	//	output, the file's contents as raw bytes
{
	//	Technical Note:  I had hoped that setlocale() + fopen() would
	//	let me read the file within the platform-independent part
	//	of the code, but alas on WinXP it works when aPathName 
	//	contains only ASCII characters but fails when aPathName 
	//	contains non-trivial Unicode characters, for example when
	//	I temporarily change "Textures" to "Textures日本".
	//	So I reluctantly resorted to platform-dependent code
	//	to read the file.

	NSData	*theFileContents;
	
	//	Read the file.
	theFileContents = [NSData dataWithContentsOfFile:[NSString stringWithUTF8String:aPathName]];
	if (theFileContents == nil)
		return "Couldn't get file contents as NSData in GetFileContents()";
	
	//	Note the number of bytes.
	*aNumRawBytes = [theFileContents length];
	if (*aNumRawBytes == 0)
		return "File contains 0 bytes in GetFileContents()";

	//	Make our own copy of the data so that the caller will be able to modify it.
	//	For example, for texture data, HalveTextureWidth() and HalveTextureHeight()
	//	will want to modify the texture in place during mipmapping.
	//	The caller will eventually free the buffer with a call to FreeRawTextureBytes().
	*someRawBytes = GET_MEMORY(*aNumRawBytes);
	if (*someRawBytes == NULL)
	{
		*aNumRawBytes = 0;
		return "Couldn't allocate buffer in GetFileContents()";
	}
	[theFileContents getBytes:(void *)(*someRawBytes) length:(*aNumRawBytes)];
	
	//	Success!
	return NULL;
}

void FreeFileContents(
	unsigned int	*aNumRawBytes,
	Byte			**someRawBytes)
{
	*aNumRawBytes = 0;
	FREE_MEMORY_SAFELY(*someRawBytes);
}


bool GetUserPref(
	char	*aKey)
{
	//	CurvedSpacesMacOSView's initWithFrame method should have already
	//	set a default value, so aKey's value should always exist.
	return [[NSUserDefaults standardUserDefaults] boolForKey:[NSString stringWithUTF8String:aKey]];
}

void SetUserPref(
	char	*aKey,
	bool	aValue)
{
	[[NSUserDefaults standardUserDefaults] setBool:aValue forKey:[NSString stringWithUTF8String:aKey]];
}


void ErrorMessage(
	ErrorText	aMessage,	//	UTF8
	ErrorText	aTitle)		//	UTF8
{
	//	NSRunAlertPanel works better than NSBeginAlertSheet
	//	because the former blocks while the latter does not.
	//	This is especially important when displaying a message
	//	during the shutdown sequence (for example the memory-leak message),
	//	when the application would quit out from under any pending
	//	NSBeginAlertSheet() calls.

	NSRunAlertPanel(	[NSString stringWithUTF8String:aTitle  ],
						[NSString stringWithUTF8String:aMessage],
						@"OK",
						NULL,
						NULL);
}
