//	CurvedSpacesMacOSView.m
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#import "CurvedSpacesMacOSView.h"
#import "CurvedSpacesMacOSUtilities.h"

#ifdef SAVE_ANIMATION
#include <QTKit/QTMovie.h>
#include <unistd.h>	//	for usleep()
#endif

//	How many intrinsic units correspond to each pixel of mouse motion?
#define UNITS_PER_PIXEL		0.001

//	When displaying the frame rate in the window's title bar,
//	how often should we update it?
#define FRAME_RATE_UPDATE_PERIOD	1000000		//	time interval in microseconds

//	Virtual key codes for the spacebar and arrow keys.
#define	     RETURN_KEY		 36
#define        SPACEBAR		 49
#define	     ESCAPE_KEY		 53
#define  LEFT_ARROW_KEY		123
#define	RIGHT_ARROW_KEY		124
#define  DOWN_ARROW_KEY		125
#define    UP_ARROW_KEY		126


#ifdef SAVE_ANIMATION
NSFileHandle	*gFlightPathFile	= nil;
#endif


@implementation CurvedSpacesMacOSView


+ (NSOpenGLPixelFormat*)pixelFormatSupportingMultisampling:(bool)aMultisampleFlag
{
	NSOpenGLPixelFormatAttribute	thePlainPixelFormatAttributes[] =
									{
										NSOpenGLPFADoubleBuffer,

										NSOpenGLPFADepthSize,			32,
										NSOpenGLPFAAlphaSize,			8,

										(NSOpenGLPixelFormatAttribute)	0
									},
									theMultisamplePixelFormatAttributes[] =
									{
										NSOpenGLPFADoubleBuffer,

										NSOpenGLPFADepthSize,			32,
										NSOpenGLPFAAlphaSize,			8,

										//	Request multisampling support.
										NSOpenGLPFASampleBuffers,		1,
										NSOpenGLPFASamples,				8,	//	Just guess a large number and let the driver reduce it?

										(NSOpenGLPixelFormatAttribute)	0
									};
	
	return [[[NSOpenGLPixelFormat alloc] initWithAttributes:(aMultisampleFlag ? theMultisamplePixelFormatAttributes : thePlainPixelFormatAttributes)] autorelease];
}

- (id)initWithFrame:(NSRect)frameRect
{
	//	Create a fallback set of user defaults (the "factory settings").
	//	These will get used if and only if the user has not provided his/her own values.
	[[NSUserDefaults standardUserDefaults] registerDefaults:
		[NSDictionary dictionaryWithObjectsAndKeys:
			@"yes", @"multisampling",
			@"yes", @"shaders",
			@"no",	@"fallback message given",
			nil
		]
	];

	//	Set harmless defaults for the ModelData,
	//	without assuming the presence of OpenGL or anything else.
	//	In particular, set all pointers to NULL so that
	//	windowShouldClose() can call ShutDownModelData() safely.
	SetUpModelData(&md);

	//	Initialize the superclass.
	//	Request a multisample buffer iff InitWindowData() set md.itsMultisamplingEnabledFlag to true.
	//	The graphics card may or may not support multisampling, but it's always safe to ask.
	self = [super initWithFrame: frameRect
					pixelFormat: [CurvedSpacesMacOSView pixelFormatSupportingMultisampling: md.itsMultisamplingEnabledFlag ]];
	if (self != nil)
	{
		//	What time are we starting at?
		Microseconds( (UnsignedWide *) &itsLastRenderTime    );
		Microseconds( (UnsignedWide *) &itsLastFrameRateTime );
		
		//	The mouse is still free.
		itsMouseSteeringFlag = false;
		
		//	Allocate and initialize the matrix file directory name.
		itsLastMatrixDirectory	= [[NSMutableString stringWithCapacity:512] retain];
		[itsLastMatrixDirectory setString:[[[NSBundle mainBundle] resourcePath] stringByAppendingString:@"/Sample Spaces/"]];
		
		//	Allocate and initialize the TIFF file directory name.
		itsLastImageDirectory	= [[NSMutableString stringWithCapacity:512] retain];
		[itsLastImageDirectory setString:NSHomeDirectory()];
		
		//	Initialize itsTimer to nil, then...
		itsTimer = nil;
		
		//	...set the swap parameter and the timer.
		//	SetUpModelData() has already set a default value for itsSyncFlag.
		[self setRedrawWithSync:md.itsSyncFlag];
	}

	return self;
}

- (void)dealloc
{
	[itsTimer invalidate];
	[itsTimer release];
	
	[itsLastMatrixDirectory	release];
	[itsLastImageDirectory	release];

	[super dealloc];
}

- (void)awakeFromNib
{
#ifdef SAVE_ANIMATION
	//	Ultimately we'll render offscreen at 1920×1080,
	//	so for now run at half those dimensions to get
	//	the aspect ratio right.
	[[self window] setContentSize:(NSSize){1920/2, 1080/2}];
//	[[self window] setContentSize:(NSSize){ 720,  480}];
	[[self window] center];
	gShowFrameRate = true;
#else
	//	Keep the window contents square when the user resizes it.
	[[self window] setContentAspectRatio:(NSSize){1.0f, 1.0f}];

	//	Zoom to the largest (square!) window that fits on the screen.
	//	MacOS likes to leave a 4-pixel margin along the bottom.
	//	To obtain a 4-pixel margin along the left side as well,
	//	set window origin coordinates (4,4) in Interface Builder.
	[[self window] zoom:self];
#endif
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	NSArray		*thePreferredLanguages;
	NSString	*thePreferredLanguage;
	bool		theShadersAreAvailable;

	//	Uncomment the following line in test versions.
//	ErrorMessage("...", "Pre-release copy for ...");

	//	Initialize the language code and dictionary to the user's default language.
	//
	//	Note #1.  preferredLocalizations determines language availability
	//	based on what localizations it finds in our program's application bundle.
	//	Thus, as a practical matter, localizing InfoPList.strings to a new language
	//	will bring that language to the system's attention.
	//	
	//	Note #2.  One would expect preferredLocalizations to return an NSArray containing
	//	all localizations supported by our mainBundle and acceptable to the user,
	//	listed in order of the user's preferences.  In practice, however,
	//	preferredLocalizations always returns a 1-element array, even when several
	//	acceptable languages are available.  Apple's documentation is evasive
	//	on this question, saying only that preferredLocalizations
	//
	//		Returns one or more localizations contained in the receiver’s bundle 
	//		that the receiver uses to locate resources based on the user’s preferences.
	//
	//	Fortunately preferredLocalizations reliably returns the user's most preferred localization,
	//	from among those supported.
	thePreferredLanguages	= [[NSBundle mainBundle] preferredLocalizations];
	thePreferredLanguage	= ([thePreferredLanguages count] > 0 ?			//	If a preferred language exists...
								[thePreferredLanguages objectAtIndex:0] :	//	...use it.
								@"en");										//	Otherwise fall back to English.
	[self changeLanguage:thePreferredLanguage];

	//	Does the OpenGL context support shaders?
	//	If not, tell the user that the fallback fixed-function code
	//	will be used and the quality of the image will be less.
	//	Display that message once, and then never again (for that user).
	[[self openGLContext] makeCurrentContext];
	theShadersAreAvailable = ShadersAreAvailable();
	[NSOpenGLContext clearCurrentContext];
	if ( ! theShadersAreAvailable
	  && ! GetUserPref("fallback message given"))
	{
		ErrorMessage(GetLocalizedText("ErrorShadersMessage"), GetLocalizedText("ErrorShadersTitle"));
		SetUserPref("fallback message given", true);
	}

	//	The user may have launched Curved Spaces directly,
	//	or may have right-clicked a matrix file and chosen Open With.
	//	In the former case -- but not the latter -- we want to...
	if (md.itsDirichletDomain == NULL)
	{
		//	...encourage the user to open a file.
		//
		//	Note:  We must call commandFileOpen here -- not in initWithFrame
		//	or awakeFromNib -- if we want the main window to get the keyboard focus.
		[self commandFileOpen:nil];
	}
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- (void)setRedrawWithSync:(bool)aSyncFlag
{
#ifdef __MAC_OS_X_VERSION_10_5
	GLint	theSwapValue;
#else
	long	theSwapValue;
#endif

	//	Record the new sync value.
	md.itsSyncFlag = aSyncFlag;

	//	Set the swap interval.
	theSwapValue = (aSyncFlag ? 1 : 0);
	[[self openGLContext] setValues: &theSwapValue forParameter: NSOpenGLCPSwapInterval];

	//	Set a timer for continuous redraw.
	//	The scene may get redrawn less often than the timer interval
	//	(especially if we're synchronizing to the monitor)
	//	but it won't get drawn more often.

	//		Out with the old...
	if (itsTimer != nil)
	{
		[itsTimer invalidate];
		[itsTimer release];
		itsTimer = nil;
	}

	//		...in with the new.
	itsTimer = [NSTimer
#ifdef SAVE_ANIMATION
			//	try for 30fps
			scheduledTimerWithTimeInterval: 1.0/30.0
#else
			scheduledTimerWithTimeInterval: (md.itsSyncFlag ? 0.01 : 0.001)
#endif
			target: self
			selector: @selector( pleaseRedraw )
			userInfo: nil
			repeats: YES];
	[itsTimer retain];
}

- (void)pleaseRedraw
{
	long long	theCurrentTime;
	double		theElapsedTime;

	//	Compute the time since the last redraw.
	Microseconds( (UnsignedWide *) &theCurrentTime );
	theElapsedTime = 0.000001 * (theCurrentTime - itsLastRenderTime);
	itsLastRenderTime = theCurrentTime;
	
	//	If the window is miniaturized, proceed no further.
	if ([[self window] isMiniaturized])
		return;

	//	Need to redraw the image?
	if (SimulationWantsUpdates(&md))
	{
		//	Update the simulation.
#ifdef SAVE_ANIMATION
		//	Pretend 1/30sec elapsed, even if the true elapsed time is more.
		//	(It shouldn't be less, because our timer is set for 1/30sec.)
		SimulationUpdate(&md, 1.0/30.0);
#else
		SimulationUpdate(&md, theElapsedTime);
#endif

		//	Request a redraw.
		[self setNeedsDisplay:YES];
	}
	
	//	If desired, update the frame rate display about once per second.
	if (gShowFrameRate == true
	 && theCurrentTime >= itsLastFrameRateTime + FRAME_RATE_UPDATE_PERIOD)
	{
		[[self window] setTitle:
			(SimulationWantsUpdates(&md) ?
				(theElapsedTime > 0.0001 ?
					[NSString localizedStringWithFormat:
						GetLocalizedTextAsNSString("fps"),
						(unsigned int)(1.0 / theElapsedTime)]
				:
					GetLocalizedTextAsNSString("fps2")
				)
			:
				[NSString localizedStringWithFormat:
					GetLocalizedTextAsNSString("fps"), 0]
			)
		];
		itsLastFrameRateTime = theCurrentTime;
	}

#ifdef SAVE_ANIMATION
	if (gFlightPathFile != nil)
	{
		[gFlightPathFile writeData:[NSData dataWithBytes:&md.itsUserPlacement   length:sizeof(md.itsUserPlacement  ) ]];
		[gFlightPathFile writeData:[NSData dataWithBytes:&md.itsCurrentAperture length:sizeof(md.itsCurrentAperture) ]];
		[gFlightPathFile writeData:[NSData dataWithBytes:&md.itsRotationAngle   length:sizeof(md.itsRotationAngle  ) ]];
	}
#endif
}

- (void)keyDown:(NSEvent *)anEvent
{
	switch ([anEvent keyCode])
	{
#ifdef START_OUTSIDE
		case RETURN_KEY:
			if (md.itsViewpoint == ViewpointExtrinsic)
				md.itsViewpoint = ViewpointEntering;
			break;
#endif

		case SPACEBAR:
			md.itsUserSpeed = 0.0;
			break;
		
		case ESCAPE_KEY:
			[self exitNavigationalMode];	//	for consistency with Torus Games
			break;
		
		case LEFT_ARROW_KEY:
			ChangeAperture(&md, true);
			break;

		case RIGHT_ARROW_KEY:
			ChangeAperture(&md, false);
			break;

		case DOWN_ARROW_KEY:
			md.itsUserSpeed -= USER_SPEED_INCREMENT;
			break;

		case UP_ARROW_KEY:
			md.itsUserSpeed += USER_SPEED_INCREMENT;
			break;

		default:
			//	Process the keystroke normally.
			[self interpretKeyEvents:[NSArray arrayWithObject:anEvent]];
			break;
	}
}

- (void)insertText:(id)aString
{
	NSString	*theNewText;
	unichar		theCharacter;

	//	The documentation promises only that aString will be either
	//	an NSString or an NSAttributedString. In practice it's an NSString
	//	for both the US English server and Hiragana server.
	if ([[aString class] isSubclassOfClass:[NSAttributedString class]])
	{
		theNewText = [aString string];
	}
	else
	if ([[aString class] isSubclassOfClass:[NSString class]])
	{
		theNewText = aString;
	}
	else
	{
		theNewText = @"?";
	}
	
	if ([theNewText length] > 0)
	{
		theCharacter = [theNewText characterAtIndex:0];

		if (theCharacter == 'f')	//	"secret" keyboard command to show frame rate
		{
			gShowFrameRate = ! gShowFrameRate;
			if ( ! gShowFrameRate )
				[[self window] setTitleWithRepresentedFilename:[[self window] representedFilename]];
		}
		else
		if (theCharacter == 's')	//	"secret" keyboard command to sync to the monitor
		{
			[self setRedrawWithSync:( ! md.itsSyncFlag )];
		}
		else
		{
			//	The user is flailing.
			//	Perhaps s/he doesn't know how to exit navigational mode?
			[self exitNavigationalMode];
		}
	}
}

- (BOOL)performKeyEquivalent:(NSEvent *)anEvent
{
	if ([anEvent modifierFlags] & NSCommandKeyMask)
	{
		//	The user just pressed a key combination such as ⌘O, ⌘S, ⌘C, ⌘? or ⌘, .
		//	Exit navigational mode to restore the system cursor.
		[self exitNavigationalMode];
	}
	
	//	Return NO to let the caller handle the key equivalent in the usual way.
	return NO;
}

- (void)mouseDown:(NSEvent *)anEvent
{
	//	For consistency with the Torus Games, we want a double-click
	//	to get us out of navigational mode. Nevertheless, for the convenience
	//	of CurvedSpaces-only users, a single-click should also do the job.
	//	The solution is to exit navigational mode on the first click, 
	//	and then ignore the second click (if any) so that it doesn't put us
	//	back into navigational mode.
	if ([anEvent clickCount] > 1)
		return;

	if (itsMouseSteeringFlag == false)
	{
		[self enterNavigationalMode];
	}
	else
	{
		//	If all Mac users had 2-button mice, they could use
		//	the left button to enter navigational mode and
		//	the right button to exit navigational mode.
		//	Alas Powerbook users are stuck with only one button.
		//	So let them left-click (i.e. single-click)
		//	for entry and exit both.  Indeed in Curved Spaces this
		//	makes good sense.  The problem will arise in the Torus Games,
		//	where the left mouse serves other purposes within the torus
		//	and therefore isn't available for exit.
		//	So in Torus Games the user will have to double-left-click
		//	or press the escape key to exit navigational mode
		//	(or right click if a 2-button mouse is available).

		[self exitNavigationalMode];
	}
}

- (void)rightMouseUp:(NSEvent *)anEvent
{
	//	Technical note:  exitNavigationalMode's call to
	//	[[self window] setAcceptsMouseMovedEvents:NO]
	//	fails when called from within rightMouseDown
	//	but succeeds when called from within rightMouseUp.
	//	Go figure.

	[self exitNavigationalMode];
}

- (void)enterNavigationalMode
{
	//	Hide the cursor and disassociate it from mouse motion
	//	so the (invisible) cursor stays parked over our window.
	//	Mouse motions will nevertheless arrive in mouseDragged calls.

	if (itsMouseSteeringFlag == false)
	{
		CGDisplayHideCursor(kCGDirectMainDisplay);	//	parameter is ignored
		CGAssociateMouseAndMouseCursorPosition(NO);
		[[self window] setAcceptsMouseMovedEvents:YES];
		itsMouseSteeringFlag = true;
	}
}

- (void)exitNavigationalMode
{
	//	Restore the usual cursor.

	if (itsMouseSteeringFlag == true)
	{
		CGDisplayShowCursor(kCGDirectMainDisplay);	//	parameter is ignored
		CGAssociateMouseAndMouseCursorPosition(YES);
		[[self window] setAcceptsMouseMovedEvents:NO];
		itsMouseSteeringFlag = false;
	}
}

- (void)mouseMoved:(NSEvent *)anEvent
{
	double			dx,
					dy;
	unsigned int	theModifierFlags;

	//	In theory we shouldn't need to test itsMouseSteeringFlag because we receive
	//	mouseMoved events only while in navigational mode.  Nevertheless, given
	//	the odd behavior of [[self window] setAcceptsMouseMovedEvents:NO]
	//	succeeding in rightMouseUp but failing in rightMouseDown, it seems prudent
	//	to double check that we're not in steering mode, just so if the user's
	//	runtime environment differs from mine there's no danger of a "stuck steering wheel".
	if (itsMouseSteeringFlag == true)
	{
		dx =   UNITS_PER_PIXEL * [anEvent deltaX];
		dy = - UNITS_PER_PIXEL * [anEvent deltaY];
		
		theModifierFlags = [anEvent modifierFlags];

		MouseMove(	&md,
					dx,
					dy,
					theModifierFlags & NSShiftKeyMask,
					theModifierFlags & NSControlKeyMask,
					theModifierFlags & NSAlternateKeyMask);
	}
}

- (void)scrollWheel:(NSEvent *)anEvent
{
	md.itsUserSpeed += [anEvent deltaY] * 0.1 * USER_SPEED_INCREMENT;
}

- (void)drawRect:(NSRect)bounds
{
	ErrorText	theError;

	if ((theError = SetUpGraphicsAsNeeded(&md)) != NULL)
	{
		ErrorMessage(theError, "OpenGL Setup Error");
		exit(1);
	}

	if ((theError = Render(&md)) != NULL)
	{
		ErrorMessage(theError, "OpenGL Rendering Error");
		exit(2);
	}

	[[self openGLContext] flushBuffer];
}

- (BOOL)validateMenuItem:(NSMenuItem *)aMenuItem
{
	SEL	theAction;
	
	theAction = [aMenuItem action];

#ifdef SAVE_ANIMATION
	if (theAction == @selector(qtStartRecording:))
		return gFlightPathFile == nil;
	if (theAction == @selector(qtStopRecording:))
		return gFlightPathFile != nil;
	if (theAction == @selector(qtPreviewMovie:))
		return gFlightPathFile == nil;
	if (theAction == @selector(qtRenderMovie:))
		return gFlightPathFile == nil;
#endif

	if (theAction == @selector(commandViewCenterpieceNone:))
	{
		[aMenuItem setState:(md.itsSceneryType == SceneryNone		? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewCenterpieceEarth:))
	{
		[aMenuItem setState:(md.itsSceneryType == SceneryEarth		? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewCenterpieceGalaxy:))
	{
		[aMenuItem setState:(md.itsSceneryType == SceneryGalaxy		? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewCenterpieceGyroscope:))
	{
		[aMenuItem setState:(md.itsSceneryType == SceneryGyroscope	? NSOnState : NSOffState)];
		return YES;
	}

	if (theAction == @selector(commandViewObserver:))
	{
		[aMenuItem setState:(md.itsShowObserver						? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewColorCoding:))
	{
		[aMenuItem setState:(md.itsShowColorCoding					? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewCliffordParallels:))
	{
		[aMenuItem setState:(	md.itsSpaceType == SpaceSpherical
							 && md.itsShowCliffordParallels			? NSOnState : NSOffState)];
		return (md.itsSpaceType == SpaceSpherical);
	}
	if (theAction == @selector(commandViewVertexFigures:))
	{
		[aMenuItem setState:(md.itsShowVertexFigures				? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewFog:))
	{
		[aMenuItem setState:(	md.itsShadersAvailableFlag    
							 && md.itsShadersEnabledFlag
							 && md.itsFogFlag							? NSOnState : NSOffState)];
		return (	md.itsShadersAvailableFlag    
				 && md.itsShadersEnabledFlag);
	}

	if (theAction == @selector(commandViewMultisampling:))
	{
		[aMenuItem setState:(	md.itsMultisamplingAvailableFlag
							 && md.itsMultisamplingEnabledFlag		? NSOnState : NSOffState)];
		return md.itsMultisamplingAvailableFlag;
	}
	if (theAction == @selector(commandViewGraphicsFixedFunction:))
	{
		[aMenuItem setState:(! (md.itsShadersAvailableFlag
							 && md.itsShadersEnabledFlag)			? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewGraphicsShaders:))
	{
		[aMenuItem setState:(	md.itsShadersAvailableFlag
							 && md.itsShadersEnabledFlag			? NSOnState : NSOffState)];
		return md.itsShadersAvailableFlag;
	}
	if (theAction == @selector(commandViewStereoscopic3D:))
	{
		[aMenuItem setState:(md.itsStereoFlag						? NSOnState : NSOffState)];
		return YES;
	}

	if (theAction == @selector(commandLanguageDe:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "de") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageEn:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "en") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageFr:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "fr") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageJa:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "ja") ? NSOnState : NSOffState)];
		return YES;
	}

	return YES;
}

- (IBAction)commandFileOpen:(id)sender
{
	NSOpenPanel	*theOpenPanel	= nil;
	
	//	Create a standard choose-a-file panel.
	theOpenPanel = [NSOpenPanel openPanel];
	
	//	Initialize to the most recently used directory.
	[theOpenPanel setDirectory:itsLastMatrixDirectory];

	//	Ask the user to choose a file.
	//	If s/he complies, open the file.
	if ([theOpenPanel runModalForTypes:(NSArray *)[NSArray arrayWithObject:@"gen"]] == NSOKButton)
		[self application:nil openFile:[theOpenPanel filename]];
	
}

- (IBAction)commandFileExportCopy:(id)sender
{
	NSData			*theTiff				= nil;
	NSPasteboard	*theGeneralPasteboard	= nil;

	theTiff = [self contentsAsTIFF];

	theGeneralPasteboard = [NSPasteboard generalPasteboard];
	
	[theGeneralPasteboard declareTypes: [NSArray arrayWithObject:NSTIFFPboardType] owner: nil];
	[theGeneralPasteboard setData: theTiff forType:NSTIFFPboardType];
}

- (IBAction)commandFileExportSave:(id)sender
{
	NSSavePanel	*theSavePanel	= nil;
	NSData		*theTiff		= nil;

	//	Create a save-a-file panel.
	theSavePanel = [NSSavePanel savePanel];
	[theSavePanel setAllowedFileTypes:[NSArray arrayWithObject:@"tiff"]];
	[theSavePanel setTitle:@"Save image as TIFF file"];

	//	Ask the user to choose a file name.
	if ([theSavePanel runModalForDirectory:itsLastImageDirectory file:
			[[[[[self window] representedFilename] lastPathComponent] stringByDeletingPathExtension] stringByAppendingPathExtension:@"tiff"]
		] == NSFileHandlingPanelCancelButton)
		return;	//	user cancelled
	
	//	Remember the directory for next time.
	[itsLastImageDirectory setString:[theSavePanel directory]];

	//	Read the OpenGL image.
	theTiff = [self contentsAsTIFF];
	if (theTiff == nil)
		return;
	
	//	Save the data under the requested filename.
	if ([theTiff writeToFile:[theSavePanel filename] atomically:YES] == NO)
		ErrorMessage("Couldn't save image under requested filename.", "Save failed");
}

- (IBAction)commandViewCenterpieceNone:(id)sender
{
	SetSceneryType(&md, SceneryNone);
}

- (IBAction)commandViewCenterpieceEarth:(id)sender
{
	SetSceneryType(&md, SceneryEarth);
}

- (IBAction)commandViewCenterpieceGalaxy:(id)sender
{
	SetSceneryType(&md, SceneryGalaxy);
}

- (IBAction)commandViewCenterpieceGyroscope:(id)sender
{
	SetSceneryType(&md, SceneryGyroscope);
}

- (IBAction)commandViewObserver:(id)sender
{
	SetShowObserver(&md, ! md.itsShowObserver );
}

- (IBAction)commandViewColorCoding:(id)sender
{
	SetShowColorCoding(&md, ! md.itsShowColorCoding );
}

- (IBAction)commandViewCliffordParallels:(id)sender
{
	SetShowCliffordParallels(&md, ! md.itsShowCliffordParallels );
}

- (IBAction)commandViewVertexFigures:(id)sender
{
	SetShowVertexFigures(&md, ! md.itsShowVertexFigures );
}

//	The following three View menu items are synonyms for Preference Panel options,
//	so as well as changing our main settings, be sure to update the Preference Panel.
//	Under normal circumstances this wouldn't be necessary, but if the user happens
//	to have the Preference Panel open, we want to keep it valid.

- (IBAction)commandViewFog:(id)sender
{
	SetFogFlag(&md, ! md.itsFogFlag );
	[itsPrefFog setState:(md.itsFogFlag ? NSOnState : NSOffState)];
}

- (IBAction)commandViewMultisampling:(id)sender
{
	SetMultisamplingEnabledFlag(&md, ! md.itsMultisamplingEnabledFlag );
	[itsPrefMultisampling setState:(md.itsMultisamplingEnabledFlag ? NSOnState : NSOffState)];
	[self replaceOpenGLContext];
}

- (IBAction)commandViewGraphicsFixedFunction:(id)sender
{
	SetShadersEnabledFlag(&md, false);
	[itsPrefShaders setState:NSOffState];
}

- (IBAction)commandViewGraphicsShaders:(id)sender
{
	SetShadersEnabledFlag(&md, true);
	[itsPrefShaders setState:NSOnState];
}

- (IBAction)commandViewStereoscopic3D:(id)sender
{
	SetStereo3DFlag(&md, ! md.itsStereoFlag );
	[itsPrefStereo3D setState:(md.itsStereoFlag ? NSOnState : NSOffState)];
	[self setStereoLabels];
}

- (IBAction)commandLanguageDe:(id)sender
{
	[self changeLanguage:@"de"];
}

- (IBAction)commandLanguageEn:(id)sender
{
	[self changeLanguage:@"en"];
}

- (IBAction)commandLanguageFr:(id)sender
{
	[self changeLanguage:@"fr"];
}

- (IBAction)commandLanguageJa:(id)sender
{
	[self changeLanguage:@"ja"];
}

- (IBAction)commandHelpHelp:(id)sender
{
	//	The Mac OS X Help Book system is terrible:
	//	it's very difficult to get it accept different pages,
	//	won't accept multiple named pages in the same directory,
	//	tends to fail or crash at the least excuse,
	//	and caches so agressively that you can't get
	//	old versions of a Help Book to go away.
	//
	//	By contrast, sending a page to the default web browser
	//	is simple and effective.  A real breath of fresh air!
	[[NSWorkspace sharedWorkspace] openFile:[[[NSBundle mainBundle] resourcePath]
		stringByAppendingString:[NSString stringWithFormat:@"/Help/CurvedSpacesWelcome-%c%c.html", gLanguageCode[0], gLanguageCode[1]]]];
}


- (BOOL)application:(NSApplication *)theApplication openFile:(NSString *)filename
{
	ErrorText		theError		= NULL;
	unsigned int	theFileSize		= 0;
	NSData			*theRawData		= nil;
	Byte			*theRawBytes	= NULL;

	//	Remember the directory for next time.
	[itsLastMatrixDirectory setString:[filename stringByDeletingLastPathComponent]];
		
	//	Read the file's raw bytes.
	theRawData = [NSData dataWithContentsOfFile:filename];
	
	//	Append a terminating zero to the raw data.
	theFileSize = [theRawData length];
	theRawBytes = (Byte *) malloc(theFileSize + 1);	//	allow room for a terminating zero
	if (theRawBytes == NULL)
		goto CleanUpOpenFile;
	[theRawData getBytes:theRawBytes length:theFileSize];
	theRawBytes[theFileSize] = 0;	//	terminating zero

	//	Try to parse the file.
	theError = LoadGenerators(&md, theRawBytes);
	if (theError != NULL)
		goto CleanUpOpenFile;

	//	Tell the window what file it represents.
	[[self window] setTitleWithRepresentedFilename:filename];

	//	Add the file to the recent document list.
	//
	//	28 July 2008.  Tatsu Takeuchi pointed out that the Open Recent menu no longer works.
	//	Somehow the new paradigm of loading the menu bar and translating it in real-time
	//	confuses the AppKit's handling of the Open Recent menu.
	//	I like the new paradigm, because it simplifies maintenance by isolating 
	//	all language-dependent text (in a single text file, one file per language) 
	//	from the language-independent user-interface design and connections 
	//	(in a single set of NIB files).  So I decided to ditch the Open Recent menu.
	//	If I ever figure out how to get it to work, I can uncomment the following line of code.
//	[[NSDocumentController sharedDocumentController] noteNewRecentDocumentURL:[NSURL fileURLWithPath:filename]];

CleanUpOpenFile:

	if (theRawBytes != NULL)
		free(theRawBytes);
	
	if (theError != NULL)
	{
		ErrorMessage(theError, "Error reading matrix file");
		return NO;
	}
	else
		return YES;
}

- (IBAction)setUpPreferencesPanel:(id)sender
{
	//	Transfer our private data to the Preference Panel's fields.
	
	//	Perspective Tab
	
	[itsPrefWindowWidth		setDoubleValue: 2.0 * md.itsWindowHalfWidth];
	[itsPrefWindowDistance	setDoubleValue:       md.itsWindowDistance ];

	[itsPrefStereo3D		setState:(md.itsStereoFlag					? NSOnState : NSOffState)];
	[itsPrefEyeSeparation	setDoubleValue: 2.0 * md.itsEyeOffset      ];
	[itsPrefIntrinsicUnit	setDoubleValue:       md.itsIntrinsicUnit  ];
	[self setStereoLabels];

	//	Advanced Tab
	
	[itsPrefFog				setEnabled:(	md.itsShadersAvailableFlag    
										 && md.itsShadersEnabledFlag)];
	[itsPrefFog				setState:(		md.itsShadersAvailableFlag    
										 && md.itsShadersEnabledFlag
										 && md.itsFogFlag				? NSOnState : NSOffState)];


	[itsPrefFrameRate		setState:(gShowFrameRate					? NSOnState : NSOffState)];

	[itsPrefSyncToMonitor	setState:(md.itsSyncFlag					? NSOnState : NSOffState)];

	[itsPrefMultisampling	setEnabled: md.itsMultisamplingAvailableFlag ];
	[itsPrefMultisampling	setState:(	md.itsMultisamplingAvailableFlag
									 && md.itsMultisamplingEnabledFlag	? NSOnState : NSOffState)];

	[itsPrefAnisotropic		setEnabled: md.itsAnisotropicAvailableFlag ];
	[itsPrefAnisotropic		setState:(	md.itsAnisotropicAvailableFlag
									 && md.itsAnisotropicEnabledFlag	? NSOnState : NSOffState)];

	[itsPrefShaders			setEnabled: md.itsShadersAvailableFlag ];
	[itsPrefShaders			setState:(	md.itsShadersAvailableFlag
									 && md.itsShadersEnabledFlag		? NSOnState : NSOffState)];
	
	//	Display the Preference Panel as a sheet.
	[NSApp beginSheet:itsPreferencesPanel modalForWindow:[self window] modalDelegate:nil didEndSelector:nil contextInfo:nil];
}

- (void) setStereoLabels
{
	//	Cocoa provides no way to enable/disable static text (!).
	//	See http://developer.apple.com/qa/qa2006/qa1461.html for discussion.
	//	So instead we must grey the static text ourselves.

	NSColor	*theStaticTextColor;
	BOOL	theEnabledFlag;
	
	if (md.itsStereoFlag)
	{
		theStaticTextColor	= [NSColor controlTextColor];
		theEnabledFlag		= YES;
	}
	else
	{
		theStaticTextColor	= [NSColor disabledControlTextColor];
		theEnabledFlag		= NO;
	}
	
	[itsPrefEyeSeparation		setEnabled:theEnabledFlag];
	[itsPrefEyeSeparationLabel	setTextColor:theStaticTextColor];
	[itsPrefEyeSeparationUnits	setTextColor:theStaticTextColor];
	
	[itsPrefIntrinsicUnit		setEnabled:theEnabledFlag];
	[itsPrefIntrinsicUnitLabel	setTextColor:theStaticTextColor];
	[itsPrefIntrinsicUnitUnits	setTextColor:theStaticTextColor];
}

- (IBAction)readOnePreference:(id)sender
{
	//	Perspective Tab

	if (sender == itsPrefWindowWidth)
	{
		md.itsWindowHalfWidth	= 0.5 * [itsPrefWindowWidth		doubleValue];
		md.itsWindowHalfHeight	= md.itsWindowHalfWidth;
	}

	if (sender == itsPrefWindowDistance)
		md.itsWindowDistance	=       [itsPrefWindowDistance	doubleValue];

	if (sender == itsPrefStereo3D)
	{
		SetStereo3DFlag(&md, [itsPrefStereo3D state] == NSOnState);
		[self setStereoLabels];
	}

	if (sender == itsPrefEyeSeparation)
		md.itsEyeOffset			= 0.5 * [itsPrefEyeSeparation	doubleValue];

	if (sender == itsPrefIntrinsicUnit)
		md.itsIntrinsicUnit		=       [itsPrefIntrinsicUnit	doubleValue];


	//	Advanced Tab

	if (sender == itsPrefFog)
		SetFogFlag(&md, [itsPrefFog state] == NSOnState);

	if (sender == itsPrefFrameRate)
	{
		gShowFrameRate = ([itsPrefFrameRate state] == NSOnState);
		if (gShowFrameRate == false)
			[[self window] setTitleWithRepresentedFilename:[[self window] representedFilename]];
	}

	if (sender == itsPrefSyncToMonitor)
		[self setRedrawWithSync:([itsPrefSyncToMonitor state] == NSOnState)];

	if (sender == itsPrefMultisampling)
	{
		SetMultisamplingEnabledFlag(&md, [itsPrefMultisampling state] == NSOnState);
		[self replaceOpenGLContext];
	}

	if (sender == itsPrefAnisotropic)
		SetAnisotropicEnabledFlag(&md, [itsPrefAnisotropic state] == NSOnState);

	if (sender == itsPrefShaders)
	{
		SetShadersEnabledFlag(&md, [itsPrefShaders state] == NSOnState);

		//	Fog is available only when shaders are enabled.
		[itsPrefFog	setEnabled:(	md.itsShadersAvailableFlag    
								 && md.itsShadersEnabledFlag)];
		[itsPrefFog	setState:(		md.itsShadersAvailableFlag    
								 && md.itsShadersEnabledFlag
								 && md.itsFogFlag				? NSOnState : NSOffState)];
	}
	

	//	OK Button
	//		shutDownPreferencesPanel wants us to test for modified text fields.
	if (sender == self)
	{
		md.itsWindowHalfWidth	= 0.5 * [itsPrefWindowWidth		doubleValue];
		md.itsWindowDistance	=       [itsPrefWindowDistance	doubleValue];
		md.itsEyeOffset			= 0.5 * [itsPrefEyeSeparation	doubleValue];
		md.itsIntrinsicUnit		=       [itsPrefIntrinsicUnit	doubleValue];

		md.itsWindowHalfHeight	= md.itsWindowHalfWidth;
	}


	//	Request a redraw.
	md.itsRedrawRequestFlag = true;
}

- (void)replaceOpenGLContext
{
	NSOpenGLPixelFormat	*thePixelFormat;
	NSOpenGLContext		*theNewContext;
	
	//	Shut down the old OpenGL context.
	//	The old OpenGL context is about to be vaporized, 
	//	but calling ShutDownGraphicsAsNeeded() gives
	//	the program a chance to clear its references 
	//	to shader programs, textures and vertex buffer objects.
	[[self openGLContext] makeCurrentContext];
	ShutDownGraphicsAsNeeded(&md);
	[NSOpenGLContext clearCurrentContext];

	//	Get a new OpenGL context with the desired pixel format.
	thePixelFormat	= [CurvedSpacesMacOSView pixelFormatSupportingMultisampling:md.itsMultisamplingEnabledFlag];
	theNewContext	= [[NSOpenGLContext alloc] initWithFormat:thePixelFormat shareContext:nil];
	[self setOpenGLContext:theNewContext];
	[theNewContext setView:self];
	
	//	Tell the new context about itsSyncFlag.
	[self setRedrawWithSync:md.itsSyncFlag];
	
	//	Reset everything.
	md.itsPreparedExtensions = false;
}

- (IBAction)shutDownPreferencesPanel:(id)sender
{
	//	Stop the Preference Panel's modal session and remove it from the window list,
	//	but don't destroy it.
	[NSApp endSheet:itsPreferencesPanel];
	[itsPreferencesPanel orderOut:self];

	//	The checkboxes and radio buttons all produce immediate results,
	//	and so do the text fields if the user hits tab or return.
	//	But if the user changes a text field and then hits OK,
	//	we have to explicitly ask that the text field be read.
	[self readOnePreference:self];
}


- (NSImage *)contentsAsImage
{
	NSBitmapImageRep	*theBitmap	= nil;
	NSImage				*theImage	= nil;
	
	//	Get a bitmap...
	theBitmap = [self contentsAsBitmapImageRep];
	
	//	...and wrap it as an NSImage.
	//
	//	Note:  The documentation for addRepresentation makes the vague comment that
	//	"After invoking this method, you may need to explicitly set features
	//	of the new image representation, such as the size, number of colors, and so on."
	//	but in practice it seems to work OK.
	theImage = [[[NSImage alloc] initWithSize:(NSSize){0.0, 0.0}] autorelease];
	[theImage addRepresentation:theBitmap];

	return theImage;
}


- (NSData *)contentsAsTIFF
{
	NSBitmapImageRep	*theBitmap	= nil;
	NSData				*theTIFF	= nil;
	
	//	Get a bitmap...
	theBitmap = [self contentsAsBitmapImageRep];

	//	...and convert it to TIFF format.
	theTIFF = [theBitmap TIFFRepresentationUsingCompression: NSTIFFCompressionLZW factor: 0.0f];
	if (theTIFF == nil)
		ErrorMessage("Couldn't create TIFF representation.", "Image Copying Error");

	return theTIFF;
}


- (NSBitmapImageRep *)contentsAsBitmapImageRep
{
	ErrorText			theError	= NULL;
	NSRect				theBounds	= {{0.0f, 0.0f}, {0.0f, 0.0f}};
	NSBitmapImageRep	*theBitmap	= nil;
	Byte				*thePixels	= NULL;
	
	//	Note the view's bounds.
	theBounds = [self bounds];
	if (NSMinX(theBounds) != 0 || NSMinY(theBounds) != 0)
	{
		theError = "View does not have origin (0,0).";
		goto CleanUpContentsAsBitmapImageRep;
	}

	//	Create an empty bitmap of the correct dimensions.
	//	If we're running Mac OS 10.4, pass NSAlphaNonpremultipliedBitmapFormat.
	if ([NSBitmapImageRep instancesRespondToSelector:@selector(initWithBitmapDataPlanes:pixelsWide:pixelsHigh:bitsPerSample:samplesPerPixel:hasAlpha:isPlanar:colorSpaceName:bitmapFormat:bytesPerRow:bitsPerPixel:)])
	{
		//	Preferred call for 10.4
		theBitmap = [[[NSBitmapImageRep alloc]
			initWithBitmapDataPlanes:	NULL
			pixelsWide:					NSWidth(theBounds)
			pixelsHigh:					NSHeight(theBounds)
			bitsPerSample:				8
			samplesPerPixel:			4
			hasAlpha:					YES
			isPlanar:					NO
			colorSpaceName:				NSDeviceRGBColorSpace
			bitmapFormat:				NSAlphaNonpremultipliedBitmapFormat
			bytesPerRow:				NSWidth(theBounds) * 4
			bitsPerPixel:				32
			] autorelease];
	}
	else
	{
		//	Fallback for 10.3.9
		theBitmap = [[[NSBitmapImageRep alloc]
			initWithBitmapDataPlanes:	NULL
			pixelsWide:					NSWidth(theBounds)
			pixelsHigh:					NSHeight(theBounds)
			bitsPerSample:				8
			samplesPerPixel:			4
			hasAlpha:					YES
			isPlanar:					NO
			colorSpaceName:				NSDeviceRGBColorSpace
			bytesPerRow:				NSWidth(theBounds) * 4
			bitsPerPixel:				32
			] autorelease];
	}
	if (theBitmap == nil)
	{
		theError = "Couldn't allocate theBitmap.";
		goto CleanUpContentsAsBitmapImageRep;
	}

	//	Read the pixels into theBitmap.
	thePixels = [theBitmap bitmapData];
	[self lockFocus];
		theError = ReadRGBAPixels(NSWidth(theBounds), NSHeight(theBounds), thePixels, false);
	[self unlockFocus];
	if (theError != NULL)
		goto CleanUpContentsAsBitmapImageRep;
	
	//	OpenGL reads the image bottom-to-top, but Cocoa seems to be expecting
	//	it top-to-bottom.  So we must reverse the rows.
	[self invertRawImage: (unsigned int *)thePixels pixelsWide:NSWidth(theBounds) pixelsHigh:NSHeight(theBounds)];
	
CleanUpContentsAsBitmapImageRep:

	if (theError != NULL)
		ErrorMessage(theError, "Error Copying OpenGL Image from Window");

	return theBitmap;
}


- (void)invertRawImage: (unsigned int *)aPixelBuffer pixelsWide:(unsigned int)aWidth pixelsHigh:(unsigned int)aHeight
{
	unsigned int	theLowerRow,
					theUpperRow,
					theCount;
	
	//	Rather than thinking of a pixel as four 1-byte components {R,G,B,A},
	//	think of it as a single 4-byte integer.  The byte order,
	//	0xRRGGBBAA or 0xAABBGGRR, makes no difference here.
	unsigned int	*theLowerPixel,
					*theUpperPixel,
					theTempPixel;
	
	theLowerRow	= 0;
	theUpperRow	= aHeight - 1;

	while (theUpperRow > theLowerRow)
	{
		theLowerPixel = aPixelBuffer + theLowerRow*aWidth;
		theUpperPixel = aPixelBuffer + theUpperRow*aWidth;

		for (theCount = aWidth; theCount-- > 0; )
		{
			theTempPixel	= *theLowerPixel;
			*theLowerPixel	= *theUpperPixel;
			*theUpperPixel	= theTempPixel;
			theLowerPixel++;
			theUpperPixel++;
		}

		theLowerRow++;
		theUpperRow--;
	}
}


- (void)changeLanguage:(NSString *)aLanguageCode
{
	//	Record the new language code and initialize a dictionary.
	RefreshLocalization([aLanguageCode UTF8String]);

	//	Reconstruct the menu bar in the new language.
	[NSApp setMainMenu:[self buildLocalizedMenuBar]];

	//	If the window title is displaying neither a filename nor a frame rate,
	//	then update the default title ("Curved Spaces") to the new language.
	//	(Note:  For Curved Spaces this presently makes no difference, 
	//	because its title is always "Curved Spaces", even in Japanese.)
	if ([[[self window] representedFilename] length] == 0
	 && gShowFrameRate == false)
	{
		[[self window] setTitle:GetLocalizedTextAsNSString("WindowTitle")];
	}
	
	//	Update the text in the Preferences Panel,
	//	whether or not it's currently visible.
	[self setPreferencePanelLabels];
}

- (NSMenu *)buildLocalizedMenuBar
{
	NSMenu		*theMenuBar,
				*theMenu,
				*theSubmenu;
	NSMenuItem	*theItemWithKeyModifier;

	//	Construct all menus using the current language.

	//	Design Note:  A more traditional approach would be to maintain 
	//	a separate .nib file for each language, but it's a pain
	//	to maintain numerous parallel .nib files as the software evolves --
	//	one must either maintain them by hand or take on the added complexity
	//	of dealing with AppleGlot to automate the process.
	//	Personally I prefer to keep a program's functionality separated 
	//	from its language-specific strings.  Thus the menus (including all
	//	selectors and key equivalents) get defined here here in this
	//	language-independent function, while all language-specific strings 
	//	(including the text of each menu item) get stored in the language files.
	//	This not only makes program maintenance easier (because a change 
	//	to the menu bar happens in one place only) but also simplifies localization
	//	(because the translator works with a simple text-only file containing 
	//	the English phrases alongside their local language equivalents).
	
	//	Main menu bar
	theMenuBar = [[[NSMenu alloc] initWithTitle:@"main menu"] autorelease];
	
	//	Application menu
	//	(On Mac OS 10.6, title will be overridden by localized CFBundleName.
	//	But on 10.4 or 10.5, the title given here may persist.)
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Curved Spaces"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("About Curved Spaces")
			action:@selector(orderFrontStandardAboutPanel:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Preferences…")
			action:@selector(setUpPreferencesPanel:) keyEquivalent:@","];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Hide Curved Spaces")
			action:@selector(hide:) keyEquivalent:@"h"];
		theItemWithKeyModifier = [[NSMenuItem alloc]
					initWithTitle:GetLocalizedTextAsNSString("Hide Others")
					action:@selector(hideOtherApplications:) keyEquivalent:@"h"];
			[theItemWithKeyModifier setKeyEquivalentModifierMask:(NSCommandKeyMask | NSAlternateKeyMask)];
			[theMenu addItem:theItemWithKeyModifier];
			[theItemWithKeyModifier release];
			theItemWithKeyModifier = nil;
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Show All")
			action:@selector(unhideAllApplications:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Quit Curved Spaces")
			action:@selector(terminate:) keyEquivalent:@"q"];
	theMenu = nil;

	//	File menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"File"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Open…")
			action:@selector(commandFileOpen:) keyEquivalent:@"o"];
		[theMenu addItem:[NSMenuItem separatorItem]];
		theSubmenu = [self addToMenu:theMenu submenuWithTitle:"Export"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Copy Image")
				action:@selector(commandFileExportCopy:) keyEquivalent:@"c"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Save Image…")
				action:@selector(commandFileExportSave:) keyEquivalent:@"s"];
		theSubmenu = nil;
	theMenu = nil;

	//	View menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"View"];
		theSubmenu = [self addToMenu:theMenu submenuWithTitle:"Centerpiece"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("None")
				action:@selector(commandViewCenterpieceNone:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Earth")
				action:@selector(commandViewCenterpieceEarth:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Galaxy")
				action:@selector(commandViewCenterpieceGalaxy:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Gyroscope")
				action:@selector(commandViewCenterpieceGyroscope:) keyEquivalent:@""];
		theSubmenu = nil;
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Spaceship")
			action:@selector(commandViewObserver:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Color Coding")
			action:@selector(commandViewColorCoding:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Clifford Parallels")
			action:@selector(commandViewCliffordParallels:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Vertex Figures")
			action:@selector(commandViewVertexFigures:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Fog")
			action:@selector(commandViewFog:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Smooth Edges")
			action:@selector(commandViewMultisampling:) keyEquivalent:@""];
		theSubmenu = [self addToMenu:theMenu submenuWithTitle:"Graphics Pipeline"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Fixed-Function")
				action:@selector(commandViewGraphicsFixedFunction:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Shaders")
				action:@selector(commandViewGraphicsShaders:) keyEquivalent:@""];
		theSubmenu = nil;
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Stereoscopic 3D")
			action:@selector(commandViewStereoscopic3D:) keyEquivalent:@""];
	theMenu = nil;

	//	Language menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Language"];
		//	To add a new language, please see the instructions
		//	in the file "To add a new language.txt".
//		[theMenu addItemWithTitle:@"Deutsch"
//			action:@selector(commandLanguageDe:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"English"
			action:@selector(commandLanguageEn:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Français"
			action:@selector(commandLanguageFr:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"日本語"
			action:@selector(commandLanguageJa:) keyEquivalent:@""];
	theMenu = nil;
	
	//	Help menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Help"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Curved Spaces Help")
			action:@selector(commandHelpHelp:) keyEquivalent:@"?"];
	theMenu = nil;

#ifdef SAVE_ANIMATION
	//	QuickTime menu (for custom builds only)
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"QuickTime"];
		[theMenu addItemWithTitle:"Start Recording…"
			action:@selector(qtStartRecording:) keyEquivalent:@"r"];
		[theMenu addItemWithTitle:@"Stop Recording"
			action:@selector(qtStopRecording: ) keyEquivalent:@"t"];
		[theMenu addItemWithTitle:@"Preview Movie…"
			action:@selector(qtPreviewMovie:  ) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Render Movie…"
			action:@selector(qtRenderMovie:   ) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:@" 720 ×  480"
			action:@selector(qtSize720x480:   ) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"1920 × 1080"
			action:@selector(qtSize1920x1080: ) keyEquivalent:@""];
		[theMenu addItemWithTitle:@" 960 ×  540"
			action:@selector(qtSize960x540:   ) keyEquivalent:@""];
	theMenu = nil;
#endif
	
	return theMenuBar;
}

- (NSMenu *)addToMenu:(NSMenu *)aParentMenu submenuWithTitle:(char *)aSubmenuTitle
{
	NSString	*theLocalizedTitle;
	NSMenuItem	*theNewParentMenuItem;
	NSMenu		*theNewMenu;

	//	Technical Note:  For top-level menus (i.e. those attached to the menu bar),
	//	theNewMenu's title gets displayed while theNewParentMenuItem's title gets ignored.
	//	For submenus, the reverse is true.  To be safe, pass the same title to both.

	//	Localize aSubmenuTitle to the current language.
	theLocalizedTitle = GetLocalizedTextAsNSString(aSubmenuTitle);

	//	Create a new item for aParentMenu, and a new menu to attach to it.
	theNewParentMenuItem	= [[NSMenuItem alloc] initWithTitle:theLocalizedTitle action:NULL keyEquivalent:@""];
	theNewMenu				= [[NSMenu alloc] initWithTitle:theLocalizedTitle];
	
	//	Attach theNewMenu to theNewParentMenuItem.
	[theNewParentMenuItem setSubmenu:theNewMenu];
	
	//	Add theNewParentMenuItem to aParentMenu.
	[aParentMenu addItem:theNewParentMenuItem];

	//	aParentMenu retains theNewParentMenuItem,
	//	so we can release theNewParentMenuItem.
	[theNewParentMenuItem release];
	
	//	aParentMenu retains theNewParentMenuItem,
	//	and theNewParentMenuItem retains theNewMenu,
	//	so we can release theNewMenu.
	[theNewMenu release];
	
	//	Pass the caller a weak reference to theNewMenu.
	return theNewMenu;
}

- (void)setPreferencePanelLabels
{
	[[itsPrefTabView tabViewItemAtIndex:0] setLabel:
		GetLocalizedTextAsNSString("PrefPerspective")];

	[[itsPrefTabView tabViewItemAtIndex:1] setLabel:
		GetLocalizedTextAsNSString("PrefAdvanced")];

	[itsPrefGeneralBox setTitle:
		GetLocalizedTextAsNSString("PrefGeneral")];

	[itsPrefWindowWidthLabel setStringValue:
		GetLocalizedTextAsNSString("PrefWindowWidth")];

	[itsPrefWindowDistanceLabel setStringValue:
		GetLocalizedTextAsNSString("PrefWindowDistance")];

	[itsPrefStereo3DBox setTitle:
		GetLocalizedTextAsNSString("PrefStereo3D")];

	[itsPrefStereo3D setTitle:
		GetLocalizedTextAsNSString("PrefEnableStereo3D")];

	[itsPrefEyeSeparationLabel setStringValue:
		GetLocalizedTextAsNSString("PrefEyeSpacing")];

	[itsPrefIntrinsicUnitLabel setStringValue:
		GetLocalizedTextAsNSString("PrefIntrinsicUnit")];

	[itsPrefFog setTitle:
		GetLocalizedTextAsNSString("PrefFog")];

	[itsPrefFrameRate setTitle:
		GetLocalizedTextAsNSString("PrefFrameRate")];

	[itsPrefSyncToMonitor setTitle:
		GetLocalizedTextAsNSString("PrefSync")];

	[itsPrefMultisampling setTitle:
		GetLocalizedTextAsNSString("PrefMultisampling")];

	[itsPrefAnisotropic setTitle:
		GetLocalizedTextAsNSString("PrefAnisotropic")];

	[itsPrefShaders setTitle:
		GetLocalizedTextAsNSString("PrefShaders")];

	[itsPrefOK setTitle:
		GetLocalizedTextAsNSString("PrefOK")];
}


- (void)reshape
{
	ResizeGLWindow(&md, (unsigned int) [self bounds].size.width, (unsigned int) [self bounds].size.height);

	//	Refresh the window.  You need to call update first to have it 
	//	come to terms with its context.  If not bizzare things happen.
	//	Adam found this solution at:
	//		http://developer.apple.com/qa/qa2001/qa1209.html
	//
	[self update];
	[self setNeedsDisplay:YES];
}

- (void)windowWillMiniaturize:(NSNotification *)aNotification
{
	//	When the user minimizes a window, Mac OS X creates a dock icon to represent it.
	//	Normally the icon is a low-resolution image of the window's contents,
	//	which Mac OS X creates automatically with no help from the programmer.
	//	Alas that automated process fails for OpenGL windows, so let's read
	//	the OpenGL image buffer ourselves and report its contents via setMiniwindowImage.
	[[self window] setMiniwindowImage:[self contentsAsImage]];
}

- (BOOL)windowShouldClose:(id)sender
{
	//	Caution:  Cocoa calls this method only when the user explictly clicks
	//	the window's close button.  When the user chooses Quit Curved Spaces
	//	we must call this method ourselves.  In the former case this method
	//	will get called twice, but that's OK.

	//	Free the ModelData here, rather than in dealloc(),
	//	so applicationWillTerminate() can reliably test for memory leaks.
	ShutDownModelData(&md);
	
	return YES;
}

- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication *)theApplication
{
	return YES;
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
	//	If the user chooses Quit from the menu,
	//	Cocoa will *not* ask windows whether they should close or not,
	//	so we must do it ourselves.
	if ([self windowShouldClose:self] == YES)
		return NSTerminateNow;
	else
		return NSTerminateCancel;
}

- (void)applicationWillTerminate:(NSNotification *)aNotification
{
	//	Free memory used for localized dictionary.
	RefreshLocalization("--");

	//	We should be clean.
	if (gMemCount != 0)
		ErrorMessage("Memory allocated does not equal memory freed.\r\nPlease report this error to\r\n\twww.geometrygames.org/contact.html", "Memory Leak");
}

#ifdef SAVE_ANIMATION
- (IBAction)qtStartRecording:(id)sender
{
	NSSavePanel	*theSavePanel;
	NSString	*theFilePath;

	[self exitNavigationalMode];
	
	if (gFlightPathFile != nil)
	{
		ErrorMessage("Flight path file unexpectedly busy in qtStartRecording.", "Internal Error");
		return;
	}
	
	theSavePanel = [NSSavePanel savePanel];
	[theSavePanel setNameFieldLabel:@"Save Flight As"];
	[theSavePanel setDirectory:[[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent]];
	if ([theSavePanel runModal] == NSFileHandlingPanelOKButton)
	{
		theFilePath = [theSavePanel filename];
		[[NSFileManager defaultManager] createFileAtPath:theFilePath contents:nil attributes:nil];
		gFlightPathFile = [[NSFileHandle fileHandleForWritingAtPath:theFilePath] retain];
		if (gFlightPathFile == nil)
		{
			ErrorMessage("Couldn't open output file in qtStartRecording.", "I/O Error");
			return;
		}
	}
}

- (IBAction)qtStopRecording:(id)sender
{
	[gFlightPathFile closeFile];
	[gFlightPathFile release];
	gFlightPathFile = nil;
}

- (IBAction)qtPreviewMovie:(id)sender
{
	[self qtRenderMovie:nil];
}

- (IBAction)qtRenderMovie:(id)sender
{
	bool				theSavedMultisamplingFlag;
	NSOpenPanel			*theOpenPanel			= nil;
	NSString			*theFilePath			= nil;
	NSFileHandle		*theFlightPathFile		= nil;
	NSSavePanel			*theSavePanel			= nil;
	NSString			*theMovieFilePath		= nil;
	NSAutoreleasePool	*theAutoreleasePool		= nil;
	QTMovie				*theMovie				= nil;
	NSData				*theUserPlacement		= nil,
						*theCurrentAperature	= nil,
						*theRotationAngle		= nil;
	double				thePreviousAperture		= 0.0;
	ErrorText			theSetUpError			= NULL,
						theRenderError			= NULL;
	long long			thePreviousTime,
						theCurrentTime;

	//	No matter what the user's current setting,
	//	we want to render with multisampling enabled (if it's available).
	theSavedMultisamplingFlag		= md.itsMultisamplingEnabledFlag;
	md.itsMultisamplingEnabledFlag	= true;
	[self replaceOpenGLContext];

	//	For some reason setTitle: gets ignored if we call it after the Open dialog.
	//	(Maybe because we've got exclusive control at that point???)
	[[self window] setTitle:
		sender != nil ?
		@"Rendering QuickTime Movie.  Please be patient…" :
		@"Dry run only:  no movie is being created."];

	//	Let the user select an input file.
	theOpenPanel = [NSOpenPanel openPanel];
	[theOpenPanel setMessage:@"Select an input file defining the flight path."];
	[theOpenPanel setDirectory:[[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent]];
	if ([theOpenPanel runModal] != NSFileHandlingPanelOKButton)
		return;
	
	//	Open the input file.
	theFilePath = [theOpenPanel filename];
	theFlightPathFile = [NSFileHandle fileHandleForReadingAtPath:theFilePath];
	if (theFlightPathFile == nil)
	{
		ErrorMessage("Couldn't open input file in qtRenderMovie.", "I/O Error");
		goto CleanUpQtRenderMovie;
	}
	
	//	Set up an output movie file.
	if (sender != nil)
	{
		theSavePanel = [NSSavePanel savePanel];
		[theSavePanel setMessage:@"Select an output file for the QuickTime movie."];
		[theSavePanel setRequiredFileType:@"mov"];
		[theSavePanel setDirectory:[[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent]];
		if ([theSavePanel runModal] != NSFileHandlingPanelOKButton)
			goto CleanUpQtRenderMovie;
		theMovieFilePath = [theSavePanel filename];

		//	A QTMovie wants to initialize from a pre-existing file, it seems.
		//	So make sure the output file exists (even if it contains an empty movie).
		//	Note:  This will overwrite any previous version of the output file.
		[[QTMovie movie] writeToFile:theMovieFilePath withAttributes:[NSDictionary dictionary]];

		//	Now that a (possibly empty) file exists, allocate and initialize the real movie.
		theMovie = [[[QTMovie alloc] initWithFile:theMovieFilePath error:nil] autorelease];
		[theMovie setAttribute:[NSNumber numberWithBool:YES] forKey:QTMovieEditableAttribute];
	}

	
	[[self openGLContext] makeCurrentContext];
	
	Microseconds((UnsignedWide *) &thePreviousTime);

	while (true)
	{
		//	Create a local autorelease pool.
		//	Each frame will require several megabytes of data,
		//	which we'll want to release before moving on to the next frame.
		theAutoreleasePool = [[NSAutoreleasePool alloc] init];

		theUserPlacement	= [theFlightPathFile readDataOfLength:sizeof(md.itsUserPlacement  )];
		theCurrentAperature	= [theFlightPathFile readDataOfLength:sizeof(md.itsCurrentAperture)];
		theRotationAngle	= [theFlightPathFile readDataOfLength:sizeof(md.itsRotationAngle  )];

		if ([theUserPlacement    length] != sizeof(md.itsUserPlacement  )
		 || [theCurrentAperature length] != sizeof(md.itsCurrentAperture)
		 || [theRotationAngle    length] != sizeof(md.itsRotationAngle  ))
			break;
		
		thePreviousAperture = md.itsCurrentAperture;
	
		[theUserPlacement    getBytes:&md.itsUserPlacement   length:sizeof(md.itsUserPlacement  )];
		[theCurrentAperature getBytes:&md.itsCurrentAperture length:sizeof(md.itsCurrentAperture)];
		[theRotationAngle    getBytes:&md.itsRotationAngle   length:sizeof(md.itsRotationAngle  )];
		
		//	Did the aperture change?
		if (md.itsCurrentAperture != thePreviousAperture)
		{
			//	Set itsDesiredAperture, just to keep things clean.
			md.itsDesiredAperture = md.itsCurrentAperture;

			//	We'll need a new VBO with the new aperture.
			md.itsPreparedVBOs = false;
		}

		theSetUpError	= SetUpGraphicsAsNeeded(&md);
		theRenderError	= Render(&md);
		[[self openGLContext] flushBuffer];
		if (theSetUpError != NULL)
		{
			ErrorMessage(theSetUpError, "Non-fatal OpenGL Setup Error");
		}
		if (theRenderError != NULL)
		{
			ErrorMessage(theRenderError, "OpenGL Rendering Error");
			goto CleanUpQtRenderMovie;	//	quit now to avoid an infinite stream of error messages
		}
		
		if (sender != nil)
		{
			[theMovie
				addImage:[self contentsAsImage]
				forDuration:QTMakeTime((long long)1, (long)30)
				withAttributes:[NSDictionary dictionaryWithObject:@"tiff" forKey:QTAddImageCodecType]
			];
			[theMovie updateMovieFile];
		}

		//	While recording a movie, we'll almost surely lag behind 30fps.
		//	But while previewing a movie, we'd probably exceed 30fps 
		//	if we didn't sleep between frames.
		while (true)
		{
			Microseconds((UnsignedWide *) &theCurrentTime);
			if ((unsigned int)(theCurrentTime - thePreviousTime) > 1000000/30)
			{
				break;
				thePreviousTime = theCurrentTime;
			}
			usleep(1000);
		}

		//	Free the local autorelease pool, which will free 
		//	the various copies of the frame that we just created,
		//	along with other locally created objects.
		[theAutoreleasePool release];
	}

CleanUpQtRenderMovie:

	[NSOpenGLContext clearCurrentContext];

	[theFlightPathFile closeFile];

	[[self window] setTitle:GetLocalizedTextAsNSString("WindowTitle")];
	
	md.itsMultisamplingEnabledFlag = theSavedMultisamplingFlag;
	[self replaceOpenGLContext];
}

- (IBAction)qtSize720x480:(id)sender
{
	//	standard definition
	[[self window] setContentSize:(NSSize){720, 480}];
}

- (IBAction)qtSize1920x1080:(id)sender
{
	//	high definition
	[[self window] setContentSize:(NSSize){1920, 1080}];
}

- (IBAction)qtSize960x540:(id)sender
{
	//	half of high definition,
	//	to get the aspect ratio right in a smaller window
	[[self window] setContentSize:(NSSize){1920/2, 1080/2}];
}
#endif

@end
