//	CurvedSpaces-Win32-Utilities.c
//
//	Provides platform-dependent supporting functions
//	for the Win32 version of Curved Spaces.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#include "CurvedSpaces-Win32.h"
#include <string.h>		//	for wcslen() and wcscpy()


static bool	ConvertEndOfLineMarkers(WCHAR *aTextBuffer, unsigned int aTextBufferLength);


#define CONVERSION_ERROR_STRING		L"<error converting text>"

WCHAR *GetLocalizedTextAsWCHAR(
	const char		*aKey,				//	zero-terminated UTF-8 string
	WCHAR			*aValueBuffer,		//	zero-terminated WCHAR string
	unsigned int	aValueBufferLength)	//	length of aValueBuffer in WCHARs (not bytes),
										//		including space for the terminating zero
{
	if (aValueBufferLength == 0)
	{
		ErrorMessage("GetLocalizedTextAsWCHAR() received zero-length output buffer.", "Internal Error");
		ExitProcess(4);
	}

	if (MultiByteToWideChar(
			CP_UTF8, MB_ERR_INVALID_CHARS,
			GetLocalizedText(aKey), -1,
			aValueBuffer, aValueBufferLength) == 0
	 || ConvertEndOfLineMarkers(aValueBuffer, aValueBufferLength) == false)
	{
		if (aValueBufferLength >= wcslen(CONVERSION_ERROR_STRING) + 1)
			wcscpy(aValueBuffer, CONVERSION_ERROR_STRING);
		else
		if (aValueBufferLength >= 2)
			wcscpy(aValueBuffer, L"?");
		else
			aValueBuffer[0] = 0;
	}

	return aValueBuffer;
}

static bool ConvertEndOfLineMarkers(
	WCHAR			*aTextBuffer,		//	zero-terminated WCHAR string
	unsigned int	aTextBufferLength)	//	length of aTextBuffer in WCHARs (not bytes),
										//		including space for the terminating zero;
										//		typically greater than the space occupied
										//		by the string itself
{
	//	Convert "\n" to "\r\n" if aValueBuffer is big enough to hold it.

	WCHAR	*r,	//	read  position
			*w;	//	write position

	//	Set r to the terminating zero.
	//	Set w a little bit further on, 
	//	to allow room to insert '\r' characters.
	r = w = aTextBuffer;
	while (*r != 0)
	{
		if (*r == L'\n')
			w++;
		r++;
		w++;
	}
	
	//	If we've got room to insert the '\r' characters...
	if (w < aTextBuffer + aTextBufferLength)
	{
		//	then read the string backwards...
		do
		{
			//	copying all characters...
			*w = *r;

			//	and inserting '\r' characters as necessary.
			if (*r == L'\n'
			 && --w >= aTextBuffer)	//	error check is logically unnecessary
			{
				*w = L'\r';
			}

		} while (--r >= aTextBuffer && --w >= aTextBuffer);
			
		return true;
	}
	else
		return false;
}


bool GetBasePath(
	char			*aPath,			//	output buffer, zero-terminated UTF-8 string
	unsigned int	aPathSize)		//	output buffer size, including space for terminating zero
{
	bool			theReturnValue	= false;
	WCHAR			*thePath		= NULL;	//	zero-terminated UTF-16 string
	unsigned int	theTerminator,
					i;

	//	Allocate a buffer for the UTF-16 version of the path.
	//	One UTF-16 character is guaranteed to be enough
	//	to handle any string that will ultimately fit
	//	into a UTF-8 path of aPathSize bytes.
	thePath = (WCHAR *) GET_MEMORY(aPathSize * sizeof(WCHAR));
	if (thePath == NULL)
		goto CleanUpGetBasePath;

	//	Get an absolute path for our executable file.
	//	GetModuleFileName() will return the length of the string,
	//	which we interpret as the position of the terminating zero.
	//	If the true path is too long for the buffer, presumably
	//	GetModuleFileName() allows room for the terminating zero,
	//	so assume a path that just barely fits has been truncated.
	//	If GetModuleFileName() fails for any other reason,
	//	it should return 0.
	theTerminator = GetModuleFileName(NULL, thePath, aPathSize);
	if (theTerminator == 0
	 || theTerminator >= aPathSize - 1)
	{
		goto CleanUpGetBasePath;
	}

	//	Remove the name of executable file itself,
	//	leaving only the path to the directory,
	//	including the trailing backslash.
	while (theTerminator > 0 && thePath[theTerminator - 1] != '\\')
		thePath[--theTerminator] = 0;
	
	//	Convert the path separator from '\' to '/'.
	for (i = 0; i < theTerminator; i++)
		if (thePath[i] == '\\')
			thePath[i] = '/';

	//	Convert to UTF-8.
	if (WideCharToMultiByte(CP_UTF8, 0, thePath, -1, aPath, aPathSize, NULL, NULL) == 0)
		goto CleanUpGetBasePath;

	//	Success!
	theReturnValue = true;

CleanUpGetBasePath:
	
	FREE_MEMORY_SAFELY(thePath);
	
	if (theReturnValue == false)
		aPath[0] = 0;
	
	return theReturnValue;
}


ErrorText GetFileContents(
	const char		*aPathName,		//	absolute path as zero-terminated UTF-8 string
	unsigned int	*aNumRawBytes,	//	output, size of file's contents
	Byte			**someRawBytes)	//	output, copy of file's contents
{
	//	Technical Note:  I had hoped that setlocale() + fopen() would
	//	let me read the file within the platform-independent part
	//	of the code, but alas on WinXP it works when aPathName 
	//	contains only ASCII characters but fails when aPathName 
	//	contains non-trivial Unicode characters, for example when
	//	I temporarily change "Textures" to "Textures日本".
	//	So I reluctantly resorted to platform-dependent code
	//	to read the file.

	ErrorText		theErrorMessage	= NULL;
	unsigned int	thePathLength	= 0;
	WCHAR			*thePathName	= NULL;
	HANDLE			theFile			= INVALID_HANDLE_VALUE;
	unsigned long	theNumBytesRead	= 0;

	//	Convert aPathName from UTF-8 to UTF-16.
	//	One UTF-16 character per UTF-8 character
	//	is guaranteed to suffice (even if the UTF-16 text
	//	includes surrogate pairs).
	thePathLength = 0;
	while (aPathName[thePathLength++] != 0)	//	counts the terminating zero as well
		;
	thePathName = (WCHAR *) GET_MEMORY(thePathLength * sizeof(WCHAR));
	if (thePathName == NULL)
	{
		theErrorMessage = "Couldn't get memory for UTF-16 path name in GetFileContents().";
		goto CleanUpGetFileContents;
	}
	if (MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aPathName, -1, thePathName, thePathLength) == 0)
	{
		theErrorMessage = "Couldn't convert path name from UTF-8 to UTF-16 in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Open the file.
	theFile = CreateFile(thePathName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (theFile == INVALID_HANDLE_VALUE)
	{
		theErrorMessage = "Couldn't open file in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Get the file size.
	*aNumRawBytes = GetFileSize(theFile, NULL);
	if (*aNumRawBytes == INVALID_FILE_SIZE)
	{
		theErrorMessage = "Couldn't get file size in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Allocate a buffer to read the file contents into.
	*someRawBytes = (Byte *) GET_MEMORY(*aNumRawBytes);
	if (*someRawBytes == NULL)
	{
		theErrorMessage = "Couldn't allocate memory for file's contents in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Read the file contents into the buffer.
	if (ReadFile(theFile, *someRawBytes, *aNumRawBytes, &theNumBytesRead, NULL) == 0
	 || theNumBytesRead != *aNumRawBytes)
	{
		theErrorMessage = "Couldn't read file's contents in GetFileContents().";
		goto CleanUpGetFileContents;
	}

CleanUpGetFileContents:
	
	if (theFile != INVALID_HANDLE_VALUE)
		CloseHandle(theFile);

	FREE_MEMORY_SAFELY(thePathName);
	
	if (theErrorMessage != NULL)
	{
		*aNumRawBytes = 0;
		FREE_MEMORY_SAFELY(*someRawBytes);
	}

	return theErrorMessage;
}

void FreeFileContents(
	unsigned int	*aNumRawBytes,
	Byte			**someRawBytes)
{
	*aNumRawBytes = 0;
	FREE_MEMORY_SAFELY(*someRawBytes);
}


bool GetUserPref(
	char	*aKey)	//	ASCII
{
	HKEY	theRegistryKey;
	LONG	theQueryError;
	DWORD	theValue,
			theNumBytes;

	//	Try to read a previously saved user pref
	//	from the Windows registry.
	if (RegOpenKeyEx(HKEY_CURRENT_USER, TEXT("Software\\Geometry Games\\Curved Spaces"), 0, KEY_READ, &theRegistryKey) == ERROR_SUCCESS)
	{
		theNumBytes = sizeof(DWORD);
		theQueryError = RegQueryValueExA(theRegistryKey, aKey, NULL, NULL, (BYTE *)&theValue, &theNumBytes);
		RegCloseKey(theRegistryKey);
		if (theQueryError == ERROR_SUCCESS && theNumBytes == sizeof(DWORD))
			return (bool)theValue;
	}

	//	No previously saved user pref was found.
	//	Return a reasonable default value.
	if (strcmp(aKey, "multisampling"			) == 0)
		return true;
	if (strcmp(aKey, "shaders"					) == 0)
		return true;
	if (strcmp(aKey, "fallback message given"	) == 0)
		return false;
	return false;
}

void SetUserPref(
	char	*aKey,	//	ASCII
	bool	aValue)
{
	HKEY	theRegistryKey;
	DWORD	theValue;

	if (RegCreateKeyEx(HKEY_CURRENT_USER, TEXT("Software\\Geometry Games\\Curved Spaces"),
						0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &theRegistryKey, NULL)
		== ERROR_SUCCESS)
	{
		theValue = (DWORD)aValue;
		RegSetValueExA(theRegistryKey, aKey, 0, REG_DWORD, (BYTE *)&theValue, sizeof(DWORD));
		RegCloseKey(theRegistryKey);
	}
}


void ErrorMessage(
	ErrorText	aMessage,	//	UTF-8
	ErrorText	aTitle)		//	UTF-8
{
	//	Translate the incoming messages from UTF-8 to UTF-16
	//	and display them with MessageBoxW().

	bool			theMessageWasGiven	= false;
	unsigned int	theMessageLength	= 0,	//	number of unicode characters including terminating zero
					theTitleLength		= 0;
	WCHAR			*theMessageBuffer	= NULL,	//	buffer for UTF-16 translation including terminating zero
					*theTitleBuffer		= NULL;
	DWORD			theTextDirection;

	theMessageLength	= MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aMessage, -1, NULL, 0);
	if (theMessageLength == 0)		//	MultiByteToWideChar() failed, possibly because of bad input
		goto CleanUpErrorMessage;
	theMessageBuffer	= GET_MEMORY(theMessageLength * sizeof(WCHAR));
	if (theMessageBuffer == NULL)	//	deep trouble
		goto CleanUpErrorMessage;
	MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aMessage, -1, theMessageBuffer, theMessageLength);

	theTitleLength		= MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aTitle, -1, NULL, 0);
	if (theTitleLength == 0)		//	MultiByteToWideChar() failed, possibly because of bad input
		goto CleanUpErrorMessage;
	theTitleBuffer		= GET_MEMORY(theTitleLength * sizeof(WCHAR));
	if (theTitleBuffer == NULL)		//	deep trouble
		goto CleanUpErrorMessage;
	MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aTitle, -1, theTitleBuffer, theTitleLength);

	theTextDirection = LanguageReadsRightToLeft() ? (MB_RTLREADING | MB_RIGHT) : 0;

	MessageBoxW(NULL, theMessageBuffer, theTitleBuffer, MB_OK | MB_TASKMODAL | theTextDirection);

	theMessageWasGiven = true;

CleanUpErrorMessage:

	FREE_MEMORY_SAFELY(theMessageBuffer);
	FREE_MEMORY_SAFELY(theTitleBuffer  );

	//	If something went wrong...
	if (theMessageWasGiven == false)
	{
		//	report the present failure as well as...
		MessageBoxA(NULL,
					"Original error message to follow...",
					"Error in ErrorMessage()",
					MB_OK | MB_TASKMODAL | MB_ICONERROR);

		//	the original error message.  The UTF-8 will be interpreted
		//	relative to the local code page, but at least any 7-bit characters
		//	should display OK.
		MessageBoxA(NULL,
					aMessage,
					aTitle,
					MB_OK | MB_TASKMODAL | MB_ICONERROR);
	}
}


bool LanguageReadsRightToLeft(void)
{
	if (SAME_LANGUAGE_CODE(gLanguageCode, "ar")		//	Arabic
	 || SAME_LANGUAGE_CODE(gLanguageCode, "fa"))	//	Persian
		return true;
	else
		return false;
}
