//	CurvedSpaces-Win32-WinMain.c
//
//	Provides WinMain() and supporting functions
//	for the Win32 version of Curved Spaces.
//
//	This code requires the import libraries
//
//		kernel32.lib  user32.lib  gdi32.lib  opengl32.lib
//		comdlg32.lib  comctl32.lib
//
//	as well as  vfw32.lib  if the AVI video feature is used.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#include "CurvedSpaces-Win32.h"
#include "CurvedSpacesIDs.h"
#include <shlobj.h>	//	for SHGetFolderPath()
#include <stdio.h>	//	for swprintf()

//	How many pixels of empty screen space should
//	we allow around the default window position?
#define WINDOW_MARGIN	16

//	By how much should we stagger successive windows?
#define WINDOW_STAGGER	32


//	Package up information for the idle-time routine.
typedef struct
{
	double	itsFramePeriod;			//	Inverse of frame rate
	bool	itsKeepGoingFlag,		//	Is at least one window present?
			itsAnimationFlag;		//	Is at least one animation active?
} IdleTimeData;


//	An "atom" uniquely identifies the main window class.
ATOM		gMainWindowClass	= 0;

//	The user's preferred directory for saving images may differ
//	from his/her preferred directory for reading manifolds.
WCHAR		gLastImageDirectory[MAX_PATH]		= {0},
			gLastManifoldDirectory[MAX_PATH]	= {0};


static void				TestUnicodeOS(void);
static void				TestSupportingFiles(WORD aPrimaryLangID);
static void				InitLocalization(WORD aPrimaryLangID);
static void				InitDefaultDirectories(void);
static bool				RegisterWindowClasses(void);
static double			MeasureFramePeriod(void);
static BOOL CALLBACK	DoIdleTime(HWND aWindow, LPARAM anIdleTimeDataPtr);
static void				DisplayFrameRate(IdleTimeData *someIdleTimeData);


int WINAPI WinMain(
	HINSTANCE	anInstance,
	HINSTANCE	aPreviousInstance,
	LPSTR		aCommandLine,
	int			aCommandShow)
{
	MSG				theMessage;
	IdleTimeData	theIdleTimeData;

	UNUSED_PARAMETER(anInstance);
	UNUSED_PARAMETER(aPreviousInstance);
	UNUSED_PARAMETER(aCommandLine);
	UNUSED_PARAMETER(aCommandShow);

	//	Test for a Unicode-capable operating system.
	TestUnicodeOS();
	
	//	Make sure this executable hasn't gotten
	//	separated from its supporting files.
	TestSupportingFiles(PRIMARYLANGID(GetUserDefaultLangID()));

	//	Load a dictionary corresponding to the user's preferred language.
	InitLocalization(PRIMARYLANGID(GetUserDefaultLangID()));
	
	//	Initialize default directories
	//	for manifolds and for saved images.
	InitDefaultDirectories();

	//	Set up the window classes.
	if (RegisterWindowClasses() == false)
		goto CleanUpAndExit;

	//	Create a window.
	if (CreateCurvedSpacesWindow() == NULL)
		goto CleanUpAndExit;

	//	Process messages.
	do
	{
		//	Fetch and dispatch all available messages.
		while (PeekMessage(&theMessage, NULL, 0, 0, PM_REMOVE) != 0)
		{
			TranslateMessage(&theMessage);
			DispatchMessage(&theMessage);
		}

		//	No more messages are available,
		//	so draw the next frame in each active animation.
		//	Note whether windows are present and whether animations are active.
		theIdleTimeData.itsFramePeriod		= MeasureFramePeriod();
		theIdleTimeData.itsKeepGoingFlag	= false;
		theIdleTimeData.itsAnimationFlag	= false;
		EnumThreadWindows(GetCurrentThreadId(), DoIdleTime, (LPARAM) &theIdleTimeData);

		//	Display the frame rate if animations are active,
		//	or "0 fps" otherwise.
		DisplayFrameRate(&theIdleTimeData);

		//	If no animations are active, sleep to avoid hogging CPU cycles.
		if (theIdleTimeData.itsAnimationFlag == false)
			Sleep(10);

		//	If no windows remain open at all, exit the message loop.
	} while (theIdleTimeData.itsKeepGoingFlag == true);

CleanUpAndExit:

	//	There's not much to clean up.
	//	Even UnregisterClass() isn't really necessary, because
	//	the classes would get unregistered automatically anyhow.
	(void) UnregisterClass(MAIN_WINDOW_CLASS_NAME,    GetModuleHandle(NULL));
	(void) UnregisterClass(DRAWING_WINDOW_CLASS_NAME, GetModuleHandle(NULL));

	//	Free language-specific resources.
	RefreshLocalization("--");

	//	Test for memory leaks.
	if (gMemCount != 0)
		ErrorMessage("Memory allocated does not equal memory freed.\r\nPlease report this error to\r\n\twww.geometrygames.org/contact.html", "Memory Leak");

	return 0;
}


static void TestUnicodeOS(void)
{
	//	We're compiling for Unicode, so at runtime we must test whether
	//	our program really is running on a Unicode-based system
	//	such as Windows NT, 2000, XP or Vista.

	if (GetVersion() & 0x80000000)	//	Windows 95/98/ME
	{
		//	Display the error message in English.
		//	With no Unicode available, a localized error message
		//	could fail spectacularly (for example, in Japanese).
		MessageBoxA(NULL, "This software requires Windows NT/2000/XP/Vista or later.", "Windows 95/98/Me Not Supported", MB_OK);

		ExitProcess(3);
	}
}


static void TestSupportingFiles(
	WORD	aPrimaryLangID)
{
	DWORD	theMessageTitleID			= 0xFFFFFFFF,
			theMessageTextID			= 0xFFFFFFFF;
	WCHAR	theMessageTitleBuffer[256]	= L"",
			theMessageTextBuffer[4096]	= L"";

	//	Test for the existence of an arbitrarily chosen file.
	//	Presumably all the supporting files will be present,
	//	or none of them will be.
	if (GetFileAttributes(L"Languages/CurvedSpaces-en.txt") == INVALID_FILE_ATTRIBUTES)
	{
		//	If the test file isn't present, display a message,
		//	if possible in the user's preferred language.
		switch (aPrimaryLangID)
		{
			case LANG_FRENCH:
				theMessageTitleID	= IDS_MISSING_FILES_TITLE_FR;
				theMessageTextID	= IDS_MISSING_FILES_MESSAGE_FR;
				break;

			case LANG_ITALIAN:
				theMessageTitleID	= IDS_MISSING_FILES_TITLE_IT;
				theMessageTextID	= IDS_MISSING_FILES_MESSAGE_IT;
				break;

			case LANG_JAPANESE:
				theMessageTitleID	= IDS_MISSING_FILES_TITLE_JA;
				theMessageTextID	= IDS_MISSING_FILES_MESSAGE_JA;
				break;

			case LANG_PORTUGUESE:
				theMessageTitleID	= IDS_MISSING_FILES_TITLE_PT;
				theMessageTextID	= IDS_MISSING_FILES_MESSAGE_PT;
				break;

			case LANG_ENGLISH:
			default:
				theMessageTitleID	= IDS_MISSING_FILES_TITLE_EN;
				theMessageTextID	= IDS_MISSING_FILES_MESSAGE_EN;
				break;
		}
		
		LoadString(GetModuleHandle(NULL), theMessageTitleID, theMessageTitleBuffer, BUFFER_LENGTH(theMessageTitleBuffer));
		LoadString(GetModuleHandle(NULL), theMessageTextID,  theMessageTextBuffer,  BUFFER_LENGTH(theMessageTextBuffer ));

		MessageBox(NULL, theMessageTextBuffer, theMessageTitleBuffer, MB_OK | MB_TASKMODAL);
	}
}


static void InitLocalization(
	WORD	aPrimaryLangID)
{
	switch (aPrimaryLangID)
	{
//		case LANG_GERMAN:	RefreshLocalization("de");	break;
		case LANG_ENGLISH:	RefreshLocalization("en");	break;
		case LANG_FRENCH:	RefreshLocalization("fr");	break;
		case LANG_JAPANESE:	RefreshLocalization("ja");	break;

		default:
			ErrorMessage("Sorry, but this software does not yet support your preferred language.  It will launch in English instead.  You may, if you wish, choose an alternative from the Language menu.\n\nIf you wish to translate this software into your native tongue, please contact Jeff Weeks at www.geometrygames.org/contact.html .", "Language not supported");
			RefreshLocalization("en");
			break;
	}
}


static void InitDefaultDirectories(void)
{
	char	thePathAsUTF8[2048];
	WCHAR	*theChar;

	if (GetAbsolutePath("Sample Spaces", NULL, thePathAsUTF8, BUFFER_LENGTH(thePathAsUTF8)) != NULL
	 || MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, thePathAsUTF8, -1, gLastManifoldDirectory, BUFFER_LENGTH(gLastManifoldDirectory)) == 0)
	{
		gLastManifoldDirectory[0] = 0;
	}
	//	In theory Win32 accepts both forward slashes / and backslashes \ as
	//	path separators (see http://en.wikipedia.org/wiki/Path_(computing) ).
	//	In practice CreateFile() happily accepts either.
	//	However, GetOpenFileName() wants its OPENFILENAME structure's
	//	lpstrInitialDir field to use a backslash, and ignores it
	//	if it uses forward slashes.  So we must convert.
	for (theChar = gLastManifoldDirectory; *theChar != 0; theChar++)
	{
		if (*theChar == L'/')
			*theChar = L'\\';
	}

	if (BUFFER_LENGTH(gLastImageDirectory) == MAX_PATH)	//	SHGetFolderPath() requires a buffer of length MAX_PATH.
		SHGetFolderPath(NULL, CSIDL_PERSONAL, NULL, 0 /* = SHGFP_TYPE_CURRENT */, gLastImageDirectory);
}


static bool RegisterWindowClasses(void)
{
	WNDCLASS	wc;
	ATOM		theDrawingWindowClass;

	wc.style			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= MainWndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance 		= GetModuleHandle(NULL);
	wc.hIcon			= LoadIcon(GetModuleHandle(NULL), (void *) IDI_ICON);
	wc.hCursor			= NULL;
	wc.hbrBackground	= NULL;
	wc.lpszMenuName		= NULL;	//	language-specific menu will be added later
	wc.lpszClassName 	= MAIN_WINDOW_CLASS_NAME;

	gMainWindowClass = RegisterClass(&wc);

	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
	wc.lpfnWndProc		= DrawingWndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= GetModuleHandle(NULL);
	wc.hIcon			= NULL;
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground	= NULL;
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= DRAWING_WINDOW_CLASS_NAME;

	theDrawingWindowClass = RegisterClass(&wc);

	return (gMainWindowClass != 0
		 && theDrawingWindowClass != 0);
}


HWND CreateCurvedSpacesWindow(void)
{
	DWORD		theWindowStyles;
	RECT		theWindowRect,
				theScreenRect;
	signed int	theHorizontalAdjustment,
				theVerticalAdjustment,
				theClientWidth,
				theClientHeight;
	HMENU		theDummyMenu;
	HWND		theWindow;
	WindowData	*wd;

	static signed int	theHorizontalOffset	= 0,
						theVerticalOffset	= 0;

	//	If you plan to make use of many windows,
	//	you may want to make them a smaller, fixed size.
	//	But because I anticipate working mainly with a single
	//	large window, the present code creates the largest
	//	square window that fits comfortably on the screen.

	//	Choose window styles.
	theWindowStyles = WS_OVERLAPPEDWINDOW | WS_VISIBLE;

	//	How big an allowance must we make for the window frame,
	//	title bar and menu?  Start with dummy values for a 256×256
	//	window and let AdjustWindowRect() tell us how much it will change.
	SetRect(&theWindowRect, 0, 0, 256, 256);
	AdjustWindowRect(&theWindowRect, theWindowStyles, true);
	theHorizontalAdjustment	= (theWindowRect.right - theWindowRect.left) - 256;
	theVerticalAdjustment	= (theWindowRect.bottom - theWindowRect.top) - 256;

	//	How big is the visible portion of the display (excluding taskbar etc.) ?
	if (SystemParametersInfo(SPI_GETWORKAREA, 0, &theScreenRect, 0) == 0)
		//	fallback in case of failure...
		SetRect(&theScreenRect, 0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN));

	//	Figure out the largest size window
	//	that will fit comfortably on the screen.
	theClientWidth  = (theScreenRect.right - theScreenRect.left)
					- theHorizontalAdjustment
					- 2 * WINDOW_MARGIN;
	theClientHeight = (theScreenRect.bottom - theScreenRect.top)
					- theVerticalAdjustment
					- 2 * WINDOW_MARGIN;

	//	Make the client area square if you want.
	if (theClientWidth > theClientHeight)
		theClientWidth = theClientHeight;
	if (theClientHeight > theClientWidth)
		theClientHeight = theClientWidth;

#ifdef ENABLE_AVI_MOVIE
	//	theClientWidth and theClientHeight must
	//	be multiples of 4 if you want to make AVI movies.
	theClientWidth	-= theClientWidth%4;
	theClientHeight	-= theClientHeight%4;
#endif

	//	Create a temporary menu bar, which will
	//	1.	occupy space at the top of the window, thus
	//		facilitating correct layout of the client area, and
	//	2.	give RefreshLanguage() an "old menu" to delete,
	//		thus avoiding risk of spurious errors.
	theDummyMenu = CreateMenu();
	AppendMenu(theDummyMenu, MF_STRING, 0, L"temp");

	//	Create the window.
	theWindow = CreateWindow(
		MAIN_WINDOW_CLASS_NAME,
		L"Temporary Title",
		theWindowStyles,
		WINDOW_MARGIN + theHorizontalOffset,
		WINDOW_MARGIN + theVerticalOffset,
		theClientWidth	+ theHorizontalAdjustment,
		theClientHeight	+ theVerticalAdjustment,
		NULL, theDummyMenu, GetModuleHandle(NULL), NULL);

	//	Give up if we didn't get a window.
	if (theWindow == NULL)
		return NULL;
	
	//	Replace the temporary title and menu bar with real ones,
	//	and adjust the mirroring if necessary.
	RefreshLanguage(theWindow, 0);

	//	An empty Curved Spaces window is useless,
	//	so prompt the user to open a file.
	if ((wd = (WindowData *) GetWindowLongPtr(theWindow, GWLP_USERDATA)) != NULL)
		DoFileOpen(wd);

	//	Stagger the window for next time.
	theHorizontalOffset += WINDOW_STAGGER;
	if (theHorizontalOffset > GetSystemMetrics(SM_CXSCREEN)/4)
		theHorizontalOffset = 0;
	theVerticalOffset += WINDOW_STAGGER;
	if (theVerticalOffset > GetSystemMetrics(SM_CYSCREEN)/4)
		theVerticalOffset = 0;

	return theWindow;
}


static double MeasureFramePeriod(void)
{
	//	QueryPerformanceCounter() is less reliable than one would hope.
	//	See
	//		http://blogs.msdn.com/oldnewthing/archive/2005/09/02/459952.aspx

	LARGE_INTEGER	theCurrentTime,		//	Can be inconsistent on certain dual-core systems.
					theClockFrequency;	//	Can vary depending on power management schemes.
	double			theFramePeriod;

	static bool				theSetUpFlag			= false;
	static LARGE_INTEGER	theLastRedrawTime		= {{0,0}};	//	initialize first member of union

	//	When deciding how far to advance the animation for the next frame,
	//	we face, in effect, a statistical question:  given the time required
	//	to render the last few frames, how long do we estimate it will take
	//	to render the next one?  For now we adopt the simplest possible answer:
	//	we guess that the next frame will take roughly as long as the previous
	//	one took.  In practice that approach seems to work pretty well.
	//	However, if jitteriness is a problem (for example if system activity
	//	slows one frame, in turn causing the animation to advance too far
	//	in the next frame) we could adopt a more sophisticated approach,
	//	for example we could keep track of the last three time intervals and
	//	use the middle value, which would in effect ignore any isolated anomalies.

	//	One-time initialization.
	if (theSetUpFlag == false)
	{
		QueryPerformanceCounter(&theLastRedrawTime);
		theSetUpFlag = true;
	}

	//	How long did it take to draw the last frame?
	QueryPerformanceCounter(&theCurrentTime);
	QueryPerformanceFrequency(&theClockFrequency);
	theFramePeriod	= (double)(theCurrentTime.QuadPart - theLastRedrawTime.QuadPart)
					/ (double)(theClockFrequency.QuadPart);
	theLastRedrawTime = theCurrentTime;
	
	return theFramePeriod;
}


static BOOL CALLBACK DoIdleTime(
	HWND	aWindow,
	LPARAM	anIdleTimeDataPtr)
{
	WindowData		*wd					= NULL;
	IdleTimeData	*theIdleTimeData	= NULL;

	//	If the window is ours...
	if (GetClassLong(aWindow, GCW_ATOM) == gMainWindowClass)
	{
		//	Prepare the input parameters.
		wd				= (WindowData *) GetWindowLongPtr(aWindow, GWLP_USERDATA);
		theIdleTimeData	= (IdleTimeData *) anIdleTimeDataPtr;
		if (wd == NULL || theIdleTimeData == NULL)	//	should never occur
			return true;	//	keep going
		
		//	Let the caller know that at least one window is present.
		theIdleTimeData->itsKeepGoingFlag = true;

		//	If the window...
		if ( ! IsIconic(aWindow)				//	isn't minimized and
		 && SimulationWantsUpdates(&wd->md))	//	wants an idle-time update, then...
		{
			//	Evolve its simulation over the given time period.
			SimulationUpdate(&wd->md, theIdleTimeData->itsFramePeriod);

			//	Declare the window contents invalid.
			//	In effect this puts a WM_PAINT message on the message queue.
			InvalidateRect(aWindow, NULL, false);

			//	Let the caller know that at least one animation is active.
			theIdleTimeData->itsAnimationFlag = true;
		}
	}

	return true;	//	keep going
}


static void DisplayFrameRate(
	IdleTimeData	*someIdleTimeData)
{
	unsigned int	theFrameRate;
	WCHAR			theWindowTitleFormat[64],
					theWindowTitle[64 + 8];	//	allow ample extra space for the digits

	static double	theTimeSinceLastDisplay	= 0.0;

	//	If the user wants to see the frame rate,
	//	write it to the window's title bar roughly once per second.

	theTimeSinceLastDisplay += someIdleTimeData->itsFramePeriod;

	if (gShowFrameRate
	 && theTimeSinceLastDisplay >= 1.0)
	{
		theTimeSinceLastDisplay = 0.0;

		if (someIdleTimeData->itsAnimationFlag)
		{
			//	Convert the frame period to a frame rate.
			if (someIdleTimeData->itsFramePeriod > 0.0001)
			{
				theFrameRate = (unsigned int) (1.0 / someIdleTimeData->itsFramePeriod);
				GetLocalizedTextAsWCHAR("fps", theWindowTitleFormat, BUFFER_LENGTH(theWindowTitleFormat));
			}
			else	//	frame rate ≥ 10000 fps
			{
				//	Display "∞ frames/second".
				theFrameRate = 0;	//	will be ignored
				GetLocalizedTextAsWCHAR("fps2", theWindowTitleFormat, BUFFER_LENGTH(theWindowTitleFormat));
			}
		}
		else	//	someIdleTimeData->itsAnimationFlag == false
		{
			//	Even though we're running the idle-time code 60 times per second,
			//	we're not actually drawing anything.  So it makes more sense
			//	to display 0 fps than 60 fps.
			theFrameRate = 0;
			GetLocalizedTextAsWCHAR("fps", theWindowTitleFormat, BUFFER_LENGTH(theWindowTitleFormat));
		}
	
		//	Note:  The codeblocks version of swprintf() doesn't check for buffer overruns.
		swprintf(	theWindowTitle,
					/*BUFFER_LENGTH(theWindowTitle),*/
					theWindowTitleFormat,
					theFrameRate);

		EnumThreadWindows(GetCurrentThreadId(), SetWindowTitle, (LPARAM) theWindowTitle);
	}
}


BOOL CALLBACK SetWindowTitle(
	HWND	aWindow,
	LPARAM	aTitle)	//	theTitle = (WCHAR *)aTitle, may be NULL
{
	WindowData	*wd = NULL;
	WCHAR		theWindowTitle[64];

	//	Note:  On my system the Input Method Environment (IME) creates
	//	secret windows that shadow the main window.  They are top-level windows,
	//	so EnumThreadWindows() finds them.  Specifically, each main window owns
	//	a secret window of class "IME", which in turn owns another secret window
	//	of class "MSCTFIME UI".  So we must test each window to see whether
	//	it's really one of ours.

	//	If the window is ours...
	if (GetClassLong(aWindow, GCW_ATOM) == gMainWindowClass)
	{
		if ((WCHAR *)aTitle != NULL)
		{
			//	Set theTitle if one is given (typically the frame rate)...
			SetWindowText(aWindow, (WCHAR *)aTitle);
		}
		else
		{
			//	...otherwise set the window's default title (typically the file name).
			wd = (WindowData *) GetWindowLongPtr(aWindow, GWLP_USERDATA);
			if (wd != NULL
			 && wd->itsFileTitle[0] != 0)
			{
				SetWindowText(aWindow, wd->itsFileTitle);
			}
			else
			{
				SetWindowText(aWindow, GetLocalizedTextAsWCHAR("WindowTitle", theWindowTitle, BUFFER_LENGTH(theWindowTitle)));
			}
		}
	}

	return true;	//	keep going
}
