//	CurvedSpaces-Win32-WndProc.c
//
//	Provides MainWndProc(), DrawingWndProc() and supporting functions
//	for the Win32 version of Curved Spaces.
//
//	© 2009 by Jeff Weeks
//	See TermsOfUse.txt

#include "CurvedSpaces-Win32.h"
#include "CurvedSpacesIDs.h"
#include <commdlg.h>
#include <math.h>		//	for sqrt()
#include <stdio.h>		//	for sprintf()
#include <shellapi.h>	//	for ShellExecute()
#include <commctrl.h>	//	for tab control in preferences panel
#include "GeometryGames-Win32-LoadEntryPoints.h"

//	What is the minimum acceptable size for the window's client area?
#define MIN_CLIENT_SIZE	256


//	Command IDs

#define IDC_FILE_OPEN						0x0000
#define IDC_FILE_OPEN_NEW					0x0001
#define IDC_FILE_EXPORT_COPY				0x0002
#define IDC_FILE_EXPORT_SAVE				0x0003
#define IDC_FILE_EXIT						0x0004

#define IDC_VIEW_CENTERPIECE_NONE			0x0100
#define IDC_VIEW_CENTERPIECE_EARTH			0x0101
#define IDC_VIEW_CENTERPIECE_GALAXY			0x0102
#define IDC_VIEW_CENTERPIECE_GYROSCOPE		0x0103
#define IDC_VIEW_OBSERVER					0x0104
#define IDC_VIEW_COLOR_CODING				0x0105
#define IDC_VIEW_CLIFFORD					0x0106
#define IDC_VIEW_VERTEX_FIGURES				0x0107
#define IDC_VIEW_FOG						0x0108
#define IDC_VIEW_MULTISAMPLING				0x0109
#define IDC_VIEW_GRAPHICS_FIXED_FUNCTION	0x010A
#define IDC_VIEW_GRAPHICS_SHADERS			0x010B
#define IDC_VIEW_STEREO_3D					0x010C

#define IDC_LANGUAGE_DE						0x0200
#define IDC_LANGUAGE_EN						0x0201
#define IDC_LANGUAGE_FR						0x0202
#define IDC_LANGUAGE_JA						0x0203

#define IDC_HELP_HELP						0x0300
#define IDC_HELP_PREFERENCES				0x0301
#define IDC_HELP_ABOUT						0x0302


//	How many intrinsic units correspond to each pixel of mouse motion?
#define UNITS_PER_PIXEL			0.001

//	Even when successful, GetOpenFileName() throws access violation exceptions,
//	apparently from within RPCRT4.dll (a Remote Procedure Call library).
//	This problem is widely discussed on the web, but unfortunately
//	the sites I found provided neither an explanation nor a fix.
//	Under normal operation the exceptions aren't a problem:
//	we ignore them, they get passed back to Windows, and GetOpenFileName()
//	returns successfully.  However, they thoroughly confuse
//	the gdb debugger (running under CodeBlocks):  they trash the stack
//	and debugging comes screeching to a halt.  To avoid this problem
//	and permit meaningful debugging, bypass the GetOpenFileName() call.
//#define BYPASS_GETOPENFILENAME


//	Support for WGL extensions.

#define WGL_DRAW_TO_WINDOW_ARB         0x2001
#define WGL_ACCELERATION_ARB           0x2003
#define WGL_SUPPORT_OPENGL_ARB         0x2010
#define WGL_DOUBLE_BUFFER_ARB          0x2011
#define WGL_COLOR_BITS_ARB             0x2014
#define WGL_ALPHA_BITS_ARB             0x201B
#define WGL_DEPTH_BITS_ARB             0x2022
#define WGL_STENCIL_BITS_ARB           0x2023
#define WGL_FULL_ACCELERATION_ARB      0x2027
#define WGL_SAMPLE_BUFFERS_ARB         0x2041
#define WGL_SAMPLES_ARB                0x2042

typedef const char	*(WINAPI * PFNWGLGETEXTENSIONSSTRINGARBPROC)(HDC hdc);
typedef BOOL		(WINAPI * PFNWGLCHOOSEPIXELFORMATARBPROC)(HDC hdc, const int *piAttribIList, const FLOAT *pfAttribFList, UINT nMaxFormats, int *piFormats, UINT *nNumFormats);
typedef BOOL		(WINAPI * PFNWGLSWAPINTERVALEXTPROC )(int interval);

static  PFNWGLSWAPINTERVALEXTPROC	wglSwapInterval	= NULL;


//	Remember the Preferences Panel's preferred page.
static unsigned int	gPreferencesPreferredPage = 0;


static bool				SetUpWindowData(HWND aWindow);
static void				ShutDownWindowData(WindowData *wd);
static bool				SetUpOpenGL(WindowData *wd);
static void				ShutDownOpenGL(WindowData *wd);
static bool				SetUpDrawingPanel(WindowData *wd);
static bool				SetUpDrawingPanelWithPixelFormat(WindowData *wd, unsigned int aPixelFormat);
static unsigned int		ChooseMultisamplePixelFormat(WindowData *wd);
static void				ShutDownDrawingPanel(WindowData *wd);
static bool				WholeWordSubstring(const char *aString, const char *aPotentialSubstring);
static void				MenuPrepare(WindowData *wd, HMENU aMenu);
static void				MenuCommand(WindowData *wd, WORD aCommand);
static ErrorText		ReadRawText(WCHAR *aFileName, Byte **someRawText);
static BOOL CALLBACK	PostCloseMessage(HWND aWindow, LPARAM anUnusedParameter);
static void				CopyTheImage(WindowData *wd);
static void				SaveTheImage(WindowData *wd);
static bool				CopyDirectoryName(WCHAR *aPathName, DWORD aPathNameLength, WCHAR *aDirectoryName, DWORD aDirectoryNameLength);
static HANDLE			FetchTheImage(WindowData *wd);
static void				ChangeLanguage(const char aLanguageCode[3]);
static void				RefreshMirroring(HWND aWindow);
static HMENU			BuildLocalizedMenuBar(void);
static void				OpenHelpBook(void);
static void				SetPrefPanelLabels(HWND aPreferencesDialog);
static void				SwapPreferencesPanel(WindowData *wd, HWND aPreferencesDialog, unsigned int aPageIndex);
static INT_PTR CALLBACK	PreferencesDialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
static INT_PTR CALLBACK PrefsPerspectiveDialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
static void				EnableStereoFields(HWND aDialog, bool aStereoFlag);
static void				FormatUnsignedInteger32(WCHAR *aBuffer, unsigned int aBufferLength, unsigned int aNumber);
static void				UnformatUnsignedInteger32(WCHAR *aBuffer, unsigned int aBufferLength, unsigned int *aNumber);
static INT_PTR CALLBACK PrefsAdvancedDialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
static void				SetSync(WindowData *wd);
static void				SteerWithMouse(WindowData *wd, WPARAM wParam, LPARAM lParam, bool aSteeringFlag);
static void				SetMinWindowSize(HWND aWindow, POINT *aMinSize);
static void				ResizeDrawingPanel(WindowData *wd, unsigned int aWidth, unsigned int aHeight);
static void				DisplayChange(WindowData *wd);
static void				PaintWindow(WindowData *wd);


//	Callback function to process messages in the main window.
LRESULT CALLBACK MainWndProc(
	HWND	hWnd,
	UINT	uMsg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData	*wd = NULL;

	//	For all messages except WM_CREATE and WM_DESTROY, locate the WindowData.
	//	If we confirm once and for all that wd != NULL, then
	//	the routines we call in the message loop can safely rely on it.
	if (hWnd != NULL)
	{
		if (uMsg != WM_CREATE
		 && uMsg != WM_DESTROY)
		{
			wd = (WindowData *) GetWindowLongPtr(hWnd, GWLP_USERDATA);
			if (wd == NULL)
				return DefWindowProc(hWnd, uMsg, wParam, lParam);
		}
		else
			wd = NULL;
	}
	else
		return 0;	//	should never occur

	//	Process the message.
	switch (uMsg)
	{
		case WM_CREATE:
			if (SetUpWindowData(hWnd) == true)
				return 0;	//	success -- continue normally
			else
				return -1;	//	failure -- abort window creation

		case WM_CLOSE:
			//	Shut down our window data, including itsRenderingContext,
			//	itsDeviceContext and itsDrawingPanel.  Otherwise Windows will destroy
			//	child windows prematurely, leaving itsRenderingContext invalid.
			ShutDownWindowData(wd);
			DestroyWindow(hWnd);
			return 0;

//		case WM_DESTROY:
//			return 0;

		case WM_KEYDOWN:
			switch (wParam)	//	virtual key code
			{
				//	Note:  Return, Escape and Space could be handled
				//	in WM_CHAR (as 0x0D, 0x1B and ' ', respectively),
				//	but the arrow keys don't generate characters 
				//	so they must be handled here.

				case VK_UP:		wd->md.itsUserSpeed += USER_SPEED_INCREMENT;	break;
				case VK_DOWN:	wd->md.itsUserSpeed -= USER_SPEED_INCREMENT;	break;
				case ' ':		wd->md.itsUserSpeed = 0.0;						break;

				case VK_LEFT:
					ChangeAperture(&wd->md, true);
					break;
				case VK_RIGHT:
					ChangeAperture(&wd->md, false);
					break;

#ifdef START_OUTSIDE
				case VK_RETURN:
					if (wd->md.itsViewpoint == ViewpointExtrinsic)
						wd->md.itsViewpoint = ViewpointEntering;
					break;
#endif

				case VK_ESCAPE:

					//	A right-click (or even a left-click) would be a better way
					//	to exit navigational mode.  Nevertheless, we don't want
					//	to leave a novice user stranded without a mouse.

					//	Done steering.  Release the mouse.
					ReleaseCapture();

					break;
			}
			return 0;

		case WM_CHAR:
			switch ((WCHAR) wParam)	//	character code
			{
				case 'f':	//	"secret" keyboard command to show frame rate
					gShowFrameRate = ! gShowFrameRate;
					EnumThreadWindows(	GetCurrentThreadId(),
										SetWindowTitle,
										(LPARAM) (gShowFrameRate ? L"…" : NULL));
					break;

				case 's':	//	"secret" keyboard command to sync to the monitor
					wd->md.itsSyncFlag = ! wd->md.itsSyncFlag;
					SetSync(wd);
					break;
			}
			return 0;

		case WM_MOUSEWHEEL:
			wd->md.itsUserSpeed += USER_SPEED_INCREMENT * ((short)HIWORD(wParam))/120.0;
			return 0;

		case WM_INITMENUPOPUP:
			MenuPrepare(wd, (HMENU) wParam);
			return 0;

		case WM_COMMAND:
			MenuCommand(wd, LOWORD(wParam));
			return 0;

//		case WM_EXITMENULOOP:
//		case WM_EXITSIZEMOVE:
//		case WM_ACTIVATE:
//			//	Warning:  wd will be NULL when a WM_ACTIVATE event
//			//	arrives during window destruction.
//			return 0;

		case WM_GETMINMAXINFO:
			SetMinWindowSize(hWnd, &((MINMAXINFO*)lParam)->ptMinTrackSize);
			return 0;

		case WM_SIZE:
			MoveWindow(wd->itsDrawingPanel, 0, 0, LOWORD(lParam), HIWORD(lParam), false);
			return 0;

		case WM_DISPLAYCHANGE:
			DisplayChange(wd);
			return 0;

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
}


//	Callback function to process messages in the drawing panel.
LRESULT CALLBACK DrawingWndProc(
	HWND	aDrawingWindow,
	UINT	uMsg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData	*wd;

	//	If we confirm once and for all that wd != NULL, then the routines
	//	we call in the message switch can safely rely on it.
	//	See MainWndProc() for an explanation of why the WindowData
	//	isn't available at WM_DESTROY time.
	if (aDrawingWindow != NULL)
	{
		if (uMsg != WM_CREATE
		 && uMsg != WM_DESTROY)
		{
			wd = (WindowData *) GetWindowLongPtr(GetParent(aDrawingWindow), GWLP_USERDATA);
			if (wd == NULL)
				return DefWindowProc(aDrawingWindow, uMsg, wParam, lParam);
		}
		else
			wd = NULL;
	}
	else
		return 0;	//	should never occur

	switch (uMsg)
	{
		case WM_CREATE:
			return 0;

		case WM_LBUTTONDBLCLK:
			//	For consistency with the Torus Games,
			//	we want a double-click to get us out of navigational mode.
			//	Nevertheless, for the convenience of CurvedSpaces-only users,
			//	a single-click should also do the job.  The solution is
			//	to exit navigational mode on the first click, and then
			//	ignore the second click (if any) so that it doesn't put us
			//	back into navigational mode.
			return 0;

		case WM_LBUTTONDOWN:
			if (GetCapture() != aDrawingWindow)
			{
				//	Capture the mouse for steering.
				SetCapture(aDrawingWindow);
				SetCursor(NULL);
				SteerWithMouse(wd, wParam, lParam, false);
			}
			else
			{
				//	Done steering.  Release the mouse.
				ReleaseCapture();
			}
			return 0;

		case WM_RBUTTONDOWN:
			//	Done steering.  Release the mouse.
			ReleaseCapture();
			return 0;

		case WM_MBUTTONDOWN:	//	includes mouse-wheel clicks
			wd->md.itsUserSpeed = 0.0;
			return 0;

		case WM_MOUSEMOVE:
			if (GetCapture() == aDrawingWindow)
				SteerWithMouse(wd, wParam, lParam, true);
			return 0;

		case WM_PAINT:
			PaintWindow(wd);
			return 0;

		case WM_SIZE:
			ResizeDrawingPanel(wd, LOWORD(lParam), HIWORD(lParam));
			return 0;

		default:
			return DefWindowProc(aDrawingWindow, uMsg, wParam, lParam);
	}
}


static bool SetUpWindowData(HWND aWindow)
{
	WindowData	*wd;
	bool		theShadersAreAvailable;

	//	Attempt to allocate the WindowData and -- whether or not the call
	//	succeeds -- attach the resulting pointer to the window.
	wd = GET_MEMORY(sizeof(WindowData));
	SetWindowLongPtr(aWindow, GWLP_USERDATA, (LONG_PTR) wd);

	//	If the memory allocation failed, return false
	//	to tell Windows to abort window construction.
	if (wd == NULL)
	{
		ErrorMessage("Couldn't get memory for WindowData.", "SetUpWindowData() Error");
		return false;
	}

	//	Record the window.
	wd->itsWindow = aWindow;

	//	Set pointers to NULL for robust error handling.
	wd->itsDrawingPanel			= NULL;
	wd->itsDeviceContext		= NULL;
	wd->itsRenderingContext		= NULL;
	wd->itsPreferencesPane		= NULL;
	wd->md.itsDirichletDomain	= NULL;
	wd->md.itsHoneycomb			= NULL;

	//	Initialize itsFileTitle to an empty string
	//	for use until a file gets read.
	wd->itsFileTitle[0] = 0;

	//	Initialize the model's internal data.
	SetUpModelData(&wd->md);

	//	Set up OpenGL.
	if (SetUpOpenGL(wd) == false)
		goto FailedToSetUpData;

	//	Does the OpenGL context support shaders?
	//	If not, tell the user that the fallback fixed-function code
	//	will be used and the quality of the image will be less.
	//	Display that message once, and then never again (for that user).
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	theShadersAreAvailable = ShadersAreAvailable();
	wglMakeCurrent(NULL, NULL);
	if ( ! theShadersAreAvailable
	  && ! GetUserPref("fallback message given"))
	{
		ErrorMessage(GetLocalizedText("ErrorShadersMessage"), GetLocalizedText("ErrorShadersTitle"));
		SetUserPref("fallback message given", true);
	}

	//	Report success.
	return true;

FailedToSetUpData:

	//	Clean up.
	ShutDownWindowData(wd);	//	wd is known to be non-NULL here

	//	Report failure.
	ErrorMessage("Failed to set up window.", "SetUpWindowData() Error");
	return false;
}


static void ShutDownWindowData(WindowData *wd)
{
	ShutDownOpenGL(wd);

	ShutDownModelData(&wd->md);

	//	Once all dependent data have been cleaned up,
	//	then free the WindowData itself.
	SetWindowLongPtr(wd->itsWindow, GWLP_USERDATA, 0);
	FREE_MEMORY(wd);
}


static bool SetUpOpenGL(WindowData *wd)
{
	bool	theLoadResult;

	//	Create a drawing panel, get a device context and
	//	attach an OpenGL rendering context.
	if (SetUpDrawingPanel(wd) == false)
		return false;

	//	On Win32 we must explicitly load functions from OpenGL 1.2 or later.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	theLoadResult = LoadOpenGLFunctions();
	wglMakeCurrent(NULL, NULL);	//	Deactivate context before displaying error message.
	if ( ! theLoadResult )
	{
		ErrorMessage("This program cannot continue because your computer's graphics library does not support \"Vertex Buffer Objects\".", "OpenGL 1.5 Required");
		ExitProcess(5);
	}
	
	//	Load an entry point for wglSwapInterval, if it's present.
	//	Caution:  My mother's Dell with Intel Extreme Graphics 
	//	does not support WGL_EXT_swap_control but nevertheless returns a non-NULL
	//	(but ultimately useless) entry point for wglGetProcAddress("wglSwapIntervalEXT").
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	wglSwapInterval = (PFNWGLSWAPINTERVALEXTPROC) wglGetProcAddress("wglSwapIntervalEXT");
	wglMakeCurrent(NULL, NULL);

	//	Set sync-to-monitor as desired.
	SetSync(wd);

	//	The following flags should already be false
	//	(either from the original window initialization
	//	or from a call to ShutDownOpenGL()), but reset them
	//	here just to be safe.
	wd->md.itsPreparedExtensions	= false;
	wd->md.itsPreparedShaders		= false;
	wd->md.itsPreparedTextures		= false;
	wd->md.itsPreparedVBOs			= false;

	return true;
}


static void ShutDownOpenGL(WindowData *wd)
{
	//	Shut down application-specific OpenGL textures, display lists
	//	and vertex shaders.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ShutDownGraphicsAsNeeded(&wd->md);
	wglMakeCurrent(NULL, NULL);

	ShutDownDrawingPanel(wd);
}


static bool SetUpDrawingPanel(WindowData *wd)
{
	unsigned int	thePixelFormat;

	//	The multisampling extension presents a chicken-and-egg problem:
	//	we need to use it to set up the OpenGL rendering context,
	//	but we can't find out whether it's available -- or even get
	//	a pixel format -- until after the rendering context has been created
	//	and made current.

	//	Set up an OpenGL drawing panel with a plain (non-multisampling) buffer.
	if (SetUpDrawingPanelWithPixelFormat(wd, 0) != true)
		return false;

	//	Try for a multi-sample pixel format if the user wants one.
	//	Note:  For sure we'll have at least the default value of
	//	itsMultisamplingEnabledFlag, but at start-up we won't yet
	//	have tested or set itsMultisamplingAvailableFlag.
	//	We hope it's available, but if it's not ChooseMultisamplePixelFormat()
	//	will fail harmlessly.
	if (wd->md.itsMultisamplingEnabledFlag)
	{
		//	Use the plain context to access the WGL_ARB_multisample extension
		//	and select a multisample-capable pixel format.
		thePixelFormat = ChooseMultisamplePixelFormat(wd);

		//	If a multisample-capable pixel format is available,
		//	destroy the plain drawing panel and replace it
		//	with a multisample-capable panel.
		if (thePixelFormat != 0)
		{
			ShutDownDrawingPanel(wd);
			
			if (SetUpDrawingPanelWithPixelFormat(wd, thePixelFormat) != true)
			{
				ErrorMessage("Failed to create multisample-capable OpenGL context.  Falling back to plain OpenGL context.", "Full-scene antialiasing not working");
				wd->md.itsMultisamplingEnabledFlag = false;
				SetUserPref("multisampling", false);
				if (SetUpDrawingPanelWithPixelFormat(wd, 0) != true)
					return false;
			}
		}
	}

	return true;
}


static bool SetUpDrawingPanelWithPixelFormat(
	WindowData		*wd,
	unsigned int	aPixelFormat)	//	needed only for multisampling;  otherwise pass 0
{
	ErrorText				theErrorMessage	= NULL,
							theErrorTitle	= "SetUpDrawingPanel() failed";	//	default title, may be overridden
	RECT					theClientRect;
	PIXELFORMATDESCRIPTOR	thePixelFormatDescriptor =
	{
		sizeof(PIXELFORMATDESCRIPTOR),	//	size
		1,								//	version number
		  PFD_DRAW_TO_WINDOW			//	format
		| PFD_SUPPORT_OPENGL
	//	| PFD_STEREO	//	could be useful in the future, but not needed for anaglyphic stereo 3D
		| PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,					//	preferred format is rgba (not palette)
		32, 							//	preferred color buffer depth (including alpha bits)
		0, 0, 0, 0, 0, 0,				//	color bits ignored
		8,								//	alpha buffer depth
		0,								//	shift bit ignored
		0,								//	no accumulation buffer
		0, 0, 0, 0, 					//	accumulation bits ignored
		16, 							//	16-bit Z-buffer
		0,								//	no stencil buffer
		0,								//	no auxiliary buffer
		0,								//	ignored
		PFD_MAIN_PLANE, 				//	underlay/overlay planes
		0,								//	ignored
		0,								//	underlay color
		0								//	ignored
	};
	
	if (wd->itsDrawingPanel != NULL)
	{
		theErrorMessage = "Window has pre-existing drawing panel.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}
	if (wd->itsDeviceContext != NULL)
	{
		theErrorMessage = "Window has pre-existing device context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}
	if (wd->itsRenderingContext != NULL)
	{
		theErrorMessage = "Window has pre-existing rendering context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	You'd think you could create itsDrawingPanel once and for all,
	//	and then replace its OpenGL rendering context as needed,
	//	but in practice that doesn't work.  So instead we must
	//	re-create itsDrawingPanel whenever we need to change
	//	the OpenGL rendering context (for example to change
	//	to a multisample-capable pixel format).
	//
	//	Note:  Calling CreateWindow() with the WS_VISIBLE style
	//	triggers a WM_SIZE message which leads to a ResizeDrawingPanel()
	//	call which adjusts the window size variables set earlier.
	GetClientRect(wd->itsWindow, &theClientRect);
	wd->itsDrawingPanel = CreateWindow(
		DRAWING_WINDOW_CLASS_NAME,
		NULL,
		WS_CHILD | WS_VISIBLE,
		theClientRect.left,	//	should be 0
		theClientRect.top,	//	should be 0
		theClientRect.right - theClientRect.left,
		theClientRect.bottom - theClientRect.top,
		wd->itsWindow, 0, GetModuleHandle(NULL), NULL);
	if (wd->itsDrawingPanel == NULL)
	{
		theErrorMessage = "Can't create a window for OpenGL graphics.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	wd->itsDeviceContext = GetDC(wd->itsDrawingPanel);
	if (wd->itsDeviceContext == NULL)
	{
		theErrorMessage = "Can't get a device context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	if (aPixelFormat == 0)
	{
		aPixelFormat = ChoosePixelFormat(wd->itsDeviceContext, &thePixelFormatDescriptor);
		if (aPixelFormat == 0)
		{
			theErrorMessage = "Can't get a pixel format.";
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}
	}

	if (DescribePixelFormat(wd->itsDeviceContext, aPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &thePixelFormatDescriptor))
	{
		if (thePixelFormatDescriptor.iPixelType	!= PFD_TYPE_RGBA
		 || (	thePixelFormatDescriptor.cColorBits	!= 24	//	Microsoft's documentation claims that cColorBits
		 	 && thePixelFormatDescriptor.cColorBits	!= 32))	//	excludes the alpha bits, but in reality it includes them.
		{
			theErrorMessage	= GetLocalizedText("ErrorColorMessage");
			theErrorTitle	= GetLocalizedText("ErrorColorTitle");
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}

		if (thePixelFormatDescriptor.cAlphaBits	!= 8)
		{
			theErrorMessage = GetLocalizedText("ErrorAlphaMessage");
			theErrorTitle	= GetLocalizedText("ErrorAlphaTitle");
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}
	}

	if (SetPixelFormat(wd->itsDeviceContext, aPixelFormat, &thePixelFormatDescriptor) == false)
	{
		theErrorMessage = "Can't set the pixel format.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	wd->itsRenderingContext = wglCreateContext(wd->itsDeviceContext);
	if (wd->itsRenderingContext == NULL)
	{
		theErrorMessage = "Can't get an OpenGL rendering context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	Let's do a "dry run" to test itsRenderingContext right here and now.
	//	That way we won't need to clutter up the rest of the code with
	//	error checking on wglMakeCurrent().  In the highly unlikely event
	//	that wglMakeCurrent() succeeds now and fails later on, nothing
	//	too terrible should happen because when wglMakeCurrent() fails
	//	to set the requested context it instead sets wglMakeCurrent(NULL, NULL),
	//	meaning that all subsequent OpenGL calls will be ignored.
	if (wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext) == false)
	{
		theErrorMessage = "Can't make OpenGL rendering context current.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	The generic Windows OpenGL driver ignores 3D graphics hardware.
	//	In practice this means that a computer from, say, Dell will run
	//	OpenGL fine because Dell provides a good driver, but a standard
	//	installation from a Windows XP CD will run OpenGL in software emulation
	//	(no doubt because Microsoft wants Microsoft-only DirectX graphics
	//	to run well while platform-independent OpenGL graphics runs poorly).
	//	I don't know any direct way to query whether hardware acceleration
	//	is active, but while our rendering context is current we can at least
	//	check the renderer name, and if it's "GDI Generic" notify the user.
	if (strcmp(GetRendererName(), "GDI Generic") == 0)
	{
		theErrorMessage = GetLocalizedText("ErrorDriverMessage");
		theErrorTitle	= GetLocalizedText("ErrorDriverTitle");
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

CleanUpSetUpDrawingPanelWithPixelFormat:

	//	Clear the OpenGL context before calling ErrorMessage().
	wglMakeCurrent(NULL, NULL);

	if (theErrorMessage != NULL)
	{
		ErrorMessage(theErrorMessage, theErrorTitle);
		ShutDownDrawingPanel(wd);
		return false;
	}
	else
		return true;
}


static unsigned int ChooseMultisamplePixelFormat(WindowData *wd)
{
	signed int							thePixelFormat;
	PFNWGLGETEXTENSIONSSTRINGARBPROC	wglGetExtensionsStringARB;
	const char							*theWGLExtensionsList;	//	always ASCII, never Unicode
	PFNWGLCHOOSEPIXELFORMATARBPROC		wglChoosePixelFormatARB;
	signed int							thePixelFormatIntAttributes[] =
	{
		WGL_DRAW_TO_WINDOW_ARB,	true,
		WGL_SUPPORT_OPENGL_ARB,	true,
		WGL_DOUBLE_BUFFER_ARB,	true,
		WGL_ACCELERATION_ARB,	WGL_FULL_ACCELERATION_ARB,
		WGL_COLOR_BITS_ARB,		32,	//	including alpha bits (contrary to some online documentation)
		WGL_ALPHA_BITS_ARB,		8,
		WGL_DEPTH_BITS_ARB,		16,
		WGL_STENCIL_BITS_ARB,	0,
		WGL_SAMPLE_BUFFERS_ARB,	true,
		WGL_SAMPLES_ARB,		8,	//	the code below assumes the number of samples is at array index 19
		0,						0
	};
	float								thePixelFormatFloatAttributes[] = {0.0f, 0.0f};
	bool								theFormatIsValid;
	unsigned int						theNumPixelFormats;

	//	If we encounter an error before finding a pixel format, we'll return 0.
	thePixelFormat = 0;

	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);

	wglGetExtensionsStringARB = (PFNWGLGETEXTENSIONSSTRINGARBPROC) wglGetProcAddress("wglGetExtensionsStringARB");
	if (wglGetExtensionsStringARB == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	theWGLExtensionsList = wglGetExtensionsStringARB(wd->itsDeviceContext);
	if (theWGLExtensionsList == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	if (WholeWordSubstring(theWGLExtensionsList, "WGL_ARB_multisample" ) == false
	 || WholeWordSubstring(theWGLExtensionsList, "WGL_ARB_pixel_format") == false)
		goto CleanUpChooseMultisamplePixelFormat;

	wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC) wglGetProcAddress("wglChoosePixelFormatARB");
	if (wglChoosePixelFormatARB == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	//	Try for 8-sample multisampling, then 6 samples, then 4, then 2, then give up.
	//	Note:  Some applications might want to voluntarily use fewer than 8 samples,
	//	to keep memory usage down.  Conversely, someday we might want to be more
	//	ambitious and try for more than 8 samples.
	do
	{
		theFormatIsValid = wglChoosePixelFormatARB(
								wd->itsDeviceContext,
								thePixelFormatIntAttributes,
								thePixelFormatFloatAttributes,
								1,
								&thePixelFormat,
								&theNumPixelFormats)
						&&
							theNumPixelFormats >= 1;

		//	Make sure the WGL_SAMPLES_ARB parameter sits where we think it sits,
		//	then decrement its value by 2.
		if (thePixelFormatIntAttributes[18] != WGL_SAMPLES_ARB)
			goto CleanUpChooseMultisamplePixelFormat;
		thePixelFormatIntAttributes[19] -= 2;

	} while (theFormatIsValid == false && thePixelFormatIntAttributes[19] > 0);

	if (theFormatIsValid == false)
	{
		thePixelFormat = 0;
		goto CleanUpChooseMultisamplePixelFormat;
	}

CleanUpChooseMultisamplePixelFormat:

	wglMakeCurrent(NULL, NULL);

	return thePixelFormat;
}


static void ShutDownDrawingPanel(WindowData *wd)
{
	wglMakeCurrent(NULL, NULL);	//	redundant, but safe

	if (wd->itsRenderingContext != NULL)
	{
		wglDeleteContext(wd->itsRenderingContext);
		wd->itsRenderingContext = NULL;
	}

	if (wd->itsDeviceContext != NULL)
	{
		ReleaseDC(wd->itsDrawingPanel, wd->itsDeviceContext);
		wd->itsDeviceContext = NULL;
	}

	if (wd->itsDrawingPanel != NULL)
	{
		DestroyWindow(wd->itsDrawingPanel);
		wd->itsDrawingPanel = NULL;
	}
}


static bool WholeWordSubstring(
	const char	*aString,				//	null-terminated
	const char	*aPotentialSubstring)	//	null-terminated
{
	const char	*thePotentialMatch,
				*s,
				*p;
	bool		theMatchFlag;

	//	Does aPotentialSubstring occur as a complete space-delimited word in aString?

	//	Irrelevant technical detail:  an empty aPotentialSubstring
	//	(consisting of a terminating zero alone) counts as a substring
	//	of any string, even an empty one.

	//	Consider the first character in each space-delimited word of aString
	//	as the starting point for a potential match.
	thePotentialMatch = aString;
	while (*thePotentialMatch == ' ')
		thePotentialMatch++;
	do
	{
		//	Compare each character in aPotentialSubstring
		//	(excluding the terminating zero) to thePotentialMatch.
		s = thePotentialMatch;
		p = aPotentialSubstring;
		theMatchFlag = true;
		while (theMatchFlag == true && *p != 0)
		{
			if (*p++ != *s++)
				theMatchFlag = false;
		}

		//	If a space or terminating 0 follows a match, return true.
		if (theMatchFlag == true
		 && (*s == ' ' || *s == 0))
			return true;

		//	Skip to the beginning of the next whole word in thePotentialMatch.
		//		Skip by non-spaces.
		while (*thePotentialMatch != ' ' && *thePotentialMatch != 0)
			thePotentialMatch++;
		//		Skip by intervening spaces.
		while (*thePotentialMatch == ' ')
			thePotentialMatch++;

		//	Continue with the loop to examine the next whole word (if any).

	} while (*thePotentialMatch != 0);

	return false;
}


static void MenuPrepare(
	WindowData	*wd,
	HMENU		aMenu)
{
	CheckMenuItem(aMenu, IDC_VIEW_CENTERPIECE_NONE,
		wd->md.itsSceneryType == SceneryNone      ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_CENTERPIECE_EARTH,
		wd->md.itsSceneryType == SceneryEarth     ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_CENTERPIECE_GALAXY,
		wd->md.itsSceneryType == SceneryGalaxy    ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_CENTERPIECE_GYROSCOPE,
		wd->md.itsSceneryType == SceneryGyroscope ? MF_CHECKED : MF_UNCHECKED);

	CheckMenuItem(aMenu, IDC_VIEW_OBSERVER,
		wd->md.itsShowObserver ? MF_CHECKED : MF_UNCHECKED);

	CheckMenuItem(aMenu, IDC_VIEW_COLOR_CODING,
		wd->md.itsShowColorCoding ? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_CLIFFORD,
		wd->md.itsSpaceType == SpaceSpherical ? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(aMenu, IDC_VIEW_CLIFFORD,
		(	wd->md.itsSpaceType == SpaceSpherical
		 && wd->md.itsShowCliffordParallels)
			? MF_CHECKED : MF_UNCHECKED);

	CheckMenuItem(aMenu, IDC_VIEW_VERTEX_FIGURES,
		wd->md.itsShowVertexFigures ? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_FOG,
		(	wd->md.itsShadersAvailableFlag
		 && wd->md.itsShadersEnabledFlag)
			? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(aMenu, IDC_VIEW_FOG,
		(	wd->md.itsShadersAvailableFlag
		 && wd->md.itsShadersEnabledFlag
		 && wd->md.itsFogFlag)
			? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_MULTISAMPLING,
		wd->md.itsMultisamplingAvailableFlag ? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(aMenu, IDC_VIEW_MULTISAMPLING,
		(wd->md.itsMultisamplingAvailableFlag && wd->md.itsMultisamplingEnabledFlag)
			? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_GRAPHICS_FIXED_FUNCTION,
		MF_ENABLED);
	CheckMenuItem(aMenu, IDC_VIEW_GRAPHICS_FIXED_FUNCTION,
		! (wd->md.itsShadersAvailableFlag && wd->md.itsShadersEnabledFlag)
			? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_GRAPHICS_SHADERS,
		wd->md.itsShadersAvailableFlag ? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(aMenu, IDC_VIEW_GRAPHICS_SHADERS,
		(wd->md.itsShadersAvailableFlag && wd->md.itsShadersEnabledFlag)
			? MF_CHECKED : MF_UNCHECKED);

	CheckMenuItem(aMenu, IDC_VIEW_STEREO_3D,
		wd->md.itsStereoFlag ? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_LANGUAGE_DE, IsValidLocale(MAKELCID(MAKELANGID(LANG_GERMAN,   SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_LANGUAGE_EN, IsValidLocale(MAKELCID(MAKELANGID(LANG_ENGLISH,  SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_LANGUAGE_FR, IsValidLocale(MAKELCID(MAKELANGID(LANG_FRENCH,   SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_LANGUAGE_JA, IsValidLocale(MAKELCID(MAKELANGID(LANG_JAPANESE, SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);

	CheckMenuItem(aMenu, IDC_LANGUAGE_DE, SAME_LANGUAGE_CODE(gLanguageCode, "de") ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_LANGUAGE_EN, SAME_LANGUAGE_CODE(gLanguageCode, "en") ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_LANGUAGE_FR, SAME_LANGUAGE_CODE(gLanguageCode, "fr") ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_LANGUAGE_JA, SAME_LANGUAGE_CODE(gLanguageCode, "ja") ? MF_CHECKED : MF_UNCHECKED);
}


static void MenuCommand(
	WindowData	*wd,
	WORD		aCommand)
{
	switch (aCommand)
	{
		//	file menu

		case IDC_FILE_OPEN:
			DoFileOpen(wd);
			break;

		case IDC_FILE_OPEN_NEW:
			(void) CreateCurvedSpacesWindow();
			break;

		case IDC_FILE_EXPORT_COPY:
			CopyTheImage(wd);
			break;

		case IDC_FILE_EXPORT_SAVE:
			SaveTheImage(wd);
			break;

		case IDC_FILE_EXIT:
			//	Send each window a WM_CLOSE message.
			//	If all windows comply, our message loop in WinMain() will terminate.
			EnumThreadWindows(GetCurrentThreadId(), PostCloseMessage, 0);
			break;
		
		//	view menu

		case IDC_VIEW_CENTERPIECE_NONE:
			SetSceneryType(&wd->md, SceneryNone);
			break;

		case IDC_VIEW_CENTERPIECE_EARTH:
			SetSceneryType(&wd->md, SceneryEarth);
			break;

		case IDC_VIEW_CENTERPIECE_GALAXY:
			SetSceneryType(&wd->md, SceneryGalaxy);
			break;

		case IDC_VIEW_CENTERPIECE_GYROSCOPE:
			SetSceneryType(&wd->md, SceneryGyroscope);
			break;

		case IDC_VIEW_OBSERVER:
			SetShowObserver(&wd->md, ! wd->md.itsShowObserver );
			break;

		case IDC_VIEW_COLOR_CODING:
			SetShowColorCoding(&wd->md, ! wd->md.itsShowColorCoding );
			break;

		case IDC_VIEW_CLIFFORD:
			SetShowCliffordParallels(&wd->md, ! wd->md.itsShowCliffordParallels );
			break;

		case IDC_VIEW_VERTEX_FIGURES:
			SetShowVertexFigures(&wd->md, ! wd->md.itsShowVertexFigures );
			break;

		case IDC_VIEW_FOG:
			SetFogFlag(&wd->md, ! wd->md.itsFogFlag );
			break;

		case IDC_VIEW_MULTISAMPLING:
			SetMultisamplingEnabledFlag(&wd->md, ! wd->md.itsMultisamplingEnabledFlag );
			DisplayChange(wd);
			break;

		case IDC_VIEW_GRAPHICS_FIXED_FUNCTION:
			SetShadersEnabledFlag(&wd->md, false);
			break;

		case IDC_VIEW_GRAPHICS_SHADERS:
			SetShadersEnabledFlag(&wd->md, true );
			break;

		case IDC_VIEW_STEREO_3D:
			SetStereo3DFlag(&wd->md, ! wd->md.itsStereoFlag );
			break;
		
		//	language menu
		
		case IDC_LANGUAGE_DE:	ChangeLanguage("de");	break;
		case IDC_LANGUAGE_EN:	ChangeLanguage("en");	break;
		case IDC_LANGUAGE_FR:	ChangeLanguage("fr");	break;
		case IDC_LANGUAGE_JA:	ChangeLanguage("ja");	break;

		//	help menu

		case IDC_HELP_HELP:
			OpenHelpBook();
			break;

		case IDC_HELP_PREFERENCES:
			DialogBoxParam(	GetModuleHandle(NULL),
							(void *) IDD_PREFERENCES_DIALOG,
							wd->itsWindow,
							(DLGPROC)PreferencesDialogProc,
							(LPARAM) wd);
			break;

		case IDC_HELP_ABOUT:
			//	Building a fancy dialog box with graphics is tricky
			//	because of the conflict between the "dialog layout units"
			//	and the pixel dimensions of bitmaps.  A simple but humble
			//	solution relays basic information in a MessageBox().
			ErrorMessage(GetLocalizedText("AboutBoxMessage"), GetLocalizedText("AboutBoxTitle"));
			break;
	}
}


void DoFileOpen(WindowData *wd)
{
	ErrorText		theErrorMessage					= NULL;
	WCHAR			theFileName[MAX_PATH]			= L"",
					*theExtension					= NULL;
	OPENFILENAME	theFileInfo						= {
						sizeof(OPENFILENAME), wd->itsWindow, NULL,
						L"Curved Spaces Generator Files (*.gen)\0*.gen\0",
						NULL, 0, 1,
						theFileName, BUFFER_LENGTH(theFileName),
						wd->itsFileTitle, BUFFER_LENGTH(wd->itsFileTitle),
						gLastManifoldDirectory, NULL, 0, 0, 0,
						L"gen", 0L, NULL, NULL};
	Byte			*theRawText						= NULL;

#ifdef BYPASS_GETOPENFILENAME
//DANGEROUS CODE:  DOES NOT CHECK BUFFER LENGTHS!
//	wcscpy(theFileName, L"E:\\Jeff\\1 General\\Software - Current\\Curved Spaces\\Curved Spaces 3\\Curved Spaces - Win\\Sample Spaces\\Basic\\3-Torus.gen");
	wcscpy(theFileName, L"E:\\Jeff\\1 General\\Software - Current\\Curved Spaces\\Curved Spaces 3\\Curved Spaces - Win\\Sample Spaces\\Basic\\HyperbolicDodecahedron.gen");
	wd->itsFileTitle[0] = 0;
#else
	//	Invite the user to select a file.
	//	If the user cancels, return with no error.
	if (GetOpenFileName(&theFileInfo) == 0)
		return;
	
	//	Remember the directory name for next time.
	CopyDirectoryName(	theFileName,
						BUFFER_LENGTH(theFileName),
						gLastManifoldDirectory,
						BUFFER_LENGTH(gLastManifoldDirectory));

	//	Remove the .gen filename extension.
	theExtension = wd->itsFileTitle;
	while (*theExtension != 0)
		theExtension++;
	if ((theExtension - wd->itsFileTitle) >= 5
	 && theExtension[-4] == '.'
	 && theExtension[-3] == 'g'
	 && theExtension[-2] == 'e'
	 && theExtension[-1] == 'n')
	{
		theExtension[-4] = 0;
	}
#endif

	//	Read the raw data from the UTF-8 or Latin-1 file.
	//	Append a terminating zero.
	theErrorMessage = ReadRawText(theFileName, &theRawText);
	if (theErrorMessage != NULL)
		goto CleanUpDoFileOpen;

	//	Read the generating matrices and set up the manifold.
	theErrorMessage = LoadGenerators(&wd->md, theRawText);
	if (theErrorMessage != NULL)
		goto CleanUpDoFileOpen;

CleanUpDoFileOpen:

	FREE_MEMORY_SAFELY(theRawText);

	if (theErrorMessage != NULL)
	{
		wd->itsFileTitle[0] = 0;
		ErrorMessage(theErrorMessage, "Couldn't open file");
	}

	//	Update the window title.
	//	If wd->itsFileTitle[0] == 0,
	//	the title will revert to "Curved Spaces".
	SetWindowTitle(wd->itsWindow, 0);
}


static ErrorText ReadRawText(
	WCHAR			*aFileName,
	Byte			**someRawText)
{
	ErrorText	theErrorMessage		= NULL;
	HANDLE		theFile				= NULL;
	DWORD		theNumBytes			= 0,
				theHighOrderBytes	= 0,
				theNumBytesRead		= 0;

	if (*someRawText != NULL)
		return "ReadRawText() was passed a non-NULL output pointer.";

	theFile = CreateFile(aFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (theFile == NULL)
	{
		theErrorMessage = "Couldn't open matrix file.";
		goto CleanUpReadRawBytes;
	}

	theNumBytes = GetFileSize(theFile, &theHighOrderBytes);
	if (theNumBytes == INVALID_FILE_SIZE
	 || theHighOrderBytes != 0)
	{
		theErrorMessage = "Couldn't get matrix file size.";
		goto CleanUpReadRawBytes;
	}
	
	*someRawText = GET_MEMORY(theNumBytes + 1);	//	Allow room for a terminating zero.
	if (*someRawText == NULL)
	{
		theErrorMessage = "Couldn't allocate memory to read matrix file.";
		goto CleanUpReadRawBytes;
	}

	if (ReadFile(theFile, *someRawText, theNumBytes, &theNumBytesRead, NULL) == 0
	 || theNumBytesRead != theNumBytes)
	{
		theErrorMessage = "Couldn't read raw bytes from matrix file.";
		goto CleanUpReadRawBytes;
	}
	
	(*someRawText)[theNumBytes] = 0;	//	terminating zero

CleanUpReadRawBytes:

	if (theFile != NULL)
		CloseHandle(theFile);

	if (theErrorMessage != NULL)
		FREE_MEMORY_SAFELY(*someRawText);

	return theErrorMessage;
}


static BOOL CALLBACK PostCloseMessage(
	HWND	aWindow,
	LPARAM	anUnusedParameter)
{
	UNUSED_PARAMETER(anUnusedParameter);

	//	Skip windows that aren't ours.
	//
	//	Note:  On my system the Input Method Environment (IME) creates
	//	secret windows that shadow the main window.  They are top-level windows,
	//	so EnumThreadWindows() finds them.  Specifically, each main window owns
	//	a secret window of class "IME", which in turn owns another secret window
	//	of class "MSCTFIME UI".
	if (GetClassLong(aWindow, GCW_ATOM) == gMainWindowClass)
		PostMessage(aWindow, WM_CLOSE, 0, 0);

	return true;	//	keep going
}


static void CopyTheImage(WindowData *wd)
{
	ErrorText	theErrorMessage		= NULL;
	HANDLE		theDibHandle		= NULL;
	bool		theClipboardIsOpen	= false;

	//	Read the image as a DIB.
	theDibHandle = FetchTheImage(wd);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Couldn't read image to copy to clipboard.";
		goto CleanUpCopyTheImage;
	}

	//	Transfer the DIB to the clipboard.

	if (OpenClipboard(NULL) == 0)
	{
		theErrorMessage = "Couldn't open clipboard.";
		goto CleanUpCopyTheImage;
	}
	theClipboardIsOpen = true;

	if (EmptyClipboard() == 0)
	{
		theErrorMessage = "Couldn't delete previous clipboard contents.";
		goto CleanUpCopyTheImage;
	}

	if (SetClipboardData(CF_DIB, theDibHandle) == NULL)
	{
		theErrorMessage = "Couldn't transfer image to clipboard.";
		goto CleanUpCopyTheImage;
	}
	//	The clipboard now owns the DIB.
	theDibHandle = NULL;

CleanUpCopyTheImage:

	if (theClipboardIsOpen == true)
		CloseClipboard();

	if (theDibHandle != NULL)
		GlobalFree(theDibHandle);

	if (theErrorMessage != NULL)
		ErrorMessage(theErrorMessage, "Copy Error");
}


static void SaveTheImage(WindowData *wd)
{
	ErrorText			theErrorMessage		= NULL;
	WCHAR				theFileName[1024]	= {0};
	OPENFILENAME		theFileInfo			= {sizeof(OPENFILENAME),
		NULL, NULL, L"32-bit Bitmap (*.bmp)\0*.bmp\0", NULL, 0, 0,
		theFileName, BUFFER_LENGTH(theFileName), NULL, 0,
		gLastImageDirectory, NULL, OFN_OVERWRITEPROMPT, 0, 0,
		L"bmp", 0L, NULL, NULL};
	HANDLE				theDibHandle		= NULL;
	unsigned int		theDIBSize			= 0;
	HANDLE				theFile				= NULL;
	BITMAPFILEHEADER	theFileHeader		= {0x4D42, 0, 0, 0, 0};
	DWORD				theNumBytes			= 0;
	Byte				*theDib				= NULL;

	//	Prompt the user for a file name.
	//	The extension will be .bmp (see theFileInfo above).
	if (GetSaveFileName(&theFileInfo) == 0)
	{
		//	The user cancelled or an error occurred.
		goto CleanUpSaveTheImage;
	}
	
	//	Remember the directory name for next time.
	CopyDirectoryName(	theFileName,
						BUFFER_LENGTH(theFileName),
						gLastImageDirectory,
						BUFFER_LENGTH(gLastImageDirectory));

	//	Read the image as a DIB.
	theDibHandle = FetchTheImage(wd);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Couldn't read image to save to file.";
		goto CleanUpSaveTheImage;
	}
	
	//	How big is the DIB?
	theDIBSize = GlobalSize(theDibHandle);
	if (theDIBSize == 0)
	{
		theErrorMessage = "Couldn't get image size.";
		goto CleanUpSaveTheImage;
	}
	
	//	Open the output file.
	theFile = CreateFile(	theFileName,
							GENERIC_WRITE,
							0,
							NULL,
							CREATE_ALWAYS,
							FILE_ATTRIBUTE_NORMAL,
							NULL);
	if (theFile == NULL)
	{
		theErrorMessage = "Couldn't create file to save image.";
		goto CleanUpSaveTheImage;
	}
	
	//	Save the DIB to the file.
	
	//		First the BITMAPFILEHEADER.
	theFileHeader.bfSize	= sizeof(BITMAPFILEHEADER) + theDIBSize;
	theFileHeader.bfOffBits	= sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	if (WriteFile(theFile, &theFileHeader, sizeof(BITMAPFILEHEADER), &theNumBytes, NULL) == 0
	 || theNumBytes != sizeof(BITMAPFILEHEADER))
	{
		theErrorMessage = "Couldn't write bitmap file header.";
		goto CleanUpSaveTheImage;
	}
	
	//		Then the DIB itself.
	theDib = (Byte *) GlobalLock(theDibHandle);
	if (theDib == NULL)
	{
		theErrorMessage = "Couldn't lock the image memory to save to file.";
		goto CleanUpSaveTheImage;
	}
	if (WriteFile(theFile, theDib, theDIBSize, &theNumBytes, NULL) == 0
	 || theNumBytes != theDIBSize)
	{
		theErrorMessage = "Couldn't write DIB to file.";
		goto CleanUpSaveTheImage;
	}
	GlobalUnlock(theDibHandle);
	theDib = NULL;

CleanUpSaveTheImage:

	if (theDib != NULL)
		GlobalUnlock(theDibHandle);

	if (theFile != NULL)
		CloseHandle(theFile);

	if (theDibHandle != NULL)
		GlobalFree(theDibHandle);

	if (theErrorMessage != NULL)
		ErrorMessage(theErrorMessage, "Save Error");
}


static bool CopyDirectoryName(
	WCHAR	*aPathName,				//	input, 	NULL-terminated string of WCHARs
	DWORD	aPathNameLength,		//	input, 	size of input  buffer in WCHARs, including terminating zero
	WCHAR	*aDirectoryName,		//	output, NULL-terminated string of WCHARs
	DWORD	aDirectoryNameLength)	//	input,	size of output buffer in WCHARs, including terminating zero
{
	//	Copy aPathName to aDirectoryName, excluding the file name itself.
	
	DWORD	thePosition,
			i;
	
	//	Will aDirectoryName accommodate a null string in case of error?
	if (aDirectoryName == NULL
	 || aDirectoryNameLength == 0)
	{
		//	Return *without* setting an empty output string.
		return false;
	}
	else
	{
		//	Preinitialize an empty output string.
		aDirectoryName[0] = 0;
	}

	//	Find the terminating zero at the end of the input path name.
	for (thePosition = 0; thePosition < aPathNameLength; thePosition++)
		if (aPathName[thePosition] == 0)
			break;
	if (thePosition == aPathNameLength)
		return false;	//	no terminating zero
	
	//	Back up to the last backslash character.
	while (thePosition > 0
	    && aPathName[thePosition] != '\\')
	{
		thePosition--;
	}
	if (thePosition == 0)
		return false;	//	no backslash found
	
	if (aDirectoryNameLength < thePosition + 2)
		return false;	//	output buffer too small
	
	//	Copy the directory name from aPathName to aDirectoryName.
	for (i = 0; i <= thePosition; i++)
		aDirectoryName[i] = aPathName[i];
	
	//	Append a terminating zero.
	aDirectoryName[thePosition + 1] = 0;
	
	//	Success
	return true;
}


static HANDLE FetchTheImage(WindowData *wd)
{
	ErrorText			theErrorMessage			= NULL;
	RECT				theClientRect			= {0,0,0,0};
	unsigned int		theImageWidth			= 0,
						theImageHeight			= 0,
						theImageSize			= 0,
						theHeaderSize			= 0,
						theTotalSize			= 0;
	HANDLE				theDibHandle			= NULL;
	Byte				*theDib					= NULL;
	BITMAPINFOHEADER	*theHeader				= NULL;
	Byte				*theImage				= NULL;
	bool				aReadFrontBufferFlag	= false;
	unsigned int		r						= 0,
						c						= 0;
	Byte				*thePixel				= NULL,
						theSwap					= 0;

	//	How big is the image?
	if (GetClientRect(wd->itsDrawingPanel, &theClientRect) == 0
	 || theClientRect.left != 0
	 || theClientRect.top  != 0
	 || (theImageWidth  = theClientRect.right)  == 0
	 || (theImageHeight = theClientRect.bottom) == 0)
	{
		theErrorMessage = "Couldn't read window size.";
		goto CleanUpFetchTheImage;
	}

	//	How many bytes do we have?
	//	(If we were reading RGB instead of RGBA we'd need to pad the rows
	//	to insure a row-stride that's a multiple of four, but for RGBA
	//	it's automatic.  Unfortunately the alpha channel will be ignored,
	//	but let's include the alpha values anyhow.)
	theImageSize = theImageWidth * theImageHeight * 4;

	//	Create a DIB.

	theHeaderSize	= sizeof(BITMAPINFOHEADER);
	theTotalSize	= theHeaderSize + theImageSize;

	theDibHandle = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, theTotalSize);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Not enough memory to copy image.";
		goto CleanUpFetchTheImage;
	}

	theDib = (Byte *) GlobalLock(theDibHandle);
	if (theDib == NULL)
	{
		theErrorMessage = "Couldn't lock the image memory to write image data.";
		goto CleanUpFetchTheImage;
	}
	theHeader	= (BITMAPINFOHEADER *) theDib;
	theImage	= theDib + theHeaderSize;

	theHeader->biSize			= sizeof(BITMAPINFOHEADER);
	theHeader->biWidth			= theImageWidth;
	theHeader->biHeight			= theImageHeight;
	theHeader->biPlanes			= 1;
	theHeader->biBitCount		= 32;
	theHeader->biCompression	= BI_RGB;
	theHeader->biSizeImage		= 0;
	theHeader->biXPelsPerMeter	= 0;
	theHeader->biYPelsPerMeter	= 0;
	theHeader->biClrUsed		= 0;
	theHeader->biClrImportant	= 0;
	
	//	When multisampling is active we need to read the front buffer,
	//	because the driver can't read the back buffer properly.
	//	The only problem with this hack is that we end up copying
	//	the Export menu's fade effect.  To work around that, repaint 
	//	the window and then sleep for 1sec before reading the buffer.
	//	(Note:  On Mac OS X reading the back buffer is fine.)
	aReadFrontBufferFlag = wd->md.itsMultisamplingAvailableFlag
						&& wd->md.itsMultisamplingEnabledFlag;
	if (aReadFrontBufferFlag)
	{
		PaintWindow(wd);
		Sleep(1000);
	}

	//	Read the pixels as RGBA.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ReadRGBAPixels(theImageWidth, theImageHeight, theImage, aReadFrontBufferFlag);
	wglMakeCurrent(NULL, NULL);

	//	Swap the byte order manually to conform to .bmp conventions.
	thePixel = theImage;
	for (r = 0; r < theImageHeight; r++)
		for (c = 0; c < theImageWidth; c++)
		{
			theSwap		= thePixel[0];
			thePixel[0]	= thePixel[2];
			thePixel[2]	= theSwap;
			
			thePixel += 4;
		}

	//	Unlock the handle.
	GlobalUnlock(theDibHandle);
	theDib = NULL;

CleanUpFetchTheImage:

	if (theDib != NULL)
		GlobalUnlock(theDibHandle);

	if (theErrorMessage != NULL)
	{
		if (theDibHandle != NULL)
		{
			GlobalFree(theDibHandle);
			theDibHandle = NULL;
		}

		ErrorMessage(theErrorMessage, "Image Export Error");
	}
	
	return theDibHandle;
}


static void ChangeLanguage(
	const char	aLanguageCode[3])
{
	RefreshLocalization(aLanguageCode);
	EnumThreadWindows(GetCurrentThreadId(), RefreshLanguage, 0);
}


BOOL CALLBACK RefreshLanguage(
	HWND	aWindow,
	LPARAM	anUnusedParameter)
{
	HMENU	theOldMenu,
			theNewMenu;

	UNUSED_PARAMETER(anUnusedParameter);

	//	Skip windows that aren't ours.
	//
	//	Note:  On my system the Input Method Environment (IME) creates
	//	secret windows that shadow the main window.  They are top-level windows,
	//	so EnumThreadWindows() finds them.  Specifically, each main window owns
	//	a secret window of class "IME", which in turn owns another secret window
	//	of class "MSCTFIME UI".
	if (GetClassLong(aWindow, GCW_ATOM) != gMainWindowClass)
		return true;

	//	Reset the window's title.
	//	(If a file is open, the title won't change.)
	SetWindowTitle(aWindow, 0);

	//	Reset the mirroring.
	RefreshMirroring(aWindow);

	//	Replace the window's old menu with a new menu in the new language.
	//	Note:  LoadMenu() creates a private menu for each window,
	//	even though they are all based on the same resource.
	//	Thus we may safely translate the new menu and destroy the old one.
	theOldMenu = GetMenu(aWindow);
	theNewMenu = BuildLocalizedMenuBar();
	SetMenu(aWindow, theNewMenu);
	DestroyMenu(theOldMenu);
	DrawMenuBar(aWindow);

	return true;	//	keep going
}


static void RefreshMirroring(HWND aWindow)
{
	//	Reset the window layout, left-to-right (LTR) or right-to-left (RTL).

	if (LanguageReadsRightToLeft())
	{
		//	Mirror the main window.
		//	Mirror child windows only if you want to -- it isn't necessary.
		SetWindowLongPtr(	aWindow,
							GWL_EXSTYLE,
							GetWindowLongPtr(aWindow, GWL_EXSTYLE) |  WS_EX_LAYOUTRTL);
	}
	else
	{
		SetWindowLongPtr(	aWindow,
							GWL_EXSTYLE,
							GetWindowLongPtr(aWindow, GWL_EXSTYLE) & ~WS_EX_LAYOUTRTL);
	}
}


static HMENU BuildLocalizedMenuBar(void)
{
	HMENU	theMainMenu		= NULL,
			theSubMenu		= NULL,
			theSubSubMenu	= NULL;
	WCHAR	theBuffer[2048];

	theMainMenu = CreateMenu();
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("File", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_FILE_OPEN,     GetLocalizedTextAsWCHAR("Open…",     theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_FILE_OPEN_NEW, GetLocalizedTextAsWCHAR("Open New…", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                 NULL                                                                     );
	theSubSubMenu = CreateMenu();
		AppendMenu(theSubMenu, MF_POPUP, (UINT_PTR)theSubSubMenu, GetLocalizedTextAsWCHAR("Export", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_FILE_EXPORT_COPY, GetLocalizedTextAsWCHAR("Copy Image",  theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_FILE_EXPORT_SAVE, GetLocalizedTextAsWCHAR("Save Image…", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = NULL;
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                 NULL                                                                     );
	AppendMenu(theSubMenu, MF_STRING,    IDC_FILE_EXIT,     GetLocalizedTextAsWCHAR("Exit",      theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("View", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = CreateMenu();
		AppendMenu(theSubMenu, MF_POPUP | MF_STRING, (UINT_PTR)theSubSubMenu, GetLocalizedTextAsWCHAR("Centerpiece", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_CENTERPIECE_NONE,      GetLocalizedTextAsWCHAR("None",      theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_CENTERPIECE_EARTH,     GetLocalizedTextAsWCHAR("Earth",     theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_CENTERPIECE_GALAXY,    GetLocalizedTextAsWCHAR("Galaxy",    theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_CENTERPIECE_GYROSCOPE, GetLocalizedTextAsWCHAR("Gyroscope", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = NULL;
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_OBSERVER,       GetLocalizedTextAsWCHAR("Spaceship",          theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_COLOR_CODING,   GetLocalizedTextAsWCHAR("Color Coding",       theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_CLIFFORD,       GetLocalizedTextAsWCHAR("Clifford Parallels", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_VERTEX_FIGURES, GetLocalizedTextAsWCHAR("Vertex Figures",     theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                       NULL                                                                              );
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_FOG,            GetLocalizedTextAsWCHAR("Fog",                theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_MULTISAMPLING,  GetLocalizedTextAsWCHAR("Smooth Edges",       theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = CreateMenu();
		AppendMenu(theSubMenu, MF_POPUP | MF_STRING, (UINT_PTR)theSubSubMenu, GetLocalizedTextAsWCHAR("Graphics Pipeline", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_GRAPHICS_FIXED_FUNCTION, GetLocalizedTextAsWCHAR("Fixed-Function", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_GRAPHICS_SHADERS,        GetLocalizedTextAsWCHAR("Shaders",        theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = NULL;
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_STEREO_3D,      GetLocalizedTextAsWCHAR("Stereoscopic 3D",    theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Language", theBuffer, BUFFER_LENGTH(theBuffer)));
	//	To add a new language, please see the instructions
	//	in the file "To add a new language.txt".
//	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_DE, L"Deutsch"  );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_EN, L"English"  );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_FR, L"Français" );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_JA, L"日本語"      );
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Help", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_HELP,        GetLocalizedTextAsWCHAR("Help",                 theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_PREFERENCES, GetLocalizedTextAsWCHAR("Preferences…",         theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                    NULL                                                                                );
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_ABOUT,       GetLocalizedTextAsWCHAR("About Curved Spaces…", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	return theMainMenu;
}


static void OpenHelpBook(void)
{
	char	theFileName[64]			= "",
			theFilePath[2048]		= "";
	WCHAR	theFilePathUTF16[2048]	= L"";

	sprintf(theFileName, "CurvedSpacesWelcome-%c%c.html", gLanguageCode[0], gLanguageCode[1]);

	if (GetAbsolutePath("Help", theFileName, theFilePath, BUFFER_LENGTH(theFilePath)) == NULL
	 && MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, theFilePath, -1, theFilePathUTF16, BUFFER_LENGTH(theFilePathUTF16)) != 0)
	{
		ShellExecute(	NULL,
						L"open",
						theFilePathUTF16,
						NULL,
						NULL,
						SW_SHOWNORMAL);
	}
}


static INT_PTR CALLBACK	PreferencesDialogProc(
	HWND	hDlg,
	UINT	msg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData		*wd				= NULL;
	HWND			theTabControl;
	unsigned int	i;
	
	if (msg != WM_INITDIALOG)
	{
		wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);
		if (wd == NULL)	//	should never occur for messages that interest us
			return false;	//	message ignored
	}

	switch (msg)
	{
		case WM_INITDIALOG:

			//	lParam points to the parent window's WindowData.
			//	Stash a copy in the dialog box as well.
			wd = (WindowData *) lParam;
			SetWindowLongPtr(hDlg, GWLP_USERDATA, (LONG_PTR) wd);
			RefreshMirroring(hDlg);
			if (wd == NULL)	//	should never occur
				return true;	//	set focus normally

			theTabControl = GetDlgItem(hDlg, IDC_PREFS_TAB_CONTROL);

			for (i = 0; i < 2; i++)
				SendMessage(theTabControl, TCM_INSERTITEM, i, (LPARAM) &(TCITEM){0, 0, 0, NULL, 0, -1, 0});

			SendMessage(theTabControl, TCM_SETCURSEL, gPreferencesPreferredPage, 0);

			SetPrefPanelLabels(hDlg);

			SwapPreferencesPanel(wd, hDlg, gPreferencesPreferredPage);

			return true;	//	set focus normally

		case WM_DESTROY:
			wd->itsPreferencesPane = NULL;
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		case WM_NOTIFY:
			switch (((LPNMHDR)lParam)->code)
			{
				case TCN_SELCHANGE:
					gPreferencesPreferredPage = SendMessage( ((LPNMHDR)lParam)->hwndFrom, TCM_GETCURSEL, 0, 0 );
					SwapPreferencesPanel(wd, hDlg, gPreferencesPreferredPage);
					break;
			}
			//	no specific return value required
			return true;	//	message processed

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDCANCEL:	//	the Close Box sends IDCANCEL
					EndDialog(hDlg, true);
					break;
			}
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		default:
			return false;	//	message ignored
	}
}


static void SetPrefPanelLabels(
	HWND	aPreferencesDialog)
{
	WCHAR	theCaption[128]	= L"",
			theLabel[64]	= L"";
	HWND	theTabControl	= NULL;

	//	Set the window caption.
	SetWindowText(aPreferencesDialog,
		GetLocalizedTextAsWCHAR("PrefTitle", theCaption, BUFFER_LENGTH(theCaption)));

	//	Set the tab labels.
	theTabControl = GetDlgItem(aPreferencesDialog, IDC_PREFS_TAB_CONTROL);
	SendMessage(theTabControl, TCM_SETITEM, 0, (LPARAM) &(TCITEM){TCIF_TEXT, 0, 0, 
		GetLocalizedTextAsWCHAR("PrefPerspective", theLabel, BUFFER_LENGTH(theLabel)), 0, -1, 0});
	SendMessage(theTabControl, TCM_SETITEM, 1, (LPARAM) &(TCITEM){TCIF_TEXT, 0, 0, 
		GetLocalizedTextAsWCHAR("PrefAdvanced",    theLabel, BUFFER_LENGTH(theLabel)), 0, -1, 0});
}


static void SwapPreferencesPanel(
	WindowData		*wd,
	HWND			aPreferencesDialog,
	unsigned int	aPageIndex)
{
	static unsigned int	theResourceId[2] =
						{
							IDD_PREFS_PERSPECTIVE_DIALOG,
							IDD_PREFS_ADVANCED_DIALOG
						};
	static DLGPROC		thePaneProcedure[4] =
						{
							PrefsPerspectiveDialogProc,
							PrefsAdvancedDialogProc
						};

	if (wd->itsPreferencesPane != NULL)
	{
		DestroyWindow(wd->itsPreferencesPane);
		wd->itsPreferencesPane = NULL;
	}

	wd->itsPreferencesPane = CreateDialogParam(	GetModuleHandle(NULL),
												(LPCTSTR) theResourceId[aPageIndex],
												aPreferencesDialog,
												thePaneProcedure[aPageIndex],
												(LPARAM) wd);
	InvalidateRect(aPreferencesDialog, NULL, true);
}


static INT_PTR CALLBACK PrefsPerspectiveDialogProc(
	HWND	hDlg,
	UINT	msg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData		*wd;
	WCHAR			theBuffer[11],
					theLabel[256];
	unsigned int	theNumber;
	RECT			theClientRect;

	switch (msg)
	{
		case WM_INITDIALOG:

			//	lParam contains a pointer to the parent window's WindowData.
			//	Stash a copy in this dialog box as well.
			wd = (WindowData *) lParam;
			SetWindowLongPtr(hDlg, GWLP_USERDATA, (LONG_PTR) wd);
			if (wd == NULL)	//	should never occur
				return true;	//	set focus normally

			//	Initialize the stereo 3D checkbox and
			//	enable corresponding text fields accordingly.
			EnableStereoFields(hDlg, wd->md.itsStereoFlag);

			//	Copy our internal variables into the dialog's edit fields.

			FormatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), (unsigned int) (2 * wd->md.itsWindowHalfWidth));
			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_WIDTH   ), theBuffer);

			FormatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), (unsigned int) wd->md.itsWindowDistance);
			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_DISTANCE), theBuffer);

			FormatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), (unsigned int) (2 * wd->md.itsEyeOffset));
			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_EYE_SPACING    ), theBuffer);

			FormatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), (unsigned int) wd->md.itsIntrinsicUnit);
			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_INTRINSIC_UNIT ), theBuffer);

			//	Initialize the labels in the current language.

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_GENERAL_LABEL),
				GetLocalizedTextAsWCHAR("PrefGeneral",        theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_WIDTH_LABEL),
				GetLocalizedTextAsWCHAR("PrefWindowWidth",    theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_DISTANCE_LABEL),
				GetLocalizedTextAsWCHAR("PrefWindowDistance", theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_STEREO_3D),
				GetLocalizedTextAsWCHAR("PrefStereo3D",       theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_EYE_SPACING_LABEL),
				GetLocalizedTextAsWCHAR("PrefEyeSpacing",     theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_INTRINSIC_UNIT_LABEL),
				GetLocalizedTextAsWCHAR("PrefIntrinsicUnit",  theLabel, BUFFER_LENGTH(theLabel)));

			return true;	//	set focus normally

		case WM_COMMAND:

			wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);
			if (wd != NULL)
			{
				switch (LOWORD(wParam))
				{
					case IDC_PREFS_PERSPECTIVE_STEREO_3D:
						SetStereo3DFlag(&wd->md, ! wd->md.itsStereoFlag );
						EnableStereoFields(hDlg, wd->md.itsStereoFlag);
						break;

					case IDC_PREFS_PERSPECTIVE_WINDOW_WIDTH:

						GetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_WIDTH   ), theBuffer, BUFFER_LENGTH(theBuffer));
						UnformatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), &theNumber);
						wd->md.itsWindowHalfWidth = ((double) theNumber) / 2;

						//	Use the known aspect ratio to deduce
						//	itsWindowHalfHeight from itsWindowHalfWidth.
						//	Note that theClientRect always has left = top = 0,
						//	so its right equals its width and its bottom equals its height.
						if (wd->itsWindow != NULL
						 && GetClientRect(wd->itsWindow, &theClientRect) != 0
						 && theClientRect.right > 0
						 && theClientRect.bottom > 0)
						{
							wd->md.itsWindowHalfHeight = wd->md.itsWindowHalfWidth
								* (double)theClientRect.bottom / (double)theClientRect.right;
						}
						else	//	should never occur
							wd->md.itsWindowHalfHeight = wd->md.itsWindowHalfWidth;

						break;

					case IDC_PREFS_PERSPECTIVE_WINDOW_DISTANCE:
						GetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_DISTANCE), theBuffer, BUFFER_LENGTH(theBuffer));
						UnformatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), &theNumber);
						wd->md.itsWindowDistance = theNumber;
						break;

					case IDC_PREFS_PERSPECTIVE_EYE_SPACING:
						GetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_EYE_SPACING    ), theBuffer, BUFFER_LENGTH(theBuffer));
						UnformatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), &theNumber);
						wd->md.itsEyeOffset = ((double) theNumber) / 2;
						break;

					case IDC_PREFS_PERSPECTIVE_INTRINSIC_UNIT:
						GetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_INTRINSIC_UNIT ), theBuffer, BUFFER_LENGTH(theBuffer));
						UnformatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), &theNumber);
						wd->md.itsIntrinsicUnit = theNumber;
						break;
				}
				PaintWindow(wd);
			}
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		default:
			return false;	//	message ignored
	}
}


static void EnableStereoFields(
	HWND	aDialog,
	bool	aStereoFlag)
{
	SendMessage(	GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_STEREO_3D),
					BM_SETCHECK,
					aStereoFlag ? BST_CHECKED : BST_UNCHECKED,
					0);

	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_EYE_SPACING_LABEL),		aStereoFlag);
	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_EYE_SPACING),			aStereoFlag);
	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_EYE_SPACING_UNIT),		aStereoFlag);

	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_INTRINSIC_UNIT_LABEL),	aStereoFlag);
	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_INTRINSIC_UNIT),			aStereoFlag);
	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_INTRINSIC_UNIT_UNIT),	aStereoFlag);
}


static void FormatUnsignedInteger32(
	WCHAR			*aBuffer,
	unsigned int	aBufferLength,
	unsigned int	aNumber)
{
	unsigned int	theNumDigits;
	WCHAR			theString[10];

	//	Format aNumber into local memory, with digits reversed.
	//	At most ten decimal digits will be required for a 32-bit unsigned integer.
	theNumDigits = 0;
	while (aNumber)
	{
		theString[theNumDigits] = (WCHAR)('0' + aNumber % 10);
		aNumber /= 10;
		theNumDigits++;
	}

	//	Logically zero is a 0-digit number,
	//	but let's make a special case an write it as a 1-digit number.
	if (theNumDigits == 0)
	{
		theString[0] = (WCHAR) '0';
		theNumDigits = 1;
	}

	//	Is the input buffer OK?
	if (aBuffer == NULL)
	{
		ErrorMessage("NULL buffer passed to FormatUnsignedInteger32().", "Bad Parameters");
		return;
	}
	if (aBufferLength < theNumDigits + 1)
	{
		ErrorMessage("Buffer too small in FormatUnsignedInteger32().", "Bad Parameters");
		return;
	}

	//	Copy the digits from theString into aBuffer, suppressing leading zeros.
	while (theNumDigits-- > 0)
		*aBuffer++ = theString[theNumDigits];
	*aBuffer = 0;	//	write a terminating 0
}


static void UnformatUnsignedInteger32(
	WCHAR			*aBuffer,
	unsigned int	aBufferLength,
	unsigned int	*aNumber)
{
	WCHAR	*theEnd;	//	will point to the first character past the end of aBuffer

	//	Initialize.
	*aNumber = 0;
	theEnd = aBuffer + aBufferLength;

	//	Scan past any initial white space and/or other junk.
	//	A negative sign is also intentionally ignored.
	while (aBuffer < theEnd
		&& (*aBuffer < ((WCHAR)'0') || *aBuffer > ((WCHAR)'9')))
	{
		aBuffer++;
	}

	//	Read the digits, if any.
	while (aBuffer < theEnd
		&& (*aBuffer >= ((WCHAR)'0') && *aBuffer <= ((WCHAR)'9')))
	{
		*aNumber *= 10;
		*aNumber += *aBuffer - ((WCHAR)'0');
		aBuffer++;
	}
}


static INT_PTR CALLBACK PrefsAdvancedDialogProc(
	HWND	hDlg,
	UINT	msg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData	*wd;
	WCHAR		theLabel[256];

	switch (msg)
	{
		case WM_INITDIALOG:

			//	lParam contains a pointer to the parent window's WindowData.
			//	Stash a copy in this dialog box as well.
			wd = (WindowData *) lParam;
			SetWindowLongPtr(hDlg, GWLP_USERDATA, (LONG_PTR) wd);
			if (wd == NULL)	//	should never occur
				return true;	//	set focus normally

			//	Initialize the checkboxes.

			EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FOG),
								wd->md.itsShadersAvailableFlag
							 && wd->md.itsShadersEnabledFlag);
			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FOG),
							BM_SETCHECK,
							(	wd->md.itsShadersAvailableFlag
							 && wd->md.itsShadersEnabledFlag
							 && wd->md.itsFogFlag	)
								? BST_CHECKED : BST_UNCHECKED,
							0);

			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FRAME_RATE),
							BM_SETCHECK,
							gShowFrameRate ? BST_CHECKED : BST_UNCHECKED,
							0);

			EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SYNC_TO_MONITOR),
							wglSwapInterval != NULL);	//	wglSwapInterval may be non-NULL but ineffective
			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SYNC_TO_MONITOR),
							BM_SETCHECK,
							(	wglSwapInterval
							 && wd->md.itsSyncFlag	)
								? BST_CHECKED : BST_UNCHECKED,
							0);

			EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_MULTISAMPLING),
							wd->md.itsMultisamplingAvailableFlag);
			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_MULTISAMPLING),
							BM_SETCHECK,
							(	wd->md.itsMultisamplingAvailableFlag
							 && wd->md.itsMultisamplingEnabledFlag	)
								? BST_CHECKED : BST_UNCHECKED,
							0);

			EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_ANISOTROPIC),
							wd->md.itsAnisotropicAvailableFlag);
			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_ANISOTROPIC),
							BM_SETCHECK,
							(	wd->md.itsAnisotropicAvailableFlag
							 && wd->md.itsAnisotropicEnabledFlag	)
								? BST_CHECKED : BST_UNCHECKED,
							0);

			EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SHADERS),
							wd->md.itsShadersAvailableFlag);
			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SHADERS),
							BM_SETCHECK,
							(	wd->md.itsShadersAvailableFlag
							 && wd->md.itsShadersEnabledFlag	)
								? BST_CHECKED : BST_UNCHECKED,
							0);

			//	Initialize the labels in the current language.

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FOG),
				GetLocalizedTextAsWCHAR("PrefFog",           theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FRAME_RATE),
				GetLocalizedTextAsWCHAR("PrefFrameRate",     theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SYNC_TO_MONITOR),
				GetLocalizedTextAsWCHAR("PrefSync",          theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_MULTISAMPLING),
				GetLocalizedTextAsWCHAR("PrefMultisampling", theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_ANISOTROPIC),
				GetLocalizedTextAsWCHAR("PrefAnisotropic",   theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SHADERS),
				GetLocalizedTextAsWCHAR("PrefShaders",       theLabel, BUFFER_LENGTH(theLabel)));

			return true;	//	set focus normally

		case WM_COMMAND:

			wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);
			if (wd != NULL)
			{
				switch (LOWORD(wParam))
				{
					case IDC_PREFS_ADVANCED_FOG:
						SetFogFlag(&wd->md, ! wd->md.itsFogFlag );
						break;

					case IDC_PREFS_ADVANCED_FRAME_RATE:
						gShowFrameRate = ! gShowFrameRate;
						EnumThreadWindows(	GetCurrentThreadId(),
											SetWindowTitle,
											(LPARAM) (gShowFrameRate ? L"…" : NULL));
						break;

					case IDC_PREFS_ADVANCED_SYNC_TO_MONITOR:
						wd->md.itsSyncFlag = ! wd->md.itsSyncFlag;
						SetSync(wd);
						break;

					case IDC_PREFS_ADVANCED_MULTISAMPLING:
						SetMultisamplingEnabledFlag(&wd->md, ! wd->md.itsMultisamplingEnabledFlag );
						DisplayChange(wd);
						break;

					case IDC_PREFS_ADVANCED_ANISOTROPIC:
						SetAnisotropicEnabledFlag(&wd->md, ! wd->md.itsAnisotropicEnabledFlag);
						break;

					case IDC_PREFS_ADVANCED_SHADERS:

						SetShadersEnabledFlag(&wd->md, ! wd->md.itsShadersEnabledFlag );

						//	Fog is available only when shaders are enabled.
						EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FOG),
											wd->md.itsShadersAvailableFlag
										 && wd->md.itsShadersEnabledFlag);
						SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FOG),
										BM_SETCHECK,
										(	wd->md.itsShadersAvailableFlag
										 && wd->md.itsShadersEnabledFlag
										 && wd->md.itsFogFlag	)
											? BST_CHECKED : BST_UNCHECKED,
										0);

						break;
				}
				PaintWindow(wd);
			}
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		default:
			return false;	//	message ignored
	}
}


static void SetSync(WindowData *wd)
{
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	if (wglSwapInterval != NULL)	//	wglSwapInterval may be non-NULL but ineffective
		wglSwapInterval(wd->md.itsSyncFlag ? 1 : 0);
	wglMakeCurrent(NULL, NULL);
}


static void SteerWithMouse(
	WindowData	*wd,
	WPARAM		wParam,
	LPARAM		lParam,
	bool		aSteeringFlag)	//	steer?  or just get set up?
{
	RECT	theWindowRect;
	POINT	theWindowCenter,
			theCursorPosition;
	double	dx,
			dy;

	UNUSED_PARAMETER(lParam);

	if (GetWindowRect(wd->itsDrawingPanel, &theWindowRect) == false
	 || GetCursorPos(&theCursorPosition) == false)
		return;

	theWindowCenter.x = (theWindowRect.right  + theWindowRect.left)/2;
	theWindowCenter.y = (theWindowRect.bottom + theWindowRect.top )/2;

	if (theCursorPosition.x != theWindowCenter.x
	 || theCursorPosition.y != theWindowCenter.y)
	{
		dx =   UNITS_PER_PIXEL * (theCursorPosition.x - theWindowCenter.x);
		dy = - UNITS_PER_PIXEL * (theCursorPosition.y - theWindowCenter.y);

		SetCursorPos(theWindowCenter.x, theWindowCenter.y);

		if (aSteeringFlag)
			MouseMove(	&wd->md,
						dx,
						dy,
						wParam & MK_SHIFT,				//	shift key down?
						wParam & MK_CONTROL,			//	control key down?
						GetKeyState(VK_MENU) & 0x8000);	//	alt key down?
	}
}


static void SetMinWindowSize(
	HWND	aWindow,
	POINT	*aMinSize)
{
	RECT	theRect;

	// Prevent the client area from going smaller than MIN_CLIENT_SIZE.
	SetRect(&theRect, 0, 0, MIN_CLIENT_SIZE, MIN_CLIENT_SIZE);
	AdjustWindowRect(&theRect, (DWORD) GetWindowLongPtr(aWindow, GWL_STYLE), true);
	aMinSize->x = theRect.right - theRect.left;
	aMinSize->y = theRect.bottom - theRect.top;
}


static void ResizeDrawingPanel(
	WindowData		*wd,
	unsigned int	aWidth,
	unsigned int	aHeight)
{
	//	Let OpenGL resize the viewport.
	//	Adjust the environmental parameters while we're at it.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ResizeGLWindow(&wd->md, aWidth, aHeight);
	wglMakeCurrent(NULL, NULL);

	//	Request a refresh.
	InvalidateRect(wd->itsDrawingPanel, NULL, false);
}


static void DisplayChange(WindowData *wd)
{
	//	Maybe the user changed the color depth, screen size or multisampling,
	//	or maybe it's just that the screen saver had been running
	//	and is now passing control back to our program.
	//	Either way, we need to refresh our OpenGL context, textures
	//	and display lists.

	//	Free the old OpenGL context along with its textures and display lists.
	ShutDownOpenGL(wd);

	//	Create a new OpenGL context with new textures and display lists.
	SetUpOpenGL(wd);
}


static void PaintWindow(WindowData *wd)
{
	ErrorText	theSetUpError				= NULL,
				theRenderError				= NULL;
	HDC			theSavedDeviceContext		= NULL;
	HGLRC		theSavedRenderingContext	= NULL;

	//	In the normal flow of events there's no need to save and restore
	//	the rendering context.  However, we may get "unexpected" WM_PAINT
	//	messages -- for example if a MessageBox takes it upon itself
	//	to send WM_PAINT messages to underlying windows -- so play it safe
	//	and leave the rendering context as we found it.

	theSavedDeviceContext		= wglGetCurrentDC();		//	typically NULL
	theSavedRenderingContext	= wglGetCurrentContext();	//	typically NULL
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
		theSetUpError = SetUpGraphicsAsNeeded(&wd->md);
		if (theSetUpError == NULL)
			theRenderError = Render(&wd->md);
		if (theSetUpError == NULL && theRenderError == NULL)
			SwapBuffers(wd->itsDeviceContext);
	wglMakeCurrent(theSavedDeviceContext, theSavedRenderingContext);

	ValidateRect(wd->itsDrawingPanel, NULL);

	if (theSetUpError != NULL)
	{
		ErrorMessage(theSetUpError, "OpenGL Setup Error");
		ExitProcess(17);
	}
	if (theRenderError != NULL)
	{
		ErrorMessage(theRenderError, "OpenGL Rendering Error");
		ExitProcess(18);
	}
}
